; x_lib 08894x - ejn (blueznl)
; pb 5.50
; generic functions
; bluez@xs4all.nl
;
; - collection of universal routines
;
#x_lib_version = "08894x"
#x_lib_copyright = "c2005..2017 ejn (blueznl)"
;
; organised in the following blocks:
;
; - generic functions
; - numeric functions
; - boolean and bitwise
; - string and string related
; - memory related
; - patterns / parsing / field
; - files
; - filesystem
; - registry
; - console
; - windows and desktop
; - icons
; - gadgets / toolbars / buttons
; - mouse and pointers
; - requesters
; - other gui related
; - threads / tasks / mailslots / timers / execute
; - instances / elevation / system folders / shortcuts
; - logfiles
; - network
; - other
;
Declare.i x_parse(string.s,search.s,flags.i=7)
Declare.i x_parsenext()
Declare.i x_parsefield(string.s,search.s,flags.i=7)
Declare.i x_totaldesktops()


; *** generic

Procedure x_history()                                                ; all changes thus far
  ;
  ; to do
  ;
  ; current
  ;
  ; - added x_exfileopen() and similar routines to handle delayed response caused by networked servers, virus scanners, and parallel (file locking) applications
  ; - renamed all low level file operations from x_... to x_ll_..., for example x_openfile() became x_ll_openfile()
  ; - added some numbers to x_lib_test() code
  ;
  ; snapshot 07512
  ;
  ; - small documentation updates
  ; - bugfix x_matchpatternx()
  ; - added mode 14 and 15 to x_stringtostring()
  ; - added \state field to x_dir_file structure
  ; - added x_validpathname()
  ; - added modes 5 and 6 to x_dir()
  ; - replaced x_dir_file.s() with x_dir_file()\name to support types and depth fields
  ; - fixed x_absolutepath()
  ; - added x_diskspace()
  ; - fixed x_registry_get on x64 systems
  ; - added mode 13 to x_stringtostring()
  ; - added x_sizeof()
  ; - added x_normalizestep()
  ; - modified x_strex()
  ; - added x_runprogram()
  ; - fixed more win64 issues
  ; - removed most dpi related stuff, didn't make too much sense
  ; - added x_ping(), x_threadping()
  ; - rewrote x_pokehex()
  ; - added x_wakeonlan()
  ; - another mode added to x_stringtostring()
  ;
  ; snapshot 06907
  ;
  ; - fixex x_window_todesktop()
  ; - added default parameters to x_log_window()
  ; - added flags to x_log_window()
  ; - improved x_littlehelp()
  ; - added x_desktopx() x_desktopy() x_desktopheight() x_desktopwidth() x_loadfont()
  ; - some changes in x_window_todesktop() x_totaldesktops()
  ; - added x_setdpimode()
  ; - added x_osscale.f and x_osrescale.f and modified some routines, all to handle better dpi scaling
  ; - removed x_sysinfo(), code moved into x_init()
  ; - x_window_getdesktop() now based on windows center instead of left upper corner
  ; - x_window_todesktop() now supports maximized screens
  ; - implemented pointer parameter for x_littlehelp()
  ;
  ; snapshot 06690
  ;
  ; - converted winsock routines from the ws_lib file to x_lib and purebasic 4.5x
  ; - added some routines from rescator (elevate, 64 bits test etc.)
  ;
  ; snapshot 06559
  ;
  ; - added x_getipaddress() x_gethostname() x_getmacaddress()
  ; - bugfix x_stringtostring() mode 10
  ; - added x_windowmousex() x_windowmousey()
  ; - added x_editorgadget_select() and related functions to manipulate text inside an editor gadget
  ; - x_peekb() deprecated, as purebasic now offers the peeka() function
  ; - added x_parse_match.s to x_parse()
  ; - added #x_parse_chars option to x_parse()
  ; - unc paths might :-) work in x_getdeepestpartpath() x_getparentpathpart()
  ; - x_getdeepestpartpath() and x_getparentpathpart() now return empty strings if they failed
  ; - changed x_getdeepestpath() and x_getparentpath() to x_getdeepestpathpart() and x_getparentpathpart()
  ; - added x_modifypath()
  ; - added copytoconsole option to x_log_set()
  ; - added excludeconsole flag to x_log()
  ; - reorganized the x_lib (it was getting a bit messy)
  ; - added x_commandcount() x_commandfield()
  ; - added x_near()
  ; - added x_folderexist()
  ; - added x_createdirectory()
  ; - bugfix x_getabsolutepath) x_getdeepestpathpart() x_getparentpathpart()
  ; - added mode '10' to x_stringtostring()
  ;
  ; snapshot 06186
  ;
  ; - added x_writepreferencestringex() and x_readpreferencestringex()
  ; - fixed all EnableExplicit failures
  ; - fixed bug in x_log_int()
  ; - all converted to 4.40b3
  ; - codecaddy works with the rewritten x_lib
  ; - x_filefolder() bugfix, did not clear list on single entry ('file')
  ; - many modifications to predict unicode behaviour
  ; - changed the naming of the mailslot routines
  ; - verified all code for unicode (in)compatibility
  ; - verified all code for 4.30 (in)compability
  ; - created x_lib_test() for simplifying future testing
  ; - replaced x_log_int() with x_log_int()
  ; - bugfix x_readstring() in unicode mode
  ; - removed x_setspingadget() and x_rollspingadget(), there's now the @PB_Spin_Numeric flag
  ; - x_setlistviewgadget() no longer necessary as SetGadgetItemText() no longer scrolls with listviewicons
  ; - minor improvement x_blockgadget()
  ; - modified x_waitkeywindow()
  ; - added desktops(), x_desktop_x y width and height replaced with x_desktops_x y width height note the s)
  ; - x_window_todesktop() can now specify a desktop (aka monitor)
  ; - removed x_multimonitorsupport variable
  ; - changed x_window_getmonitor() to x_window_getdesktop()
  ; - removed some multi monitor detection stuff that 4.30 now has build in
  ; - bugfix x_matchpattern()
  ; - added backslash to paths returned by x_dir() and brethern
  ; - registry routines fully rewritten, changed parameter sequence
  ; - x_getcurrentpath() removed, use GetCurrentDirectory() instead
  ; - x_matchpattern() bugfix, wrong pointer and string returned if pattern generated partial and full matches
  ;
  ; save 05700
  ;
  ; - fixed x_getlowestpathpart() and changed it to x_getdeepestpathpartpart()
  ; - i've choosen to convert all types to the platform native one (integer aka .i) unless a quad is explictly required
  ; - convert to 4.30 or newer
  ; - bugfix x_sendmessagetomailslot() mode parameter now works
  ; - fixed x_window_todesktop(), now supports unicode (structure in 4.20 only existed in non-unicode form)
  ; - added macro x_log_debug()
  ; - added check to x_exist() to detect empty filenames
  ; - bugfixes x_dir() and x_dir_filefolder(), were broken after conversion to pb4.xx and adding default parameters
  ; - removed mode 5 from x_dir()
  ; - renamed getlowestfolderpart() to getlowestpathpart()
  ; - added x_getlasterror()
  ; - fixed x_findwindow() was missing name and repeated calls
  ; - added x_runprogram()
  ;
  ; save 05447
  ;
  ; - started rewrite to pb4.10
  ; - added x_toolbar_togglebutton() again... saves typing
  ; - changed x_toolbar_togglebutton() and x_toolbar_radiobutton(), they now use toolbar number instead of handle
  ; - several bugfixes
  ; - this is actually the first pb4.10 version of x_lib that the first pb4.10 version of codecaddy uses!
  ;
  ; save 05430
  ;
  ; - rewrote api file handling x_read() x_write() etc.
  ; - encryption stuff removed (have to rethink)
  ; - changed x_findwindow()
  ; - removed x_button_togglebutton(), see buttongadget() options
  ; - removed x_toolbar_getbuttonstate(), see purebasic's gettoolbarbuttonstate()
  ; - removed x_toolbar_setbuttonstate(), settoolbarbuttonstate() handles that now
  ; - removed x_toolbar_togglebutton(), toolbarimagebutton() now has a toggle mode
  ; - improved documentation x_editorgadget()
  ; - removed x_getlisticoncolumntitle(), use getgadgetitemtext()
  ; - removed x_setlisticoncolumntitle(), use setgadgetitemtext()
  ; - changed silentrequester(), now needs window number
  ; - bigfix x_dir_simple(), removed . and .. entries that were included with 4.10
  ; - renamed x_disablegadget() in x_blockgadget() and rewrote the procedure a little
  ; - rewrote x_rol(), x_ror(), x_rolb() and x_rorb() to use assembly
  ; - added x_openfile() etc. to support shared files
  ; - removed all x_gg() routines (needs a full rewrite)
  ; - changed x_window_to_primary()
  ; - removed x_window_interiorsize()
  ; - removed x_toolbarheight() x_menuheight() x_statusbarheight(), all replaced by purebasic primitives
  ; - removed x_absf()
  ; - changed x_filefolder() to x_dir_filefolder()
  ;
  ; save 04899
  ;
  ; - added x_getlowestfolderpart()
  ; - changed x_dir_file() array to x_dir_file() list, changed parameters for x_dir()
  ; - x_registry_get() routines now return succes or not, contents of the key are now in x_registry_l and x_registry_s
  ; - added x_registry_getx() x_registrysetx() etc.
  ; - x_reg() replaced by x_registry() routines
  ; - x_command() and all related removed
  ; - x_peeks() x_pokes() changed
  ; - remove all homebuild fastfile routines
  ; - 1400 lines converted... the easy ones thus far...
  ; - started rewrite to pb 4.10
  ;
  ; save 04405
  ;
  ; - modified x_window_forcetofront()
  ; - added x_sendmessagetomailslot()
  ; - removed x_log_copytohandle() stuff and replaced it with x_log_copytomailslot() functionality
  ; - modified x_matchpatternx()
  ; - added x_matchpattern_n
  ; - added x_matchpattern_p and x_matchpattern_l
  ; - bugfix x_peekhex()
  ; - added several x_crypt() routines
  ; - added message support to x_log() routines
  ; - changed syntax of x_log_set()
  ; - added #HWND_TOPMOST to x_setrequesterposition()
  ; - added windownr option To window_forcetofront()
  ; - modified x_log_write() and x_log_set() to support automatic logfile trimming
  ; - added x_log_inhibit() and inhibit flag to easily support programs using logfiles for read-only media
  ; - added x_disablegadget() x_enddisablegadgets() x_gg_disableobject() x_gg_enddisableobjects()
  ; - added x_gadgetx() x_gadgety()
  ; - added x_rollspingadget()
  ; - multiple changes in x_gg()
  ; - no longer resize invisible gadgets in x_gg_draw()
  ; - bugfix x_dir() now supports empty paths
  ; - changed lots of x_gg_tag() stuff into x_gg_select() stuff
  ; - bugfix x_gg_objectsnap() x_gg_followcursor()
  ; - added multiple #x_gg constants
  ; - added x_gg_tag() etc.
  ; - added multiselection support (for tag field) to a number of x_gg procedures
  ; - bugfixes in x_matchpattern() and x_matchpatternx()
  ; - added x_matchpattern_s.s
  ;
  ; save 03244
  ;
  ; - added x_setrequesterposition()
  ; - bugfix x_gg_cloneobject()
  ; - added x_gg_copyobject() x_gg_deleteobject()
  ; - added x_gg_editorgadget() x_gg_stringgadget() x_gg_textgadget()
  ; - added x_gg_buttongadget() x_gg_buttonimagegadget() x_gg_checkboxgadget()
  ; - added x_gg_hidegroup(), x_gg_clonegroup(), x_gg_freegroupgadgets()
  ; - added x_gg_creategroupgadgets(), x_gg_editorgadgetx()
  ; - changed several x_gg() routines
  ; - added x_silentrequester()
  ; - added x_setcursor()
  ; - added x_leftmousebutton(), x_rightmousebutton(), x_mousewheel()
  ; - fixed bug in x_gg_draw() (avoid unneccessary gadget updates now works)
  ;
  ; save 02983
  ;
  ; - some bugfixes on x_gg() routines
  ; - added comments
  ; - added x_rect_over()
  ; - added x_gg_overobject()
  ; - added x_gg_overobject()
  ; - added x_gg_overwhichobject()
  ; - added x_rect_over()
  ; - added x_rect_overlap()
  ; - removed x_gg_dragsplit()
  ; - added x_gg_followcursor() and modified x_gg_calc() to match
  ; - changed x_gg_draw() check on unchanged buttons to improve redraw speed
  ;
  ; save 02648
  ;
  ; - removed x_constant() routines
  ; - changes in x_init()
  ; - changes in constants (added related function name to constant name where adequate)
  ; - removed x_countchar()
  ; - added x_stringtostring()
  ; - added x_editorgadget()
  ; - added x_gg() routines
  ;
  ; save 02396
  ;
  ; - added x_setspingadget()
  ; - added x_log_hidewindow()
  ; - changed x_log_set(), x_log_window(), x_log_waitwindow(), now supports hidden window
  ; - another bugfix in x_matchpattern() it did not handle empty strings properly
  ; - added x_icon() routines
  ; - changed x_peekhex()
  ; - added x_getcurrentpath() though x_sysinfo() does return it as well...
  ; - changed x_forcewindowtofront() to x_window_forcetofront()
  ; - added x_getlisticoncolumntitle() and x_setlisticoncolumntitle()
  ;
  ; save 02265
  ;
  ; - x_dir() now uses -1 For Default values, Not 0! change your code!
  ; - fixed bug in x_val()
  ; - x_val() now supports negative numbers (first non-space char must be a negative sign for this)
  ; - fixed some small bugs in x_window_...() routines
  ; - changed some global varnames x_retval_w / h into x_retval_width / height
  ; - some additional comments
  ; - turned a few local vars to protected equivalents
  ; - bugfix x_dir() for pb 3.93
  ; - added x_insertstring()
  ; - added mouseclick to x_waitkeywindow()
  ; - changed x_window_exteriorsize()
  ; - added x_window_interiorsize()
  ; - changed x_window_todesktopmax() x_window_toprimary()
  ; - added x_statusbarheight() x_toolbarheight() x_menuheight() x_dragbarheight()
  ; - added x_filefolder()
  ; - added x_pokehex(), x_peekhex(), x_forcewindowtofront()
  ; - added x_timer(), x_sin(), x_cos(), x_regdelete(), x_rnd()
  ; - added comment behind procedure name for easier recognition
  ; - bugfix x_matchpattern()
  ;
  ; save 01800
  ;
  ; - added x_bool() function (simply a if not zero then var is true function)
  ; - added x_log() in different tastes (see x_log_set() for more info)
  ; - more fixes in multi monitor code (expanding support for other os's)
  ; - fixes in x_exist() for win98 / ... (?)
  ; - changed x_exist() back to old version
  ; - added x_getfileinfo() to fill in stuff taken out from x_exist()
  ;
  ; save 01510
  ;
  ; - removed x_swap() which didn't make much sense anyway :-)
  ; - bugfix x_limit()
  ; - added x_window_exteriorsize() retrieves size of window exterior
  ; - removed x_boxondesktop()
  ; - changed x_window_todesktop()
  ; - changed parameters and code in x_window_desktopmax()
  ; - added x_command...() functions
  ; - minor change x_dir()
  ; - added x_constants...() functions
  ;
  ; save 01500
  ;
  ; - changed x_min() x_max() to normal purebasic, as of 3.90 psychopanta's code malfunctioned?!?
  ; - removed x_countstring() there's now the standard keyword countstring()
  ; - x_openfile() etc. can now use winapi calls (#x_ff_use = 2)
  ; - hidedebugger() replaced With closedebugger()
  ; - added x_strex() and x_str()
  ; - x_exist() now also retrieves other info on a file (filesize, when created, written to, accessed)
  ; - some small fixes to x_openfile() and related routines
  ; - x_openfile now takes parameters for sharing etc.
  ; - added x_peekb() returns a long with unsigned byte value (the & $FF thingy)
  ; - added x_pokes() pokes strings without terminating zeroes
  ; - added x_peeks() read strings up to specified terminator
  ; - added x_sysinfo() info on current Path, module, system
  ; - changed x_desktop\x and x_desktop\y to x_desktops_x etc.
  ; - improved x_monitor_detect()
  ; - added x_boxondesktop() checks if a rectangle is visible
  ; - added x_window_todesktop limits current window to desktop
  ; - added x_window_todesktopmax() spreads current window over whole desktop
  ; - added x_window_toprimarycenter() centers current window on primary screen
  ; - added x_findwindow() finds a window using wildcards
  ; - new (full) rewrite of x_matchpattern()
  ; - changed x_pbtoolbar...() etc. to x_toolbar...() etc.
  ; - added result_limit parameter to x_dir()
  ; - x_parse...() doesn't change case anymore unless #x_parse_lcase is used
  ;
  ; save 01350
  ;
  ; - buffered reading from files possible using x_fileopen() etc.
  ; - buffered writing taken out again, windows itself was faster :-)
  ; - and added writing again, but as simple wrappers for flush + pb primitives
  ; - changed naming of toolbar functions for future enhancements
  ; - few small bugfixes
  ; - removed x_maxf() x_minf() psycho's code didn't work properly :-(
  ; - small changes for 3.90 (allocatememory stuff)
  ;
  ; save 01312
  ;
  ; - removed some old vars / arrays
  ; - added toolbar functions
  ;
  ; save 01300
  ;
  ; - mostly cosmetic
  ;
  ; save 01091
  ;
  ; - added x_parse() routines (sort of findstring /stringfield variant)
  ; - added x_val()
  ; - bugfix x_matchpattern()
  ;
  ; save 01073
  ;
  ; - added multi monitor detection
  ;
  ; save 01061
  ;
  ; - bugfix x_matchpattern()
  ;
  ; save 01024
  ;
  ; - added x_dir()
  ; - added x_matchpattern()
  ;
  ; save 00320
  ;
  ; - added x_fileexist()
  ;
  ; save 00289
  ;
  ; - added registry commands
  ; - added rfindstring, rcountchar
  ;
  ; save 00001
  ;
  ; - first version in purebasic
  ;
EndProcedure

Procedure x_init()                                                   ; must be called before using any other x_lib routines
  Protected x.OSVERSIONINFO, hdc.i, logfont.LOGFONT, nonclientmetrics.NONCLIENTMETRICS, font_h.i
  ;
  ;
  ; *** initialize variables, constants and parameters, retrieve some basic information
  ;
  ; in:                                - none
  ; retval:                            - none
  ; out:      x_osversion.s            - windows version
  ;           x_osbuild.s              - windows build
  ;           x_ospack.s               - windows package (?)
  ;           x_osname.s               - windows name (windows xp, windows 7, etc.)
  ;           x_osdpi.i                - dpi of the gui
  ;           x_osscale.f              - scaling factor (current dpi compared with default 96 dpi)
  ;           x_osfontname.s           - name of system font
  ;           x_osfontsize.i           - size of system font
  ;
  ; notes:
  ;
  ; - always call this one at the start of your program
  ; - always call x_end() when done
  ;
  ;
  ; *** standard stuff (may be or is included with later versions of purebasic)
  ;
  #False = 0
  #True = 1
  ;
  ;
  ; *** unicode related
  ;
  CompilerIf #PB_Compiler_Unicode = #True
    #x_charsize = 2
  CompilerElse
    #x_charsize = 1
  CompilerEndIf
  ;
  Structure CHAR
    c.c
  EndStructure
  ;
  ;
  ; *** processor related
  ;
  CompilerIf #PB_Compiler_Processor = #PB_Processor_x64
    #x_integersize = 8
    #x_integerbitsize = 64
  CompilerElse
    #x_integersize = 4
    #x_integerbitsize = 32
  CompilerEndIf
  ;
  ;
  ; *** nr / freenr list
  ;
  
  Global x_nr_n.i
  Global x_nr.i
  x_nr_n = 1                            ; arbitrary value, first unique number given OUT
  Global NewList x_nr_list.q()          ; see x_nr() For more details
  ;
  ;
  ; *** x_parse() and x_matchpattern()
  ;
  Global x_parse.s                          ; (next) string to parse
  Global x_parsed.s                         ; last parsed section
  Global x_parse_match.s                    ; section that caused the match
  Global x_parse_search.s                   ; search term for x_parse() etc.
  Global x_parse_flags.i                    ; flags used during parsing
  Global x_parse_eol.i
  Global x_parse_field_n.i
  Global x_matchpattern_p.i
  Global x_matchpattern_l.i
  Global x_matchpattern_n.i
  Global x_matchpattern_s.s
  Global x_matchpattern_m.s
  ;
  #x_parse_notrim      = %00000             ; do not trim input and output (default)
  #x_parse_matchcase   = %00000             ; check for upper and lower case
  #x_parse_anycase     = %00001             ; don't care about case
  #x_parse_trim        = %00010             ; trim input and result
  #x_parse_eol         = %00100             ; accept eol as a match
  #x_parse_lcase       = %01000             ; turn everything to lower case
  #x_parse_chars       = %10000             ; each single char in the parse string can be used as a match
  ;
  #x_parse_limit = 1000                     ; arbitrary value for max. number of fields in x_parsefield()
  ;
  Global Dim x_parse_field.s(#x_parse_limit)
  ;
  ;
  ; *** x_dir()
  ;
  Structure x_dir_file
    name.s                                    ; fpath and / or filename and / or fldername, x_dir_file()\name replaces the old x_dir_file()
    size.i                                    ; size of file
    type.i                                    ; type of file (#PB_DirectoryEntry_Directory = folder, #PB_DirectoryEntry_File = file, -1 = unknown)
    deep.i                                    ; 0 = root
    state.i                                   ; additional data field, not used by the x_dir() routines, free for the user to (ab)use :-)
  EndStructure
  ;
  Global NewList x_dir_file.x_dir_file()    ; used to store results of x_dir() and x_dir_simple()
  Global NewList x_dir_folder.s()           ; used inside x_dir() to keep track of folders
  Global x_dir_n.i                           ; number of entries found (files and / or folders, depending on mode)
  ;
  ;
  ; *** x_blockgadget()
  ;
  Structure x_bgwindow
    windownr.l
    mousehook_h.l
  EndStructure
  ;
  Structure x_bggadget
    gadgetnr.l
    gadget_h.l
    windownr.l
  EndStructure
  ;
  Global NewList x_bgwindow.x_bgwindow()
  Global NewList x_bggadget.x_bggadget()
  ;
  ;
  ; *** x_openfile()
  ;
  #x_file_r        = $0001           ; open for reading only
  #x_file_w        = $0002           ; open for writing only
  #x_file_rw       = $0003           ; open for read and write
  #x_file_shared   = $0004           ; non-exclusive access (file is shared)
  #x_file_new      = $0008           ; create a new empty file
  #x_file_nocache  = $0010           ; decrease write caching time (data can still be cached but not lazily flushed)
  #x_file_readonly = $0100           ; create read only file (?)
  #x_file_hidden   = $0200           ; create hidden file
  #x_file_system   = $0400           ; create system file
  ;
  ;
  ; *** x_log()
  ;
  Global x_log_autocr.i, x_log_autotimestamp.i, x_log_copytodebug.i, x_log_filename.s
  Global x_log_file_h.i, x_log_indent.i, x_log_keepopen.i, x_log_isopen.i, x_log_copytowindow.i
  Global x_log_resizablewindow.i, x_log_hiddenwindow.i, x_log_inhibit.i, x_log_maxsize.i
  Global x_log_copytomailslot.i, x_log_copytoconsole.i, x_log_mailslotname.s, x_log_leader.s
  Global x_log_window_nr.i, x_log_listicon_nr.i, x_log_listicon_h.i
  ;
  ; default settings for logging
  ;
  x_log_autotimestamp.i = 0       ; 0 none 1 always 2 only when specified (using x_timestamp() before an x_log() call)
  x_log_indent.i = 0              ; insert 0 spaces in front of each log entry (indent)
  x_log_file_h.i = 0              ; see x_openfile() and x_log_set(), handle of logfile
  x_log_filename = "log.log"      ; default logfile name
  x_log_keepopen.i = #True        ; keep it open
  x_log_autocr.i = #True          ; end each log entry with a crlf
  x_log_isopen.i = #False         ; log not yet open
  x_log_copytomailslot.i = #False ; don't copy to a (remote) mailslot
  x_log_mailslotname.s = ""       ; name of the mailslot to send the message to
  x_log_leader.s = ""             ; optional leader inserted in every string send to mailslot
  x_log_copytodebug.i = #False    ; do not copy output to debug window
  x_log_copytoconsole.i = #False  ; do not copy output to console
  ;
  ;
  ; *** x_silentrequester()
  ;
  #MB_USERICON      = $80
  #QUICKNQUIET      = $C0
  #IDI_APPLICATION  = 32512
  #IDI_HAND         = 32513
  #IDI_QUESTION     = 32514
  #IDI_EXCLAMATION  = 32515
  #IDI_ASTERISK     = 32516
  #IDI_WINLOGO      = 32517
  #IDI_WARNING      = 32515
  #IDI_ERROR        = 32513
  #IDI_INFORMATION  = 32516
  ;
  ;
  ; *** winsock
  ;
  Global x_ws_library.i
  Global x_ws_wsastartup.WSADATA
  ;
  ; standard port numbers
  ;
  #x_ws_pop3 = 110                          ; pop3 port 110
  #x_ws_smtp = 25                           ; smtp port 25
  #x_ws_http = 80                           ; http port 80
  #x_ws_ftp  = 21                           ; ftp port 21
  ;
  ; winsock related structures
  ;
  Structure x_ws_fd_set
    ;
    ;   used for the select command, can be used for checking multiple sockets for some winsock commands
    ;   this is a struct with a variable length, can't do that in pure
    ;   let's stick to a max of 16... i will basically only use one anyway :-)
    ;
    ;   +0  - word        number of entries in the set
    ;   +2  - word        socketnr. #1
    ;   +4  - word        socketnr. #2
    ;   +6  - ...         etc.
    ;
    fd_number.w      ; number of entries in the set
    fd_socket.w[16]  ; one word per socket
  EndStructure
  ;
  Structure x_ws_sockaddr
    ;
    ; variant on SOCKADDR
    ;
    in_family.w
    in_port.w
    in_addr.l
    in_zero.b[8]
  EndStructure
  ;
  ; structure used for imcp echo / ping result
  ;
  Structure x_ping_reply
    reply.ICMP_ECHO_REPLY
    buffer.b[16384]
  EndStructure
  ;
  Structure x_threadping_status
    bip.l                                      ; ipv4 address in bip formaat
    status.i                                   ; status van ping
  EndStructure
  ;
  Global x_ping_timeout.i = 255                ; default timeout for x_ping() and x_threadping()
  ;
  #x_threadping_inprogress = 0
  #x_threadping_timeout = -1
  #x_threadping_error = -2
  ;
  ; *** fonts and dpi
  ;
  Global x_osfont.NONCLIENTMETRICS
  Global x_osdpi.i                             ; windows os dpi setting (default 96)
  Global x_osscale.f                           ; gui resize factor when 'xp-style-dpi' is turned on
  Global x_defaultguifontname.s
  Global x_defaultguifontsize.i
  Global x_defaultsystemfontname.s
  Global x_defaultsystemfontsize.i
  Global x_defaultmenufontname.s
  Global x_defaultmenufontsize.i
  Global x_defaultstatusfontname.s
  Global x_defaultstatusfontsize.i
  Global x_defaultmessagefontname.s
  Global x_defaultmessagefontsize.i
  Global x_defaultpurebasicfontid.i
  Global x_defaultpurebasicfontname.s
  Global x_defaultpurebasicfontsize.i
  ;
  hdc = GetDC_(0)
  x_osdpi = GetDeviceCaps_(hdc,#LOGPIXELSX)
  ReleaseDC_(0,hdc)
  x_osscale = x_osdpi/96
  ;
  ;   x_osfont\cbSize = SizeOf(x_osfont)
  ;   SystemParametersInfo_(#SPI_GETNONCLIENTMETRICS,SizeOf(x_osfont),@x_osfont,0)
  ;   x_osfontname = PeekS(@x_osfont\lfMessageFont\lfFaceName[0])
  ;   x_osfontsize = (0-PeekL(@x_osfont\lfMessageFont\lfHeight)-3)
  ;
  logfont.LOGFONT
  x_defaultpurebasicfontid = GetGadgetFont(#PB_Default)
  GetObject_(x_defaultpurebasicfontid, SizeOf(LOGFONT), @logfont)
  x_defaultpurebasicfontname.s = PeekS(@logfont\lfFaceName[0])
  x_defaultpurebasicfontsize.i = logfont\lfHeight
  ;
  logfont.LOGFONT
  font_h = GetStockObject_(#DEFAULT_GUI_FONT)
  GetObject_(font_h, SizeOf(LOGFONT), @logfont)
  x_defaultguifontname.s = PeekS(@logfont\lfFaceName[0])
  x_defaultguifontsize.i = logfont\lfHeight
  font_h = GetStockObject_(#SYSTEM_FONT)
  GetObject_(font_h, SizeOf(LOGFONT), @logfont)
  x_defaultsystemfontname.s = PeekS(@logfont\lfFaceName[0])
  x_defaultsystemfontsize.i = logfont\lfHeight
  ;
  nonclientmetrics.NONCLIENTMETRICS
  nonclientmetrics\cbSize = SizeOf(nonclientmetrics)
  SystemParametersInfo_(#SPI_GETNONCLIENTMETRICS,SizeOf(nonclientmetrics),@nonclientmetrics,0)
  x_defaultmenufontname.s = PeekS(@nonclientmetrics\lfMenuFont\lfFaceName[0])
  x_defaultmenufontsize.i = nonclientmetrics\lfMenuFont\lfHeight
  x_defaultstatusfontname.s = PeekS(@nonclientmetrics\lfStatusFont\lfFaceName[0])
  x_defaultstatusfontsize.i = nonclientmetrics\lfStatusFont\lfHeight
  x_defaultmessagefontname.s = PeekS(@nonclientmetrics\lfMessageFont\lfFaceName[0])
  x_defaultmessagefontsize.i = nonclientmetrics\lfMessageFont\lfHeight
  ;
  ; *** os
  ;
  Global x_osversion.s
  Global x_osbuild.s
  Global x_osname.s
  Global x_ospack.s
  ;
  x\dwOSVersionInfoSize = SizeOf(x)
  GetVersionEx_(@x)
  x_osversion = Str(x\dwMajorVersion)+"."+Str(x\dwMinorVersion)
  x_osbuild = Str(x\dwBuildNumber)
  x_ospack = PeekS(@x\szCSDVersion,128)
  ;
  Select OSVersion()
      Case #PB_OS_Windows_NT3_51
        x_osname = "NT3.51"
      Case #PB_OS_Windows_95
        x_osname = "95"
      Case #PB_OS_Windows_NT_4
        x_osname = "NT4"
      Case #PB_OS_Windows_98
        x_osname = "98"
      Case #PB_OS_Windows_ME
        x_osname = "ME"
      Case #PB_OS_Windows_2000
        x_osname = "2000"
      Case #PB_OS_Windows_XP
        x_osname = "XP"
      Case #PB_OS_Windows_Server_2003
        x_osname = "Server 2003"
      Case #PB_OS_Windows_Server_2008
        x_osname = "Server 2008"
      Case #PB_OS_Windows_Vista
        x_osname = "Vista"
      Default
        x_osname = "(unknown)"
  EndSelect
  ;
  ;
  ; *** other
  ;
  Global x_val_type.i
  Global x_retval.i
  Global x_peeks_read.i
  Global x_registry_type.l
  Global x_registry_q.q
  Global x_registry_s.s
  Global x_retval_x.i, x_retval_y.i
  Global x_retval_width.i, x_retval_height.i
  Global x_desktops_n.i
  Global x_desktops_x.i, x_desktops_y.i
  Global x_desktops_width.i, x_desktops_height.i
  Global x_window_h.i, x_process_id.i, x_window_name.s
  Global x_requesterposition_x.i, x_requesterposition_y.i
  Global x_requesterposition_callback.i
  Global x_requesterposition_hook.i, x_requesterposition_window.i
  Global x_x_firstinstance_instancename.s, x_firstinstance_mutex_h.i
  Global x_firstinstance_wakeupmessage.i
  Global x_retval_string.s
  Global x_osversion.s, x_osname.s, x_osbuild.s, x_ospack.s
  Global x_windowmousedx.i, x_windowmousedy.i
  Global x_diskspace_total.q
  Global x_diskspace_free.q
  Global x_diskspace_used.q
  ;
  #x_gadget_notify = $0001
  #x_gadget_wordwrap = $0002
  ;
  #CFM_BACKCOLOR = $4000000
  ;
  Global x_pi.d                       ; pi (what else?)
  x_pi.d = ACos(-1)
  ;
  Global x_lorem_ipsum.s
  x_lorem_ipsum = "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do eiusmod tempor incididunt ut "
  x_lorem_ipsum = x_lorem_ipsum + "labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris "
  x_lorem_ipsum = x_lorem_ipsum + "nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit "
  x_lorem_ipsum = x_lorem_ipsum + "esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt "
  x_lorem_ipsum = x_lorem_ipsum + "in culpa qui officia deserunt mollit anim id est laborum."
  ;
  x_totaldesktops()
  ;
EndProcedure

Procedure x_end()                                                    ; clean up api / system stuff that any x_lib procedure may have been using
  ;
  ; *** free up any loose ends used by the x_library
  ;
  ; thus far no *real* need :-) as most is handled by purebasic itself... but keep calling this routine
  ; before exiting your program, because i may change things...
  ;
EndProcedure


; *** numeric functions

Procedure.i x_nr()                                                   ; generates unique number
  ; Global x_nr.i, x_nr_n.i, x_nr_list()
  ;
  ; *** generates a new unique number
  ;
  ; in:                - none
  ; retval:  n         - unused unique number
  ; out:     x_nr      - as x_retval
  ;
  ; pure sometimes uses 'numbers' to identify elements instead of windows handles
  ; these numbers are arbitrary and fully defined by the user, this can cause trouble
  ; when using multiple libraries or code from different people as they might be reusing
  ; these unique numbers, for this purpose, i've added a x_nr procedure that returns
  ; a new unique number on every call
  ;
  ; you decide yourself if you want to free them or not :-)
  ; i would suggest doing so if you don't 'recycle' them using x_freenr() because other procedures
  ; may end up with numbers that are 'too high' for the pb functions being called
  ;
  If ListSize(x_nr_list()) > 0          ; which means there's some stuff on the list
    x_nr = x_nr_list()                  ; then use that number
    DeleteElement(x_nr_list())          ; and take it from the list
  Else
    x_nr = x_nr_n
    x_nr_n = x_nr_n+1
  EndIf
  ProcedureReturn x_nr
  ;
EndProcedure

Procedure x_freenr(n.i)                                              ; frees number generated by x_nr()
  ; Global x_nr.i, x_nr_n.i, x_nr_list()
  ;
  ; *** recycles unique numbers
  ;
  ; put this number into the 'free numbers' list so it can be reused by x_nr()
  ;
  AddElement(x_nr_list())
  x_nr_list()=n
  ;
EndProcedure

Procedure.i x_bool(var.i)                                            ; true if var is non-zero
  ;
  ; *** turns anything non-zero into a #true (ie. 1)
  ;
  ; in PureBasic the constant #True is 1, #False = 0
  ; inside an IF statement anything non-zero is considered 'true'
  ;
  If var = 0
    ProcedureReturn #False
  Else
    ProcedureReturn #True
  EndIf
EndProcedure

Procedure.i x_not(var.i)                                             ; true if var is zero
  ;
  ; *** not statement
  ;
  ; not isn't missing anymore in purebasic, but i've kept this one around for compatibility reasons
  ;
  If var = #False
    ProcedureReturn #True
  Else
    ProcedureReturn #False
  EndIf
EndProcedure

Procedure.i x_min(var1.i,var2.i)                                     ; smallest of two vars
  If var1 < var2
    ProcedureReturn var1
  Else
    ProcedureReturn var2
  EndIf
EndProcedure

Procedure.i x_max(var1.i,var2.i)                                     ; largest of two vars
  If var1 > var2
    ProcedureReturn var1
  Else
    ProcedureReturn var2
  EndIf
EndProcedure

Procedure.i x_limit(var.i,low.i,high.i)                              ; limits v.i to a range from low.i to high.i
  ;
  ; *** limit a value to extremes
  ;
  If low > high
    Swap low,high
  EndIf
  If var < low
    ProcedureReturn low
  ElseIf var > high
    ProcedureReturn high
  Else
    ProcedureReturn var
  EndIf
EndProcedure

Procedure.d x_limit_d(var.d,low.d,high.d)                            ; limits v.d to a range from low.d to high.d
  ;
  ; *** limit a value to extremes
  ;
  If low > high
    Swap low,high
  EndIf
  If var < low
    ProcedureReturn low
  ElseIf var > high
    ProcedureReturn high
  Else
    ProcedureReturn var
  EndIf
EndProcedure

Procedure.i x_near(var.d,target.d,range.d)                           ; returns true if var is within range of target
  ;
  ; *** returns true if var.d > target.d-range.d and var.d < target.d+range.d
  ;
  ; similar in concpet to x_limit_d() but here v.d is compared to a base number target.d plus/minus range range.d
  ;
  If var > target-range And var < target+range
    ProcedureReturn #True
  Else
    ProcedureReturn #False
  EndIf
EndProcedure

Procedure.i x_abs(n.i)                                               ; return absolute value as int
  ;
  ; *** return absolute value
  ;
  ;  you could use the normal abs() function though that one appears to return a float, so it may not always work
  ;
  If n < 0
    ProcedureReturn 0 - n
  Else
    ProcedureReturn n
  EndIf
EndProcedure

Procedure.i x_sgn(n.i)                                               ; returns sign (int)
  ;
  ; *** returns sign of integer var
  ;
  If n < 0
    ProcedureReturn -1
  ElseIf n = 0
    ProcedureReturn 0
  Else
    ProcedureReturn 1
  EndIf
EndProcedure

Procedure.i x_sgn_d(n.d)                                             ; returns sign (floats)
  If n < 0
    ProcedureReturn -1
  ElseIf n = 0
    ProcedureReturn 0
  Else
    ProcedureReturn 1
  EndIf
EndProcedure

Procedure.d x_sin(a.i)                                               ; sin() with angle in degrees
  ;
  ; *** as sin() but using degrees instead of radians
  ;
  ProcedureReturn Sin(a/180*x_pi)
EndProcedure

Procedure.d x_cos(a.i)                                               ; cos() with angle in degrees
  ;
  ; as cos() but using degrees instead of radians
  ;
  ProcedureReturn Cos(a/180*x_pi)
EndProcedure

Procedure.i x_random(low.i,high.i)                                   ; generates a random int in the range low to high
  ;
  ; *** return a random var from and including 'low.i' to and including 'high.i'
  ;
  ; (I find the pb random function somewhat confusing now and again...)
  ;
  ProcedureReturn low+Random(high-low)
EndProcedure

Procedure.d x_rnd()                                                  ; generates a random float between 0 and 1
  ;
  ; *** returns a regular rnd number like many other basics (from 0 to 1)
  ;
  ProcedureReturn Random(10000000)/10000000
EndProcedure

Procedure x_normalizestep(floor.i,ceiling.i,steps.i=20,skip25.i=0)   ; find closest 'human' stepsize to travers from floor to ceiling (1, 2, 2.5, 5, ...)
  Protected n.i, d.i, stepsize.i
  ;
  ; *** calculate a 'human' step size to segmentize a range from floor to ceiling
  ;
  ; in:      floor.i     - lowest number in range
  ;          ceiling.i   - highest number in range
  ;          steps.i     - number of steps preferred
  ;          skip25.i    - don't return 2.5, 25, etc.
  ; retval:  .i          - stepsize, ie. 1, 2, 5, 19, 25, 50, 100, 200, 250 etc.
  ;
  ; this procedure tries to find an acceptable step size to count from a number nearby floor to a number nearby ceiling
  ; it prefers something in the following range: 1 2 5 10 20 25 50 100 200 250 500 100 etcetera (note the missing 2.5)
  ;
  ; if there is no exact match, it tries a number higher and a number lower, then gives some preference to the lower
  ; number
  ;
  ; in other words: if a human would have a collection of 1245 objects, and would seperate those objects in groups
  ; for easier administration and handling, that human would probably chose package sizes of 1, 2, 5, 10, 20, 25, 50,
  ; 100, etc. of units per package... this routine does that, for example 12 packs of 100 units, with a remainder of 45
  ;
  If steps = < 0
    steps =1
  EndIf
  ;
  If ceiling <= floor
    ProcedureReturn 1
  Else
    stepsize = (ceiling-floor)/steps
    If stepsize = 0
      stepsize = 1
    EndIf
  EndIf
  ;
  n = 1
  d = 1
  While n < stepsize
    If n < stepsize
      d = n
      n = n*2
    EndIf
    If n = 2 Or skip25 <> 0
      If n < stepsize
        d = n
        n = n*2.5
      EndIf
    Else
      If n < stepsize
        d = n
        n = n*1.25
      EndIf
      If n < stepsize
        d = n
        n = n*2
      EndIf
    EndIf
    If n < stepsize
      d = n
      n = n*2
    EndIf
  Wend
  ;
  If Abs(stepsize-n)*1.5 < Abs(stepsize-d)
    d = n
  EndIf
  ;
  ProcedureReturn d
EndProcedure


; *** boolean and bitwise

Procedure.i x_bitmask(bit.i)                                         ; return long with specific bit set
  ;
  ; *** return a long with bit 'n' set
  ;
  ; examples:
  ;
  ; x_bitmask(2) = $04 = %00000100
  ; x_bitmask(0) = $01 = $00000001
  ;
  ProcedureReturn 1<<bit
EndProcedure

Procedure.i x_binmask(n.i)                                           ; return long with lower 'n' bits set
  Protected nn.i, i.i
  ;
  ; *** set 'n' bits of a long with 1, this quickly creates a 'mask' for other purposes
  ;
  ; examples:
  ;
  ; x_binmask(8) = $FF = %11111111 for masking bytes
  ; x_binmask(7) = $7F = %01111111 for masking lower 7 bits of a byte
  ;
  ; notes:
  ;
  ; - to create a mask with bit 7 set use x_bitmask(7) = $80 = %10000000
  ; - to create a mask with bits 0 to 7 set use x_bitmask(8) = %11111111
  ;
  ; in other words, x_bitmask sets the specific bit to 1, x_binmask sets all bits below the specific bit to 1
  ;
  i = 1
  nn = 1
  While nn < n
    i = i<<1 | 1
    nn = nn+1
  Wend
  ;
  ProcedureReturn i
EndProcedure

Procedure.l x_rol_l(x.l)                                             ; rol a long
  ;
  ; *** shift all bits one left, move msb to bit 0
  ;
  ; in ugly 3.94 purebasic:
  ;
  ; ProcedureReturn (x<<1 | (x>>31 & $01))
  ;
  ; in 32 bits asm (hey, a milestone! my very first x86 assembly :-))
  ;
  ! MOV dword EAX,[p.v_x]
  ! ROL EAX,1
  ;
  ProcedureReturn
EndProcedure

Procedure.l x_ror_l(x.l)                                             ; ror a long
  ;
  ; *** shift all bits one right, move lsb to bit 31
  ;
  ; ProcedureReturn ((x>>1 & $7FFFFFFF) | x<<31)
  ;
  ! MOV dword EAX,[p.v_x]
  ! ROR EAX,1
  ;
  ProcedureReturn
EndProcedure

Procedure.b x_rol_b(x.b)                                             ; rol a byte
  ;
  ; *** shift all bits one left, move msb to bit 0
  ;
  ; ProcedureReturn (x<<1 | (x>>7 & $01))
  ;
  ! MOV byte AL,[p.v_x]
  ! ROL AL,1
  ! MOV byte [p.v_x],AL
  ;
  ProcedureReturn x
EndProcedure

Procedure.b x_ror_b(x.b)                                             ; ror a byte
  ;
  ; *** shift all bits one right, move lsb to bit 7
  ;
  ; ProcedureReturn ((x>>1 & $7F) | x<<7)
  ;
  ! MOV byte AL,[p.v_x]
  ! ROR AL,1
  ! MOV byte [p.v_x],AL
  ;
  ProcedureReturn x
EndProcedure

Procedure.i x_roln(x.i,n.i,r.i)                                      ; rol r bits n steps
  ; Global x_retval.i
  ;
  ; *** shift the lower r bits n steps to the left, move bits shifted out to the lsb side, all non-shifted bits are cleared
  ;
  ; in:     x.l        - value to rol
  ;         n.l        - steps to shift
  ;         r.l        - size of object (in bits) to rol (8 = byte, 32 = long)
  ; retval:            - resulting value
  ;
  n = n % r
  If n > 0
    ProcedureReturn (x<<n & x_binmask(r)) | ((x & x_binmask(r) )>>(r-n))
  Else
    ProcedureReturn (x & x_binmask(r))
  EndIf
  ;
EndProcedure

Procedure.i x_rorn(x.i,n.i,r.i)                                      ; rol r bits n steps
  ;
  ; *** shift the lower r bits n steps to the right, move bits shifted out to the msb side, all non-shifted bits are cleared
  ;
  n = n % r
  If n > 0
    ProcedureReturn ((x & x_binmask(r) )>>n ) | (x<<(r-n)) & x_binmask(r)
  Else
    ProcedureReturn (x & x_binmask(r))
  EndIf
  ;
EndProcedure


; *** string and string related

Procedure.s x_str(var.i,digits.i)                                    ; convert int to string with n digits
  ;
  ; *** convert int to string with 'n' digits
  ;
  ; for example x_str(12,5) results in "00012"
  ;
  ProcedureReturn RSet(Str(var),digits,"0")
EndProcedure

Procedure.s x_string(n.i,char.s)                                     ; create string of n chars
  ;
  ; *** create string of n chars
  ;
  ; just a dumb procedure because i can't remember the equivalent in pb of string$ as used in gfa...
  ;
  ProcedureReturn LSet("",n,char)
EndProcedure

Procedure.s x_strex(var.i,format.s)                                  ; convert int to string using a format field, format characters space and -+#.#
  Protected s.s, fl.i, sl.i, *f.CHAR, *s.CHAR, p.i, sb.i
  ;
  ; *** format integer to string
  ;
  ; in:     var.i    - int var
  ;         f.s      - format
  ; retval: .s       - string
  ;
  ; convert int var to string using a pattern
  ;
  ; pattern elements:
  ;
  ; '<' - no overflow (must be first character)
  ; '#' - number or leading zero
  ; ' ' - space, number or (when there's no '+' or '-' used in the format) sign
  ; '+' - positive or negative indicator
  ; '.' - decimal sign
  ; 'x' - strip trailing digits
  ;
  ; examples:
  ;
  ; x_strex( 1234,     "###") =     "***"
  ; x_strex( 1234,   "##.##") =   "12.34"
  ; x_strex(-1234,   "##.##") =   "*****"
  ; x_strex(    1,  ".#####") =  ".00001"
  ; x_strex(   -1, "+   .##") = "-   .01"
  ;
  ; x_strex( 1234,     "###") =     "***"
  ; x_strex( 1234,      "<#") =    "1234"
  ; x_strex( 1234,     "##x") =     "123"
  ; x_strex( 1234,    "<#.#x") =   "12.3"
  ;
  ;
  s = Str(var)
  sl = Len(s)
  ;
  ; change in v7029: added support for < and x
  ;
  If Left(format,1) = "<"
    format = x_string(sl,"<")+format
  EndIf
  fl = Len(format)
  ;
  *f.CHAR = @format+fl*#x_charsize            ; using two pointers and two counters
  *s.CHAR = @s+sl*#x_charsize                 ; *f and *s are the pointers, fl and sl are the counters
  ;
  If PeekC(@format) = '+'                     ; remember if there is a sign in a fixed place
    p = 2
  ElseIf PeekC(@format) = '-'
    p = 1
  Else
    p = 0
  EndIf
  ;
  While fl > 0
    fl = fl-1
    *f = *f-#x_charsize                           ; we could be in unicode mode...
    If *f\c = '.'                                 ; skip a dot if we pass one
    Else
      sl = sl-1
      If sl >= 0
        *s = *s-#x_charsize
        sb = *s\c                                 ; sb contains the digit, AND is used as a flag
      EndIf
      Select *f\c
          Case '-'                                ; ah, a sign in a fixed place
            If sb = '-'
              sb = -1                             ; if sb = -1 we've managed to store the sign
            Else
              *f\c = ' '
            EndIf
          Case '+'                                ; same thing for the +
            If sb = '-'
              *f\c = sb
              sb = -1
            EndIf
          Case '#'                                ; if we have data that is not a sign we're gonna fill it in
            If sb = '-' Or sl < 0
              *f\c = '0'                          ; otherwise it's going to be a zero
            Else
              *f\c = *s\c
            EndIf
          Case ' ', '<'                           ; if there is no fixed spot for a sign we'll store the minus
            If sb = '-'                           ; immediately on the first space we encounter
              If p = 0
                *f\c = '-'
                sb = -1
              EndIf
            ElseIf sb <> -1 And sl >= 0           ; otherwise we'll just gonna fil it in
              *f\c = *s\c
            EndIf
      EndSelect
    EndIf
  Wend
  ;
  format = ReplaceString(format,"<","")
  format = ReplaceString(format,"x","")
  ;
  If sb = '-' Or sl > 0                       ; if the sign wasn't stored or there was some data left
    format = LSet("",Len(format),"*")         ; we'll put in some stars to indicate overflow
  EndIf
  ;
  ProcedureReturn format
EndProcedure

Procedure.i x_val(string.s)                                          ; string to val, recognizes hex dec bin oct, no floats
  Protected p.i, l.i, b.i, t.i, c.i, s.i, m.i, p0.i, p_start.i
  ; Global x_val_type.i
  ;
  ; *** as normal val() except it accepts also &H &O &0 &B % \ $ 0X
  ;
  string = UCase(Trim(string))
  l = Len(string)
  p_start = @string
  p = p_start
  s = PeekC(p) & $FF
  t = 0
  ;
  If s = '-'
    m = -1
    p = p+#x_charsize
    s = PeekC(p) & $FF
  Else
    m = 1
    If s = '+'
      p = p+#x_charsize
      s = PeekC(p) & $FF
    EndIf
  EndIf
  ;
  If s = '$'
    p = p+#x_charsize
    b = 16
  ElseIf s = '%'
    p = p+#x_charsize
    b = 2
  ElseIf s = '\'
    p = p+#x_charsize
    b = 8
  ElseIf Left(string,2) = "&B"
    p = p+2*#x_charsize
    b = 2
  ElseIf Left(string,2) = "&O"
    p = p+2*#x_charsize
    b = 8
  ElseIf Left(string,2) = "&0"
    p = p+2*#x_charsize
    b = 8
  ElseIf Left(string,2) = "&H"
    p = p+2*#x_charsize
    b = 16
  ElseIf Left(string,2) = "0X"
    p = p+2*#x_charsize
    b = 16
    ;
    ; ElseIf Left(string,1) = "0"           ; i don't like this one, as i often use
    ;    p = p+2*#x_charsize                ; preceding zeroes in front of decimals while
    ;    b = 8                              ; c(++) would turn those into octals... brrr...
    ;                                       ; well, it's up to you to uncomment these lines
  Else
    b = 10
  EndIf
  ;
  While p < p_start+l*#x_charsize
    c = (PeekB(p) & $FF) - 48
    p = p+#x_charsize
    If c > 9
      c = c - 7
    EndIf
    If c >= 0 And c < b
      t = t*b+c
    Else
      p = p+#x_charsize
    EndIf
  Wend
  t = t * m
  x_val_type = b
  ;
  ProcedureReturn t
EndProcedure

Procedure.d x_vald(string.s)                                         ; string to val, accepts commas and periods
  string = ReplaceString(string,",",".")
  ProcedureReturn ValD(string)
EndProcedure

Procedure.s x_chop(string.s,left.i,right.i)                          ; chop a bit of the left and right side of a string
  Protected l.i
  ;
  ; *** chop 'n' characters from the left or right of a string
  ;
  l = Len(string)
  If left+right >= l
    string = ""
  Else
    If right > 0
      string = Left(string,l-right)
    EndIf
    If left > 0
      string = Mid(string,left+1,l)
    EndIf
  EndIf
  ;
  ProcedureReturn string
  ;
EndProcedure

Procedure.s x_insertstring(string.s,pos.i,len.i,insert.s)            ; insert in or replace part of string.s with insert.s at position pos.i
  Protected l.i
  ;
  ; *** insert a string into another string
  ;
  ; in:     string.s              - string to work on
  ;         pos.i > 0             - pos of first char to replace / insert (for example '2' means insert between 1st and 2nd char, counts like mid())
  ;         pos.i < 0             - as pos.i = 1
  ;         pos.i > len(string)   - append at end
  ;         len.i > 0             - nr of characters to replace
  ;         len.i = 0             - insert, don't replace anything
  ;         len.i < 0             - use the length of insert.s
  ;         insert.s              - string to insert
  ; retval:                       - processed string
  ;
  l = Len(string)
  If pos > l
    string = string+insert
  ElseIf pos < 0
    string = insert+string
  ElseIf len = 0
    string = Left(string,pos-1)+insert+Right(string,l-pos+1)
  Else
    If len < 0
      len = Len(insert)
    EndIf
    string = Left(string,pos-1)+insert+Right(string,l-pos-len+1)
  EndIf
  ;
  ProcedureReturn string
  ;
EndProcedure

Procedure.i x_rfindstring(string.s,find.s)                           ; find string find.s in string.s starting from the right side
  Protected l.i , p.i
  ;
  ; *** find a string in another string, starting from the right
  ;
  ; retval:      n - found string at position n
  ;              0 - string not found
  ;
  l = Len(find.s)
  p = Len(string.s)
  If l > 0 And p > 0
    p = p + l
    Repeat
      p = p - 1
    Until p = 0 Or Mid(string,p,l) = find
  EndIf
  ;
  ProcedureReturn p
EndProcedure

Procedure.s x_propercase(s.s)                                        ; emulates the vb propercase() function
  Protected *p, f.i, b.i
  ;
  ; *** make all lowercase except for the first chars of each word
  ;
  *p = @s
  f = 1
  b = PeekC(*p)
  While b <> 0
    If b = 32
      f = 1
    ElseIf f = 1 And b >= 97 And b<=122
      PokeC(*p,b & $DF)
      f = 0
    ElseIf f = 0 And b >= 65 And b <= 90
      PokeC(*p,b | $20)
    Else
      f = 0
    EndIf
    *p = *p+#x_charsize
    b = PeekC(*p)
  Wend
  ProcedureReturn s
EndProcedure

Procedure.s x_stringtostring(string.s, mode.i)                       ; different string conversions
  Protected s.s, n.i, nn.i, nnn.i
  ;
  ; *** weird string conversion routine
  ;
  ; convert a string to another string, all sorts of conversions may end up here
  ; none of these routines are optimized for speed, use with caution
  ;
  ; in:   string.s    - string to convert
  ;       mode.l =  0 - do nothing
  ;                 1 - add spaces around pipe
  ;                 2 - replace multiple spaces with single space
  ;                 3 - remove all spaces
  ;                 4 - remove all spaces, cr's, lf's, and tabs
  ;                 5 - convert cr to space+pipe+space
  ;                 6 - convert pipe to cr
  ;                 7 - change tab in pipe, remove lf and tab, remove spaces around pipe
  ;                 8 - add spaces around pipe characters, remove all multiple spaces, and all cr's, lf's, and tabs
  ;                 9 - change space+pipe+space to pipe, strip leading, trailing and double pipe, trail with one pipe, then turn all pipe to cr
  ;                10 - remove all duplicates and empty elements from a pipe separated string
  ;                11 - replace all CRLF with CR, then all LF with CR
  ;                12 - replace any non-alphanum characters with spaces, then strip any multiple, leading and trailing spaces, turn all lower case
  ;                13 - replace (semi) colon with space, remove double spacing, turn all elements into values, remove all duplicates
  ;                14 - widen pattern, as used for pattern matching, trim spaces, surround with asterix, for example apple|pear becomes *apple*|*pear*
  ;                15 - smart widen pattern, widen pattern if it doesn't contain any wildcards
  ;                16 - change crlf to pipe
  ;                17 - change pipe to crlf
  ;
  If string > ""
    Select mode
        Case 0
          string = string
        Case 1
          string = ReplaceString(string,"|"," | ")
        Case 2
          While FindString(string,"  ",1) > 0
            string = ReplaceString(string,"  "," ")
          Wend
        Case 3
          string = ReplaceString(string," ","")
        Case 4
          string = ReplaceString(string,#CR$,"")
          string = ReplaceString(string,#LF$,"")
          string = ReplaceString(string,Chr(9),"")
          string = ReplaceString(string," ","")
        Case 5
          string = ReplaceString(string,#CR$," | ")
        Case 6
          string = ReplaceString(string,"|",#CR$)
        Case 7
          string = ReplaceString(string,#CR$,"|")
          string = ReplaceString(string,#LF$,"")
          string = ReplaceString(string,Chr(9),"")
          s = ""
          x_parse(string,"|",#x_parse_trim)
          While x_retval = #True
            s = s+x_parsed+"|"
            x_parsenext()
          Wend
          s = s+x_parsed
          string = s
        Case 8
          string = ReplaceString(string,"|"," | ")
          string = ReplaceString(string,#CR$,"")
          string = ReplaceString(string,#LF$,"")
          string = ReplaceString(string,Chr(9),"")
          string = x_stringtostring(string,2)
        Case 9
          While FindString(string," | ",1) > 0
            string = ReplaceString(string," | ","|")
          Wend
          While FindString(string,"||",1) > 0
            string = ReplaceString(string,"||","|")
          Wend
          While Left(string,1)="|"
            string = x_chop(string,1,0)
          Wend
          While Right(string,1)="|"
            string = x_chop(string,0,1)
          Wend
          string = ReplaceString(string+"|","|",#CR$)
        Case 10
          n = x_parsefield(string,"|",#x_parse_eol|#x_parse_trim)
          If n > 1
            For nn = 1 To n-1
              For nnn = nn+1 To n
                If LCase(x_parse_field(nn)) = LCase(x_parse_field(nnn))
                  x_parse_field(nnn)=""
                EndIf
              Next nnn
            Next nn
            string = ""
            For nn = 1 To n
              If x_parse_field(nn) > ""
                string = string+x_parse_field(nn)+"|"
              EndIf
            Next nn
            If Right(string,1) = "|"
              string = x_chop(string,0,1)
            EndIf
          EndIf
        Case 11
          string = ReplaceString(string,#CRLF$,#CR$)
          string = ReplaceString(string,#LF$,#CR$)
        Case 12
          string = Trim(LCase(string))
          n = Len(string)
          For nn = 1 To n
            If FindString("abcdefghijklmnopqrstuvwxyz012345678_",Mid(string,nn,1)) > 0
            Else
              string = ReplaceString(string,Mid(string,nn,1)," ")
            EndIf
          Next nn
          While CountString(string,"  ") > 0
            string = ReplaceString(string,"  "," ")
          Wend
        Case 13
          string = Trim(ReplaceString(string,","," "))
          string = Trim(ReplaceString(string,";"," "))
          n = x_parsefield(string," ",#x_parse_eol|#x_parse_trim)
          If n > 1
            For nn = 1 To n
              If Trim(x_parse_field(nn)) = ""
                x_parse_field(nn) = ""
              Else
                x_parse_field(nn) = Str(Val(x_parse_field(nn)))
              EndIf
            Next nn
            For nn = 1 To n-1
              For nnn = nn+1 To n
                If LCase(x_parse_field(nn)) = LCase(x_parse_field(nnn))
                  x_parse_field(nnn) = ""
                EndIf
              Next nnn
            Next nn
            string = ""
            For nn = 1 To n
              If x_parse_field(nn) > ""
                string = string+x_parse_field(nn)+" "
              EndIf
            Next nn
            string = Trim(string)
          EndIf
        Case 14
          string = Trim(string)
          Repeat
            s = string
            string = ReplaceString(string," |","|")
            string = ReplaceString(string,"| ","|")
            string = ReplaceString(string,"||","|")
          Until string = s
          string = "*"+string+"*"
          string = ReplaceString(string,"|","*|*")
          s = string
          Repeat
            string = ReplaceString(string,"**","*")
          Until string = s
        Case 15
          If FindString(string,"*") Or FindString(string,"?")
          Else
            string = x_stringtostring(string,14)
          EndIf
        Case 16
          string = ReplaceString(string,#CRLF$,"|")
        Case 17
          string = ReplaceString(string,"|",#CRLF$)
    EndSelect
  EndIf
  ;
  ProcedureReturn string
  ;
EndProcedure

Procedure.s x_datetime(string.s, seperator.s="-")                    ; turn any timestamp into a standard format
  Protected date.s, day.i, month.i, year.i, time.s, hour.i, minute.i, second.i, x.s
  ;
  ; *** reformat datetime stamps to dd-mm-yy hh:mm:ss
  ;
  ; in:     string.s    - datetime to convert
  ;         seperator.s - seperator character to use in output
  ; retval:             - reformatted datetime of 19 characters
  ;
  ; (exports from) sql and other dabases may return datetime stamps in various formats
  ; this procedure always turns them into DD-MM-YYYY HH:MM:SS
  ;
  day = 0
  year = 0
  month = 0
  hour = 0
  minute = 0
  second = 0
  ;
  If seperator <> "-" And seperator <> "/" And seperator <> "."
    seperator = "-"
  EndIf
  ;
  If string > ""
    ;
    ; reformat date
    ;
    ReplaceString(string,".",seperator,#PB_String_InPlace)
    ReplaceString(string,"/",seperator,#PB_String_InPlace)
    ReplaceString(string,"-",seperator,#PB_String_InPlace)
    ;
  EndIf
  ;
  ; process date
  ;
  date = StringField(string,1," ")
  If date > ""
    day = Val(StringField(date,1,seperator))
    month = Val(StringField(date,2,seperator))
    x = StringField(date,3,seperator)
    If x = ""
      year = Year(Date())
    ElseIf Len(x) = 2
      year = Val(x)
      If year >= 70
        year = 1900+year
      Else
        year = 2000+year
      EndIf
    Else
      year = Val(x)
    EndIf
    ;
    ; process time
    ;
    time = StringField(string,2," ")
    If time > ""
      hour = Val(StringField(time,1,":"))
      minute = Val(StringField(time,2,":"))
      second = Val(StringField(time,3,":"))
    EndIf
  EndIf
  ;
  If day >= 1 And day <= 31 And month >= 1 And month <= 12 And year >= 0 And year <= 9999
    x = Right("0"+Str(day),2)+seperator + Right("0"+Str(month),2)+seperator + Right("000"+Str(year),4)+" "
    x = x + Right("0"+Str(hour),2)+":" + Right("0"+Str(minute),2)+":" + Right("0"+Str(second),2)
    ProcedureReturn x
  Else
    ProcedureReturn "**"+seperator+"**"+seperator+"**** **:**:**"
  EndIf
  ;
EndProcedure

Procedure.i x_parsedate(string.s)                                    ; parse any timestamp into date format
  ;
  ; *** convert timestamp to date format
  ;
  string = x_datetime(string,"/")
  ProcedureReturn ParseDate("%dd/%mm/%yyyy %hh:%ii:%ss",string)
EndProcedure

Procedure.s x_formatdate(date.i, seperator.s="-", mark.i = #False)   ; convert date to default datetime string
  Protected string.s
  ;
  ; *** convert date to datetime string
  ;
  ; in:    date.i          - date in purebasic date format
  ;        seperator.s     - seperator used in returned sttring
  ;        mark.i = #True  - return "*" instead of 00/00/0000 etc. on invalid dates
  ;
  ; pretty much a simple replacement for formatdate with the standard format as output mask
  ;
  If date <= 0 And mark = #True
    ProcedureReturn "*"
  Else
    string = FormatDate("%dd-%mm-%yyyy %hh:%ii:%ss", date.i)
    string = ReplaceString(string,"-",seperator)
    ProcedureReturn string
  EndIf
EndProcedure

Procedure.i x_isoweeknumber(date.i)
  Protected year.i, firstday.i, isoday.i, isowk.i
  ;
  ; *** returns the iso week number
  ;
  ; notes:
  ;
  ; - week 1 is the week containing jan 4
  ; - weeks start on monday
  ;
  ; by ti99/4a
  ;
  year = Year(date)
  firstday = DayOfWeek(Date(year, 1, 1, 0, 0, 0))
  If firstDay = 0
    firstday = 7
  EndIf
  If firstday <= 4
    isoday = DayOfYear(date) + (firstDay - 1)
  Else
    isoday = DayOfYear(date) - (8 - firstDay)
  EndIf
  isowk = Round(ISOday / 7, #PB_Round_Up)
  If Not ISOwk
    isowk = x_isoweeknumber(Date(year - 1, 12, 31, 0, 0, 0))
  EndIf
  ProcedureReturn isowk
EndProcedure


; *** memory related

Procedure.i x_peekb(addr.i)                                          ; as peekb() but returns an unsigned long instead of a signed byte
  ;
  ; notes:
  ;
  ; - as peekb but returns value 0 .. 255 instead of -127 .. +127
  ; - deprecated as there is now PeekA()
  ;
  ProcedureReturn PeekA(addr)
EndProcedure

Procedure.l x_pokes(addr.i,string.s,Length.i=-1,flags.i=-1)          ; writes a string into memory without the terminating zero if the string is too long
  Protected n.i,m.i
  ; Global x_retval.i
  ;
  ; *** similar to pokes() but will not write a terminating zero if the string is too long to fit, length is specified in bytes not in chars
  ;
  ; in:      addr.l                    - address to write to
  ;          string.s                  - string to write (either unicode or regular ascii)
  ;          [ length.i = n ]          - number of BYTES, not CHARS (see note below)
  ;                       -1           - auto detect
  ;          [ flags.i = #PB_Ascii ]   - write as regular ascii (default mode if program is compiled in ascii mode)
  ;                      #PB_Unicode   - write as unicode (default mode if program is compiled in unicode)
  ;                      #PB_UTF8      - write as UTF8
  ;                      -1            - automatically choose between unicode and ascii, see above
  ; retval:  .i                        - number of bytes written
  ; out:     x_retval.i                - same as retval
  ;
  ; notes:
  ;
  ; - x_pokes() 'lenght' parameter specifies the number of BYTES, not the number of chars!
  ; - x_pokes() ONLY writes one or two terminating zeroes if there is enough space!
  ; - purebasic's pokes() 'length' parameter specifies the number of CHARS, not the number of bytes!
  ; - purebasic's pokes() ALWAYS writes a terminating zero into memory (2 if writing unicode)!
  ; - no support for UTF8 in memory (use ASCII)
  ;
  ;
  If flags = -1
    CompilerIf #PB_Compiler_Unicode
      flags = #PB_Unicode
    CompilerElse
      flags = #PB_Ascii
    CompilerEndIf
  EndIf
  ;
  x_retval = StringByteLength(string,flags)
  If flags = #PB_Unicode
    x_retval = x_retval+2
  Else
    x_retval = x_retval+1
  EndIf
  ;
  m = AllocateMemory(x_retval)
  If flags = -1
    PokeS(m,string)
  Else
    PokeS(m,string,-1,flags)
  EndIf
  ;
  If length > -1 And length < x_retval
    x_retval = length
  EndIf
  CopyMemory(m,addr,x_retval)
  FreeMemory(m)
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.s x_peeks(addr.i,length.i=-1,flags.i=-1,terminator.s="")   ; read string from mem until a null is found or max length is reached
  Protected string.s, p.l
  ; Global x_retval.i, x_peeks_read.i
  ;
  ; *** read a string from memory until terminating condition is met
  ;
  ; in:     addr.i             - location in memory
  ;         [ length.i = n ]   - max length in BYTES
  ;                    = -1    - default: ignore
  ;         [ flags.i  = n ]   - #PB_Ascii, #PB_Unicode, #PB_UTF8
  ;                    = -1    - default: use #PB_Ascii if program is compiled in ascii mode, #PB_Unicode if compiled in unicode mode
  ;         terminator.s       - seperator
  ; retval: .s                 - string found
  ; out:    x_retval.i         - length of string as found in memory in BYTES
  ;         x_peeks_read.i     - as x_retval.i
  ;
  ; notes:
  ;
  ; - terminating condition can be string, a null character, or hitting maximal length
  ; - purebasic's peeks() uses chars not bytes! (in 4.02 the included helpfile is wrong)
  ; - x_peeks() uses bytes not chars!
  ; - a terminating zero in unicode mode is actually TWO zeroes, ie. $ 00 00!
  ; - no support for UTF8 in memory, use ASCII
  ;
  If flags = -1
    CompilerIf #PB_Compiler_Unicode
      flags = #PB_Unicode
    CompilerElse
      flags = #PB_Ascii
    CompilerEndIf
  EndIf
  ;
  If length > 0
    If flags = #PB_Unicode
      string = PeekS(addr,length/2,flags)
    Else
      string = PeekS(addr,length,flags)
    EndIf
  Else
    string = PeekS(addr,-1,flags)
  EndIf
  ;
  x_retval = StringByteLength(string,flags)
  If x_retval < length Or length = -1
    If flags = #PB_Unicode
      x_retval = x_retval+2
    Else
      x_retval = x_retval+1
    EndIf
  EndIf
  ;
  If terminator > ""
    p = FindString(string,terminator,1)
    If p > 0
      string = Left(string,p-1)
      x_retval = StringByteLength(string+terminator,flags)
    EndIf
  EndIf
  ;
  x_peeks_read = x_retval
  ProcedureReturn string
EndProcedure

Procedure.i x_pokehex(addr.i,s.s)                                    ; write a hex sequence to memory
  Protected l.i, n.i, p.i, x.s, v.i
  ;
  l = Len(s)
  p = 0
  ;
  While n+1 < l
    n = n+1
    x = Mid(s,n,1)
    If ( x >= "0" And x <= "9" ) Or ( x >= "A" And x <= "F" ) Or ( x >= "a" And x <= "f" )
      n = n+1
      x = Mid(s,n,1)
      If ( x >= "0" And x <= "9" ) Or ( x >= "A" And x <= "F" ) Or ( x >= "a" And x <= "f" )
        v = Val("$"+Mid(s,n-1,2))
        PokeB(addr+p,v)
        p = p+1
      EndIf
    EndIf
  Wend
  ;
  ProcedureReturn p
EndProcedure

Procedure.s x_peekhex(*addr.byte,length.i,mode.i=7)                  ; read a sequence of bytes from mem and store them in hex format in a string
  Protected s.s, b.l, n.l
  ;
  ; *** read a sequence of bytes from memory and store them in hex format in a string
  ;
  ; in:      *addr.byte             - address where to peek
  ;          length.i               - number of bytes
  ;          [ mode.i = 0 ]         - just a hex string
  ;                     1           - add spaces between the bytes
  ;                     2           - add string sign in front
  ;                     3           - add both
  ;                     7 (default) - add string, add spaces, and show the char itself
  ; retval:  .s                     - string containing hex presentation of peeked data
  ;
  ;
  If (mode & 2) <> 0
    s = "$"
  EndIf
  ;
  n = 0
  While n < Length
    b = *addr\b
    If (mode & 1) <> 0
      s = s+" "
    EndIf
    s = s+Right("0"+Hex(b),2)
    n = n+1
    *addr = *addr+1
    If (mode & 4) <> 0
      If b > 31 And b < 'z'
        s = s+" [ "+Chr(b)+" ]"
      EndIf
    EndIf
  Wend
  ProcedureReturn Trim(s)
EndProcedure

Procedure.i x_stringbytelength(s.s,format.i)                         ; returns memory space needed for storing a string including terminating zeroe(s)
  ; Global x_retval.i
  ;
  ; *** returns memory space needed for storing a string including the terminating zero / zeroes
  ;
  ; in:     s.s                        - string
  ;         format.i = #PB_UTF8        - how much space is needed for an UTF8 string zero terminated
  ;                    #PB_Unicode     - same in widechar mode
  ;                    #PB_Ascii       - same in flat ascii
  ; retval: .i                         - memory space needed
  ;
  ; notes:
  ;
  ; - PB stringbytelength does not take the terminating zero or zeroes into account
  ;
  x_retval = StringByteLength(s,format)
  Select format
      Case #PB_Unicode
        x_retval = x_retval+2
      Case #PB_Ascii, #PB_UTF8
        x_retval = x_retval+1
      Default
        ;
        ; not yet implemented
        ;
  EndSelect
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure x_sizeof(type.i)                                           ; returns the size of the given variable type
  Protected a.a, b.b, c.c, d.d, i.i, f.f, l.l, u.u, w.w, q.q, size.i
  ;
  ; *** return the size of the given variable type
  ;
  ; in:    type     > 0            - use one of the predefined purebasic types (#pb_byte, #pb_word etc.)
  ;                 = 0            - not in use, length 0
  ;                 < 0            - specify length (note, to specify a field length of 6 bytes, specify it as -6 etc.)
  ; retval:         n              - size in bytes
  ;                 999999         - unknown type
  ;
  ; example: x_sizeof(#pb_byte)
  ;          x_sizeof(-6)
  ;
  ; typically use sizeof() instead of this procedure, only required for special circumstances
  ;
  If type > 0
    Select type.i
        Case 0
          size = 0
        Case #PB_Byte
          size = SizeOf(b)
        Case #PB_Word
          size = SizeOf(w)
        Case #PB_Long
          size = SizeOf(l)
        Case #PB_Float
          size = SizeOf(f)
        Case #PB_Character
          size = SizeOf(c)
        Case #PB_Double
          size = SizeOf(d)
        Case #PB_Quad
          size = SizeOf(q)
        Case #PB_Integer
          size = SizeOf(i)
        Case #PB_Ascii
          size = SizeOf(a)
        Case #PB_Unicode
          size = SizeOf(u)
        Default
          size = 999999
    EndSelect
  Else
    size = type
  EndIf
  ProcedureReturn size
EndProcedure


; *** patterns / parsing / fields

Procedure.i x_parsenext()                                            ; use after x_parse(), look for the next occurance of search.s
  Protected p.i, n.i, c.s, pp.i
  ; Global x_parse_flags.i, x_parsed.s, x_parse.s, x_parse_eol.i, x_retval.i, x_parse_search.s
  ;
  ; *** look for the next occurance of the searchstring x_parse_search.s as last used in x_parse() or x_parsestring()
  ;
  If x_parse_flags & #x_parse_trim <> 0                               ; trim
    x_parse = Trim(x_parse)
  EndIf
  ;
  If x_parse_flags & #x_parse_lcase <> 0
    x_parse = LCase(x_parse)
    x_parse_search = LCase(x_parse_search)
  EndIf
  ;
  If Len(x_parse) = 0                                                 ; empty string
    x_retval = #False
    x_parse_eol = #True
    x_parsed = ""
    x_parse = ""
    x_parse_match = ""
  Else
    ;
    If x_parse_flags & #x_parse_chars = 0                             ; regular search for specified string
      If x_parse_flags & #x_parse_anycase <> 0                        ; accept both upper and lower case
        p = FindString(LCase(x_parse.s),LCase(x_parse_search.s),1)
      Else                                                            ; match case
        p = FindString(x_parse.s,x_parse_search.s,1)
      EndIf
    Else                                                              ; matching any of the characters in the search string is considered a match
      p = 0
      n = 0
      Repeat                                                          ; check for all occurences
        n = n+1
        c = Mid(x_parse_search,n,1)
        If x_parse_flags & #x_parse_anycase <> 0                      ; accept both upper and lower case
          pp = FindString(LCase(x_parse.s),LCase(c.s),1)
        Else                                                          ; match case
          pp = FindString(x_parse.s,c.s,1)
        EndIf
        If pp > 0 And ( pp < p Or p = 0)                              ; found a valid nearby match
          p = pp
          x_parse_match = c                                           ; character that triggered the match
        EndIf
      Until n = Len(x_parse_search)
    EndIf
    ;
    If p = 0                                                          ; didn't find the search string
      x_parsed = x_parse
      x_parse = ""
      x_parse_match = ""
      x_parse_eol = #True
      If x_parse_flags & #x_parse_eol <> 0                            ; eol is considered a match
        x_retval = #True
      Else
        x_retval = #False
      EndIf
    Else                                                              ; x_retval
      x_parsed = Left(x_parse,p-1)
      If x_parse_flags & #x_parse_chars = 0                           ; parsing for strings of more than one char
        x_parse_match = Mid(x_parse,p,Len(x_parse_search))
        x_parse = Mid(x_parse,p+Len(x_parse_search),Len(x_parse))
      Else                                                            ; parsing for single characters
        x_parse = Mid(x_parse,p+1,Len(x_parse))
      EndIf
      x_retval = #True
      ;
      If Len(x_parse) = 0                                             ; the match just came before the eol so there's nothing left
        x_parse_eol = #True
      Else
        x_parse_eol = #False
      EndIf
    EndIf
    ;
    If x_parse_flags & #x_parse_trim <> 0
      x_parse = Trim(x_parse)
      x_parsed = Trim(x_parsed)
      x_parse_match = Trim(x_parse_match)
    EndIf
    ;
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_parsestring(search.s)                                  ; use after x_parse(), look for the next occurance of search.s
  ; Global x_parse_search.s
  ;
  ; *** look for the next occurance of search.s (overwrites the previous one set in x_parse())
  ;
  x_parse_search.s = search
  ProcedureReturn x_parsenext()
EndProcedure

Procedure.i x_parse(string.s,search.s,flags.i=7)                     ; find search.s in string.s, processed is in x_parsed.s, remaining in x_parse.s
  ; Global x_parse_flags.i, x_parsed.s, x_parse.s, x_parse_search.s, x_retval.i
  ;
  ; *** parse a string (retrieve parameters etc.) a variation on findstring
  ;
  ; in:     string.s                               - string to parse
  ;         search.s                               - string or seperator to find
  ;         [ flags.i = 7 ]                        - no case check, trim, use eol
  ;                   = 0                          - slice only
  ;                   = #x_parse_anycase %00001  1 - no case check
  ;                   = #x_parse_trim    %00010  2 - trim leading / trailing spaces
  ;                   = #x_parse_eol     %00100  4 - hitting eol is considered a match
  ;                   = #x_parse_lcase   %01000  8 - force output to lowercase
  ;                   = #x_parse_chars   %10000 16 - each single char in the search string can be used for a match
  ;
  ; retval: 0 or #False                            - not found
  ;         1 or #True                             - found
  ; out:    x_parsed.s                             - part of string BEFORE the x_retval searchstring
  ;         x_parse.s                              - part of string AFTER the x_retval searchstring (can be reparsed)
  ;         x_parse_match.s                        - part of string BETWEEN x_parsed and x_parse
  ;         x_parse_eol.i = 1 (#True)              - flag to indicate end of line was reached
  ;         x_retval.i = 1 (#true)                 - as retval
  ;
  x_parse_flags = flags
  x_parse = string
  x_parse_flags = flags
  x_parse_search = search
  x_parse_match = ""
  x_parsed = ""
  ProcedureReturn x_parsenext()
EndProcedure

Procedure.i x_parsefield(string.s,search.s,flags.i=7)                ; fills the x_parse_field() array with segments separated by search.s
  Protected p1.s, p2.s
  ; Global x_parse_flags.i, x_parsed.s, x_parse.s, x_parse_search.s, x_parse_field_n.i, x_retval.i
  ; Global Dim x_parse_field.s()
  ;
  ; *** parse a string repeatedly and return an array with results
  ;
  ; in:          ...               - see x_parse()
  ; retval:      n                 - number of segments found
  ; out:         x_retval.i        - number of segments found
  ;              x_parse_field_n.i - number of segments found
  ;              x_parse_field.s() - results (starting at 1)
  ;              x_parsed          - first field found (equal to x_parse_field(1))
  ;              x_parse           - original string (after eventual trimming)
  ;
  ; see also:    x_parse()         - x_lib function
  ;              x_matchpattern()  - x_lib function
  ;              stringfield()     - purebasic function (why did i ever write this function? perhaps for an older version of purebasic?)
  ;
  ; notes:
  ;
  ; - can't recall why i wrote this, wouldn't purebasic's stringfield not work just as well? (well, yeah, but there's case, multiple chars, etc....)
  ; - x_parsefield() fills x_parse.s and x_parsed.s the same way as x_parse() so x_parsenext() etc. can be used
  ; - limited to a 1000 fields, set with constant #x_parse_limit in x_init()
  ; - the array starts with element 1 (should turn this into a linked list someday)
  ;
  Dim x_parse_field.s(#x_parse_limit)
  x_parse_field_n = 0
  ;
  x_parse = string
  x_parsed = ""
  x_parse_search = search
  x_parse_flags = flags
  ;
  x_parsenext()
  p1 = x_parsed
  p2 = x_parse
  While x_retval = #True And x_parse_field_n < #x_parse_limit
    x_parse_field_n = x_parse_field_n+1
    x_parse_field(x_parse_field_n) = x_parsed
    x_parsenext()
  Wend
  x_parsed = p1
  x_parse = p2
  ;
  x_retval = x_parse_field_n
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_matchpattern(string.s,pattern.s,ulcase.i=1)                                    ; check if a string matches one or more patterns  (wildcards *? separator |)
  Protected p_p.i, p_l.i, s_l.i, match.i, s_p.i, ff_max.i, ff_min.i, p_sub.s, p_char.s, done.i, x.i
  ; Global x_retval.i, x_matchpattern_p.i, x_matchpattern_l.i, x_matchpattern_n.i, x_matchpattern_s.s, x_matchpattern_m.s
  ;
  ; *** check if string matches one or more patterns using * and ? as wildcards and | as seperator
  ;
  ; (this is probably something I should rewrite one day to make use of the regular expression library of pb4.xx)
  ;
  ; in:     string.s                             - string to check
  ;         pattern.s                            - pattern including wildcards * and ? and multiple patterns seperated by |
  ;         ulcase.i    = 0 (#x_parse_matchcase) - check case
  ;                     = 1 (#x_parse_anycase)   - don't care about case (default)
  ;
  ; retval: 0 or #False                          - no match
  ;         1 or #True                           - match
  ;
  ; out:    x_matchpattern_p.i                   - first character of the match in string.s
  ;         x_matchpattern_l.i                   - length of the match in string.s
  ;         x_matchpattern_s.s                   - part of the string that matched the pattern
  ;         x_matchpattern_n.i                   - number of the pattern that caused the match (starting at 1)
  ;         x_matchpattern_m.s                   - pattern that caused the match
  ;
  ; notes:
  ;
  ; - x_matchpattern_p and l EXCLUDE bordering wildcards, ie. 'abcdef' contains pattern '*cd*' at position 3 with lenght 2!
  ; - ? must be filled by a single character, this is different in windows / dos where it can be empty
  ; - note that an 'empty' pattern will match an empty original
  ; - rewrite number three, and it's still horrible :-) and i'm only making it worse :-)
  ; - x_matchpattern_p points to the first character in string that matches the first non-wildcard character in pattern
  ; - x_matchpattern_l gives the length of the part of string that matches pattern excluding surrounding wildcards
  ;
  ; example:
  ;
  ;   string.s = "appel"
  ;   x_matchpattern(string,"?ppe*",1)
  ;   Debug Mid(string,x_matchpattern_p,x_matchpattern_l)   ; this returns 'ppe' in spite of the wildcards
  ;
  If ulcase = 1
    pattern = LCase(pattern)
    string = LCase(string)
  EndIf
  ;
  x_matchpattern_p = -1
  x_matchpattern_l = -1
  x_matchpattern_n = 1
  ;
  p_p = 0
  p_l = Len(pattern)
  s_l = Len(string)
  ;
  If s_l = 0                                               ; 2265 handle emtpy strings
    If p_l = 0
      match = #True
    ElseIf FindString(pattern,"||",1) > 0
      match = #True
    ElseIf Right(pattern,1) = "|"
      match = #True
    Else
      match = #False
    EndIf
  Else
    Repeat
      x_matchpattern_p = -1                                ; 5126 bugfix to make sure x_matchpattern_p isn't set on a partial match
      s_p = 1
      ;
      Repeat
        ff_max = 0
        ff_min = 0
        p_sub.s = ""
        p_char.s = ""
        done = #False
        match = #False
        Repeat
          p_p = p_p+1
          p_char = Mid(pattern,p_p,1)
          If p_char = "*"
            ff_max = s_l
          ElseIf p_char = "?"
            ff_min = ff_min+1
            ff_max = x_max(ff_max,ff_min)
          ElseIf p_char = "|"
            done = #True
          Else
            p_sub = p_sub+p_char
            Repeat
              p_char = Mid(pattern,p_p+1,1)
              If p_char = "*" Or p_char = "?"
                done = #True
              ElseIf p_char = "|"
                done = #True
                ; p_p = p_p+1                              ; 2265 took this one out as a quickfix... appears to work
              Else
                p_sub = p_sub+p_char
                If p_p < p_l
                  p_p = p_p+1
                Else
                  done = #True
                EndIf
              EndIf
            Until done = #True
          EndIf
          If p_p >= p_l
            done = #True
          EndIf
        Until done = #True
        ;
        If p_sub = ""
          If s_l-s_p+1 <= ff_max And s_l-s_p+1 >= ff_min
            match = #True
          Else
            x_matchpattern_n = x_matchpattern_n+1
          EndIf
          s_p = s_l+1
        Else
          x = FindString(string,p_sub,s_p)
          If x = 0                                         ; 4150 added this little twist to find the matching string (a hack, yes, as usual)
            x_matchpattern_p = -1
          Else
            If x_matchpattern_p = -1
              x_matchpattern_p = x
            EndIf
            x_matchpattern_l = (x+Len(p_sub))-x_matchpattern_p
          EndIf
          If x > 0
            If x-s_p > ff_max
              match = #False
            ElseIf x-s_p < ff_min
            Else
              match = #True
            EndIf
            s_p = x+Len(p_sub)
          Else
            s_p = s_l+1
          EndIf
        EndIf
      Until match = #False Or s_p = s_l+1
      ;
      If p_p >= p_l
      Else
        If match = #False
          While p_p < p_l And p_char <> "|"
            p_p = p_p+1
            p_char = Mid(pattern,p_p,1)
          Wend
        Else
          While p_p < p_l And p_char <> "|"
            p_p = p_p+1
            p_char = Mid(pattern,p_p,1)
            If p_char = "|"
            ElseIf p_char <> "*"
              match = #False
            EndIf
          Wend
        EndIf
      EndIf
    Until match = #True Or p_p >= p_l
  EndIf
  ;
  If match = #False
    x_matchpattern_s = ""
    x_matchpattern_m = ""
    x_matchpattern_l = 0
    x_matchpattern_p = -1
    x_matchpattern_n = -1
  Else
    x_matchpattern_s = Mid(string,x_matchpattern_p,x_matchpattern_l)
    x_matchpattern_m = StringField(pattern,x_matchpattern_n,"|")
  EndIf
  ;
  x_retval = match
  ProcedureReturn match
EndProcedure

Procedure.i x_matchpatternx(string.s,pattern.s,ulcase.i=1,widen.i=1,filter.s="")             ; modify pattern before searching, filter results
  Protected match.i, n.i, x.s, l.i,old.s
  ; Global x_retval.i, x_matchpattern_n.i, x_matchpattern_p.i, x_matchpattern_l.i, x_matchpattern_s.s
  ;
  ; *** as regular matchpattern but modifies pattern before searching, and filters out results with specific characters in the matched section
  ;
  ; in:     string.s                             - string to check
  ;         pattern.s                            - pattern including wildcards * and ? and multiple patterns seperated by |
  ;         ulcase.i    = 0 (#x_parse_matchcase) - check case
  ;                     = 1 (#x_parse_anycase)   - don't care about case
  ;         widen.i  = 1                         - strip trailing / leading spaces and surround by *
  ;         filter.s                             - result doesn't match pattern if characters listed in filter are found in the match
  ;
  ; retval: 0 or #False                          - no match
  ;         1 or #True                           - match
  ;
  ; see x_matchpattern() for more additional information
  ;
  ; typically this can help turning searches in a more 'human readable' form, for example searches like
  ;
  ;   "appel | peer"
  ;
  ; with widen = 1 will change those patterns into
  ;
  ;   "*appel*|*peer*"
  ;
  ; however, things are a little smarter (more confusing? :-)) than that:
  ;
  ; 1. the function looks for the listed pattern
  ; 2. if it finds the pattern, it then checks if the result contains any of the characters listed in the filter
  ; 3. if that filter character also exists in the pattern it's considered a match
  ; 4. if that filter character does exist in the found string but did not exist in the pattern it's a miss
  ;
  ; for example, the next code will NOT return a match, as the '+' sign in filter filters out all matches containing it:
  ;
  ;   ; no match
  ;   string.s  = "+appel"
  ;   pattern.s = "*"
  ;   filter.s  = "+"
  ;
  ; however if the pattern also contains that character ANYWHERE then it WILL return a match (doesn't make sense? good! :-))
  ;
  ;   ; match
  ;   string.s  = "+appel"
  ;   pattern.s = "+*"
  ;   filter.s  = "+"
  ;
  ; and with anywhere I mean anywhere, it doesn't have to be the part that caused the match!
  ; (it's a feature I needed for simple searches for single patterns where the patterns could include wildcard characters ? and *)
  ;
  ;   ; match (!) because SOMEWHERE in the list of patterns the filtered character is listed!
  ;   string.s  = "appeltaart"
  ;   pattern.s = "a*taart|p"
  ;   filter.s  = "p"
  ;
  ;
  ; step 1: modify the pattern before actually doing the test
  ;
  If widen > 0
    pattern = x_stringtostring(pattern,14)
  EndIf
  ;
  ; step 2: match and filter
  ;
  match = #False
  If x_matchpattern(string,pattern,ulcase)
    match = #True
    n = 0
    While n < Len(filter)
      n = n+1
      x.s = Mid(filter,n,1)
      If CountString(x_matchpattern_s,x) > 0 And CountString(pattern,x) <= 0  ; a filter character exists in the string but not in the pattern
        match = #False
      EndIf
    Wend
  EndIf
  ;
  x_retval = match
  ProcedureReturn x_retval
EndProcedure

Procedure.s x_commandfield(string.s,index.i,stripdquotes.i=#True)                            ; segmentizes a complete command line recognizing double quotes
  Protected l.i, field.i, inside.i, p_start.i, p.i
  ; Global x_retval_string.s
  ;
  ; *** returns space seperated fields, keeps double quoted sections whole
  ;
  ; in:       string.s                - string to parse
  ;           index.i                 - field to return
  ;           stripquotes.i = #true   - include encompassing quotes in return value (default #false)
  ; retval:   .s                      - string matching the index number
  ;
  ; see also x_commandcount()
  ;
  ; a 'command line string' contains multiple sections, of which some are enclosed in parenthesis
  ; this routine splits up and spits out :-) space seperated sections, where spaces in doublequoted
  ; parts are ignored
  ;
  ; duh... what this means is that you can, from inside the code, quickly test command line like structures,
  ; like this:
  ;
  ;   debug x_commandfield(3,"server:abc.com port:4010 user:test password:test connect") ; returns user:test
  ;
  ; you can use #DQUOTE$ to include doublequote characters, but you could cheat a little by doing this:
  ;
  ;   s.s = replacestring("'this' 'is a test'")
  ;   debug x_commandfield(2,s) ; would return 'is a test' without the quote characters
  ;
  ; so yes, you cannot pass SINGLE quotes as parameters... but who cares except the french ;-)
  ;
  l = Len(string)
  field = 0
  inside = #False
  p_start = 1
  p = 0
  While p < l And field <> index
    p = p+1
    Select Asc(Mid(string,p,1))
        Case ' '                        ; it's a space
          If Not inside                 ; but we're not inside so it counts as a seperator
            If p = p_start              ; we just started so it must be two spaces following eachother
              p_start = p+1             ; next char should be a non-space, otherwise we end up here again
            Else
              field = field+1           ; so we just passed a valid field
              If field <> index         ; if it wasn't the field we were looking for
                p_start = p+1           ; then we'll set the start pointer to the start of the next field
              EndIf
            EndIf
          EndIf
        Case '"'                        ; toggle for doublequoted sections
          inside = 1-inside
    EndSelect
  Wend
  If field <> index And p = l       ; reached the end of the string
    field = field+1
    p = p+1
  EndIf
  If field = index                                                   ; we had a match
    If stripdquotes = #True And Mid(string,p_start,1)=#DQUOTE$       ; strip doublequotes
      x_retval_string = Trim(Mid(string,p_start+1,p-p_start-2))
    Else
      x_retval_string = Mid(string,p_start,p-p_start)                ; include doublequotes
    EndIf
  Else
    x_retval_string = ""
  EndIf
  ;
  ProcedureReturn x_retval_string
EndProcedure

Procedure.i x_commandcount(string.s)                                                         ; returns the number of fields in a commandline
  Protected l.i, field.i, inside.i, p_start.i, p.i
  ; Global x_retval.i
  ;
  ; *** returns the number of space seperated fields, keeps double quoted sections whole
  ;
  ; in:       string.s                - string to parse
  ; retval:   .s                      - string matching the index number
  ;
  ; see x_commandfield()
  ;
  ; a 'command line string' contains multiple sections, of which some are enclosed in parenthesis
  ; this routine counts how many are valid, which can be used with commandfield()
  ;
  l = Len(string)
  field = 0
  inside = #False
  p_start = 1
  p = 0
  While p < l
    p = p+1
    Select Asc(Mid(string,p,1))
        Case ' '                        ; it's a space
          If Not inside                 ; but we're not inside so it counts as a seperator
            If p = p_start              ; we just started so it must be two spaces following eachother
              p_start = p+1             ; next char should be a non-space, otherwise we end up here again
            Else
              field = field+1           ; so we just passed a valid field
              p_start = p+1             ; then we'll set the start pointer to the start of the next field
            EndIf
          EndIf
        Case '"'                        ; toggle for doublequoted sections
          inside = 1-inside
    EndSelect
  Wend
  If p+1 <> p_start
    field = field+1
  EndIf
  ;
  x_retval = field
  ProcedureReturn field
EndProcedure


; *** files

Procedure.i x_exopenfile(nr.i,name.s,flags.i=0,timeout.i=2000,retries.i=5)                   ; retrying alternative for openfile()
  Protected retval.i, t.i
  ;
  ; *** retrying alternative for openfile()
  ;
  ; in:        nr.i                      - file nr for source, use nr or #pb_any
  ;            name.s                    - file name
  ;            flags.i                   - openfile() flags
  ;            timeout.i                 - timeout in msec, default 2000
  ;            retries.i                 - max number of retries within timeout period, default 5, each retry takes 2000/5=400 msec by default
  ; retval:    0                         - fail
  ;            <> 0                      - success
  ;
  ; ran into some problems with the copyfile() command, suspect winapi or virus scanner issues, this retries file actions before giving up
  ;
  ; this is a replacment for:
  ;
  ;   openfile(nr.i,name.s,flags.i)
  ;
  t = GetTickCount_()+timeout
  retval = OpenFile(nr,name,flags)
  While retval = 0 And GetTickCount_() < t
    Delay(timeout/retries)
    retval = OpenFile(nr,name,flags)
  Wend
  ;
  ProcedureReturn retval
  ;
EndProcedure

Procedure.i x_exreadfile(nr.i,name.s,flags.i=0,timeout.i=2000,retries.i=5)                   ; retrying alternative for readfile()
  Protected retval.i, t.i
  ;
  ; *** retrying alternative for readfile()
  ;
  ; in:        nr.i                      - file nr for source, use nr or #pb_any
  ;            name.s                    - file name
  ;            flags.i                   - openfile() flags
  ;            timeout.i                 - timeout in msec, default 2000
  ;            retries.i                 - max number of retries within timeout period, default 5, each retry takes 2000/5=400 msec by default
  ; retval:    0                         - fail
  ;            <> 0                      - success
  ;
  t = GetTickCount_()+timeout
  retval = ReadFile(nr,name,flags)
  While retval = 0 And GetTickCount_() < t
    Delay(timeout/retries)
    retval = ReadFile(nr,name,flags)
  Wend
  ;
  ProcedureReturn retval
  ;
EndProcedure

Procedure.i x_excreatefile(nr.i,name.s,flags.i=0,timeout.i=2000,retries.i=5)                 ; retrying alternative for createfile()
  Protected retval.i, t.i
  ;
  ; *** retrying alternative for createfile()
  ;
  ; in:        nr.i                      - file nr for source, use nr or #pb_any
  ;            name.s                    - file name
  ;            flags.i                   - openfile() flags
  ;            timeout.i                 - timeout in msec, default 2000
  ;            retries.i                 - max number of retries within timeout period, default 5, each retry takes 2000/5=400 msec by default
  ; retval:    0                         - fail
  ;            <> 0                      - success
  ;
  t = GetTickCount_()+timeout
  retval = CreateFile(nr,name,flags)
  While retval = 0 And GetTickCount_() < t
    Delay(timeout/retries)
    retval = CreateFile(nr,name,flags)
  Wend
  ;
  ProcedureReturn retval
  ;
EndProcedure

Procedure.i x_exdeletefile(name.s,timeout.i=2000,retries.i=5)                                ; retrying alternative for deletefile()
  Protected retval.i, t.i
  ;
  ; *** retrying alternative for deletefile()
  ;
  ; in:        name.s                    - file name
  ;            timeout.i                 - timeout in msec, default 2000
  ;            retries.i                 - max number of retries within timeout period, default 5, each retry takes 2000/5=400 msec by default
  ; retval:    0                         - fail
  ;            <> 0                      - success
  ;
  t = GetTickCount_()+timeout
  If FileSize(name) = -2                                                  ; file is a folder, cannot delete, so fail
    retval = 0
  ElseIf FileSize(name) = -1                                              ; file was already deleted
    retval = 1
  Else
    DeleteFile(name)
    retval = FileSize(name)         
    While retval <> -1 And GetTickCount_() < t
      Delay(timeout/retries)
      DeleteFile(name)
      retval = FileSize(name)       
    Wend
    If retval = -1                                                        ; a filesize of -1 means the file has been deleted
      retval = 1                                                          ; so report success
    Else
      retval = 0                                                          ;  otherwise report a fail
    EndIf
  EndIf
  ;
  ProcedureReturn retval
  ;
EndProcedure

Procedure.i x_excopyfile(s_nr.i,s_name.s,d_nr.i,d_name.s,timeout.i=2000,retries.i=5)         ; retrying alternative for copyfile()
  Protected t.i, s_handle.i, d_handle.i, retval.i
  ;
  ; *** retrying alternative for copyfile()
  ;
  ; in:        source_nr.i               - file nr or #pb_any for source
  ;            source_name.s             - file to copy
  ;            destination_nr.i          - file nr or #pb_any for destination
  ;            destination_name.s        - file to create / overwrite
  ;            timeout.i                 - timeout in msec, default 2000
  ;            retries.i                 - max number of retries within timeout period, default 5, each retry takes 2000/5=400 msec by default
  ; retval:    0                         - fail
  ;            <> 0                      - success
  ;
  retval = 0                                                              ; report failure when trying to copy a folder or non-existing file
  s_handle = 0
  d_handle = 0
  ;
  If FileSize(s_name) < 0                                                 ; fail if there is no source
  ElseIf s_name = d_name                                                  ; fail if source and destination have the same name
  ElseIf s_nr = d_nr And s_nr <> #PB_Any                                  ; fail if source and destination have the same nr except for #pb_any
  Else
    t = GetTickCount_()+timeout
    s_handle = x_exreadfile(s_nr,s_name,0,timeout,retries)                ; try to read source
    While s_handle = 0 And GetTickCount_() < t
      Delay(timeout/retries)
      s_handle = x_exreadfile(s_nr,s_name,0,timeout,retries)
    Wend
    If s_handle <> 0                                                      ; succes opening the source
      d_handle = x_excreatefile(d_nr,d_name,0,timeout,retries)            ; try to create destination
      If d_handle = 0
        d_handle = x_excreatefile(d_nr,d_name,0,timeout,retries)     
        Delay(timeout/retries)
      EndIf
    EndIf
    If s_handle <> 0 And s_nr <> #PB_Any
      s_handle = s_nr
    EndIf
    If d_handle <> 0 And d_nr <> #PB_Any
      d_handle = d_nr
    EndIf
    If s_handle <> 0 And d_handle <> 0                                    ; success reading source and creating destination
      While Not Eof(s_handle)                                             ; copy data
        WriteByte(d_handle,ReadByte(s_handle))
      Wend
      retval = 1
    EndIf
    If s_handle <> 0                                                      ; close source
      CloseFile(s_handle)
    EndIf
    If d_handle <> 0                                                      ; close destination
      CloseFile(d_handle)
    EndIf
  EndIf
  ; 
  ProcedureReturn retval
  ;
EndProcedure

Procedure.i x_ll_openfile(filename.s,mode.i)                              ; create or open an existing file
  Protected share_mode.i, access_mode.i, flags.i, creation_mode.i
  ; Global x_retval.i
  ;
  ; *** open a file for buffered read, handles are shared with normal openfile() etc.
  ;
  ; the x_ll_... procedures ara all using lower level winapi calls and are not buffered or otherwise wrapped by purebasic
  ;
  ; in:        filename.s                - filename
  ;            mode.l = #x_file_r        - read
  ;                     #x_file_rw       - read/write
  ;                     #x_file_w        - write
  ;                     #x_file_shared   - allow other files to read / write to the same file while open
  ;                     #x_file_hidden   - hidden file
  ;                     #x_file_readonly - read only file
  ;                     #x_file_system   - system file
  ;                     #x_file_nocache  - don't cache
  ; retval:    0                         - failed
  ;            <> 0                      - succeeded, contains handle
  ;
  ;
  If mode & #x_file_rw
    access_mode = #GENERIC_READ | #GENERIC_WRITE
  ElseIf mode & #x_file_w
    access_mode = #GENERIC_WRITE
  Else
    access_mode = #GENERIC_READ
  EndIf
  ;
  If mode & #x_file_shared
    share_mode = #FILE_SHARE_WRITE|#FILE_SHARE_READ
  Else
    share_mode = 0
  EndIf
  ;
  ; does it matter if i switch these options on or off?
  ; (i think it's slightly faster with random always switched on)
  ;
  ; flags = #FILE_FLAG_SEQUENTIAL_SCAN
  flags = #FILE_FLAG_RANDOM_ACCESS
  ;
  If mode & #x_file_nocache
    flags = flags | #FILE_FLAG_WRITE_THROUGH
  EndIf
  If mode & #x_file_hidden
    flags = flags | #FILE_ATTRIBUTE_HIDDEN
  EndIf
  If mode & #x_file_readonly
    flags = flags | #FILE_ATTRIBUTE_READONLY
  EndIf
  If mode & #x_file_system
    flags = flags | #FILE_ATTRIBUTE_READONLY
  EndIf
  ;
  If mode & #x_file_w
    creation_mode = #OPEN_ALWAYS
  Else
    creation_mode = #OPEN_EXISTING
  EndIf
  If mode & #x_file_new
    creation_mode = #CREATE_ALWAYS
  EndIf
  ;
  x_retval = CreateFile_(@filename,access_mode,share_mode,0,creation_mode,flags,0)
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_ll_createfile(filename.s,mode.i)                            ; create a new empty file
  ; Global x_retval.i
  ;
  mode = mode | #x_file_new | #x_file_w
  If mode & #x_file_readonly
    mode = mode ! #x_file_readonly
  EndIf
  x_retval = x_ll_openfile(filename.s,mode)
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_ll_readfile(filename.s,mode.i)                              ; open an existing file as read only
  ; Global x_retval.i
  ;
  mode = mode | #x_file_readonly
  If mode & #x_file_w
    mode = mode ! #x_file_w
  EndIf
  x_retval = x_ll_openfile(filename.s, mode)
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_ll_closefile(file_h.i)                                      ; close file opened with x_openfile()
  ;
  ProcedureReturn CloseHandle_(file_h)
  ;
EndProcedure

Procedure.q x_ll_fileseek(file_h.i,pos.q)                                 ; seek a certain position in a file
  Protected hi.l, lo.l
  ;
  ; *** wrappers for standard file api calls, 32 bits handles max 2 gig, this should do more...
  ;
  ; in:      file_h.i    - handle from x_openfile()
  ;          pos.q       - new position
  ; retval:  .q          - new position
  ;
  ; notes:
  ;
  ; - as of version 4.x purebasic's FileSeek() function works with quads
  ;
  hi.l = pos >> 32
  lo.l = pos & $FFFFFFFF
  lo.l = SetFilePointer_(file_h,lo.l,@hi.l,#FILE_BEGIN)
  ProcedureReturn hi.l << 32 + lo.l
  ;
EndProcedure

Procedure.i x_ll_eof(file_h.i)                                            ; detect end of file
  Protected hi.l, lo.l, size.q, pos.q
  ;
  ; *** detect end of file
  ;
  hi.l = 0
  lo.l = GetFileSize_(file_h,@hi.l)
  size.q = hi.l << 32 + lo.l
  ;
  hi.l = 0
  lo.l = 0
  lo.l = SetFilePointer_(file_h,lo,@hi,#FILE_CURRENT)
  pos.q = hi.l << 32 + lo.l
  ;
  If pos < size
    ProcedureReturn #False
  Else
    ProcedureReturn #True
  EndIf
EndProcedure

Procedure.q x_ll_lof(file_h.l)                                            ; return length of file
  Protected hi.l, lo.l
  ;
  ; notes:
  ;
  ; - as of version 4.x purebasic's Lof() function works with quads
  ; - needs some verification when dealing with 64 bit os'es... nooooo! not another rewrite! :-(
  ;
  hi.l = 0
  lo.l = GetFileSize_(file_h,@hi.l)
  ProcedureReturn (hi.l << 32 + lo.l)
EndProcedure

Procedure.q x_ll_loc(file_h.i)                                            ; retrieve position in file
  Protected hi.l, lo.l
  ;
  ; notes:
  ;
  ; - as of version 4.x purebasic's Loc() function works with quads
  ; - needs some verification when dealing with 64 bit os'es... not another rewrite! :-(
  ;
  hi.l = 0
  lo.l = SetFilePointer_(file_h,0,@hi.l,#FILE_CURRENT)
  ProcedureReturn (hi.l << 32 + lo.l)
EndProcedure

Procedure.b x_ll_readbyte(file_h.i)                                       ; api read: read one byte (8 bits)
  Protected byte.b, bytes_read.i
  ;
  bytes_read.i = 0                                                     ; pretty irrelevant in this case :-)
  byte.b = 0
  ReadFile_(file_h,@byte,1,@bytes_read,0)
  ;
  ProcedureReturn byte.b
EndProcedure

Procedure.w x_ll_readword(file_h.i)                                       ; api read: read one word (16 bits)
  Protected word.w, bytes_read.i
  ;
  bytes_read.i = 0
  word.w = 0
  ReadFile_(file_h,@word,2,@bytes_read,0)
  ;
  ProcedureReturn word.w
EndProcedure

Procedure.l x_ll_readlong(file_h.i)                                       ; api read: read one long (32 bits)
  Protected long.l, bytes_read.i
  ;
  bytes_read.i = 0
  long.l = 0
  ReadFile_(file_h,@long,4,@bytes_read,0)
  ;
  ProcedureReturn long.l
EndProcedure

Procedure.q x_ll_readquad(file_h.i)                                       ; api read: read one long (32 bits)
  Protected quad.q, bytes_read.i
  ;
  bytes_read.i = 0
  quad.q = 0
  ReadFile_(file_h,@quad,8,@bytes_read,0)
  ;
  ProcedureReturn quad.q
EndProcedure

Procedure.i x_ll_readdata(file_h.i,buffer_p.i,bytes_to_read.i)            ; api read: read a block of data (n bytes)
  Protected bytes_read.i
  ; Global x_retval.i
  ;
  ; in:     file_h.i            - handle of file
  ;         buffer_p.i          - start of memory block to store data in
  ;         bytes_to_read.i     - number of bytes to read
  ; retval: .i                  - number of bytes read
  ; out:    x_retval.i = n      - number of bytes read
  ;
  bytes_read.i = 0
  x_retval = ReadFile_(file_h,buffer_p,bytes_to_read,@bytes_read,0)
  If x_retval = 0 Or bytes_read  = 0
    x_retval = 0
    ProcedureReturn 0
  Else
    x_retval = bytes_read
    ProcedureReturn x_retval
  EndIf
EndProcedure

Procedure.s x_ll_readstring(file_h.i,mode.i=#PB_UTF8,size.i=4096)         ; api read: read a string
  Protected exit.i, bytes_read.i, p.q, *buffer_p.BYTE, buffer_m.i, buffer_l.i, chunksize.i
  ; Global x_retval.i, x_retval_string.s
  ;
  ; *** read a text string from a file, read until a null, cr, lf, crlf, or eof is hit
  ;
  ; in:      file_h.l          - handle received from x_openfile()
  ;          [ flags.l ]       - #PB_Ascii, #PB_UTF8 (default) or #PB_Unicode
  ;          [ size.l ]        - max size of string in characters
  ; retval:  .s                - string
  ; out:     x_retval_string.s - as retval
  ;
  If mode = #PB_Unicode                ; read in blocks of 2 when in unicode mode (utf16)
    chunksize = 2
    size = (size/2)*2                  ; make sure buffer size is a multiple of 2 when in unicode mode
  Else
    chunksize = 1
  EndIf
  ;
  *buffer_p.BYTE                       ; pointer to next element in buffer
  buffer_m = AllocateMemory(size)      ; reserve memory and store address
  buffer_l = 0                         ; length of data in buffer
  *buffer_p = buffer_m                 ; set pointer to start of buffer
  ;
  bytes_read = 0
  exit = #False
  While exit = #False And buffer_l < size
    If ReadFile_(file_h,*buffer_p,chunksize,@bytes_read,0) = 0 Or bytes_read < chunksize
      exit = #True
    ElseIf *buffer_p\b = 0
      exit = #True
    ElseIf *buffer_p\b = #CR
      exit = #True
      p = x_ll_loc(file_h)
      If ReadFile_(file_h,*buffer_p,chunksize,@bytes_read,0) = 0 Or bytes_read < chunksize
      ElseIf *buffer_p\b <> #LF
        x_ll_fileseek(file_h,p)
      Else
      EndIf
    ElseIf *buffer_p\b = #LF
      exit = #True
    Else
      buffer_l = buffer_l+chunksize
      *buffer_p = *buffer_p+chunksize
    EndIf
  Wend
  x_retval_string = x_peeks(buffer_m,buffer_l,mode)
  FreeMemory(buffer_m)
  ;
  ProcedureReturn x_retval_string
EndProcedure

Procedure.i x_ll_writebyte(file_h.i,byte.b)                               ; api write: write one byte (8 bits)
  ; Global x_retval.i
  ;
  ; in:      file_h.i   - file handle
  ;          byte.b     - byte to write
  ; retval:  0          - error
  ;          <> 0       - no error
  ; out:     x_retval.i - number of bytes written
  ;
  x_retval = 0
  ProcedureReturn WriteFile_(file_h,@byte,1,@x_retval,0)
EndProcedure

Procedure.i x_ll_writeword(file_h.i,word.w)                               ; api write: write a word (16 bits)
  ; Global x_retval.i
  ;
  x_retval = 0
  ProcedureReturn WriteFile_(file_h,@word,2,@x_retval,0)
EndProcedure

Procedure.i x_ll_writelong(file_h.i,long.l)                               ; api write: write a long (32 bits)
  ; Global x_retval.i
  ;
  x_retval = 0
  ProcedureReturn WriteFile_(file_h,@long,4,@x_retval,0)
EndProcedure

Procedure.i x_ll_writedata(file_h.i,buffer_p.i,bytes_to_write.i)          ; api write: write a block of memory to the file
  ; Global x_retval.i
  ;
  x_retval = 0
  ProcedureReturn WriteFile_(file_h,buffer_p,bytes_to_write,@x_retval,0)
EndProcedure

Procedure.i x_ll_writestring(file_h.i,string.s,mode.i=#PB_UTF8)           ; api write: write a string to the file
  Protected string_l.l, string_p.l, r.l
  ; Global x_retval.i
  ;
  ; *** write string to file
  ;
  ; in:      file_h.l      - handle of file to write to, generated by x_openfile()
  ;          string.s      - string to write
  ;          [ mode.l ]    - as pb's WriteString(), #PB_Ascii, #PB_UTF8 (default), #PB_Unicode
  ; retval:  .l = 0        - error
  ;             <> 0       - no error
  ; out:     x_retval.i    - number of bytes written
  ;
  x_retval = 0
  ;
  string_l = StringByteLength(string,mode)
  string_p = AllocateMemory(string_l)
  x_pokes(string_p,string,string_l,mode)
  ;
  r = WriteFile_(file_h,string_p,string_l,@x_retval,0)
  ;
  FreeMemory(string_p)
  ;
  ProcedureReturn r
EndProcedure

Procedure.i x_ll_writestringn(file_h.i,string.s,mode.i=#PB_UTF8)          ; api write: write a string plus crlf to the file
  ProcedureReturn x_ll_writestring(file_h,string+Chr(13)+Chr(10),mode)
EndProcedure

Procedure.s x_readpreferencestringex(name.s,defaultstring.s)              ; read multiple lines from preferences file into a formatted string
  Protected n.i
  ; Global x_retval.i
  ;
  ; *** read multiple lines from a preferences file into a string, elements in the resulting string are seperated by a pipe character
  ;
  ; in:      name.s           - field name in preference file
  ;          defaultstring.s  - default string returned if field not found
  ; retval:  .s               - string, containing multiple elements seperated by pipe characters
  ; out:     x_retval.i       - number of elements in string
  ;          x_string.s       - as retval
  ;
  ; this procedure reads multiple lines from a preferences file, the lines have the format "name" + underscore + <number>
  ; the elements in the resulting string will be seperated by pipe characters '|'
  ; these lines can be written to a preference file using x_writepreferencestrngex()
  ;
  x_retval = ReadPreferenceInteger(name,0)
  x_retval_string = ""
  n = 0
  While n < x_retval
    n = n+1
    If n > 1
      x_retval_string = x_retval_string+"|"
    EndIf
    x_retval_string = x_retval_string+ReadPreferenceString(name+"_"+Str(n),"")
  Wend
  ;
  If x_retval = 0
    x_retval_string = defaultstring
    x_retval = x_parsefield(x_retval_string,"|",#x_parse_eol|#x_parse_trim)
  EndIf
  ProcedureReturn x_retval_string
EndProcedure

Procedure.i x_writepreferencestringex(name.s,string.s)                    ; write multiple lines to preferences file from a formatted string
  Protected n.i
  ;
  ; *** store a single string containing multiple elements seperated by pipe characters as multiple lines in a preferences file
  ;
  ; in:      name.s      - field name in preference file
  ;          string.s    - string, containing multiple elements seperated by pipe characters
  ; retval:  .i          - number of elements in string
  ; out:     x_retval.i  - as retval
  ;
  ; see also x_readpreferencesstringex()
  ;
  x_retval = x_parsefield(string,"|",#x_parse_eol|#x_parse_trim)
  WritePreferenceInteger(name,x_retval)
  n = 0
  While n < x_retval
    n = n+1
    WritePreferenceString(name+"_"+Str(n),x_parse_field(n))
  Wend
  ProcedureReturn x_retval
EndProcedure


; *** filesystem

Procedure.s x_getdeepestpathpart(s.s)                                  ; returns deepest folder part from a given path
  Protected p.i
  ;
  ; *** get the 'deepest' folder from the given string, ie. c:\windows\system32\ would return system32\
  ;
  ; notes:
  ;
  ; - assume that given string is a path, regardless if it ended on a backslash or not
  ; - to stay systematic I would suggest you'd always pass it that terminating backslash :-)
  ; - returns an empty string if it could not find a 'deepest' path
  ; - might :-/ handle unc paths
  ;
  If Mid(s,2,1) = ":"                 ; handle root references
    s = x_chop(s,2,0)
  EndIf
  If Left(s,2)="\\"                   ; handle unc paths
    p = FindString(s,"\",3)
    If p > 0
      s = x_chop(s,p,0)
    Else
      s = ""
    EndIf
  EndIf
  ;
  If Right(s,1) = "\"                 ; strip trailing backslash
    s = x_chop(s,0,1)
  EndIf
  If s > ""                           ; if there is something left... :-)
    p = x_rfindstring(s,"\")          ; find a preceeding backslash
    If p > 0                          ; found a section that must be a subfolder
      s = Mid(s,p+1)+"\"
    Else
      s = s+"\"                       ; none found, assume remainder is a path
    EndIf
  EndIf
  ProcedureReturn s
EndProcedure

Procedure.s x_getparentpathpart(s.s)                                   ; returns parent folder part of a given path
  Protected q.s, p.i
  ;
  ; *** get the 'parent' folder from the given string, ie. c:\windows\system32\ would return c:\windows\
  ;
  ; notes:
  ;
  ; - assume that given string is a path, regardless if it ended on a backslash or not
  ; - if it cannot find a parent path it returns an empty string
  ; - might :-/ hancle unc paths
  ;
  If Mid(s,2,1) = ":"                 ; local root
    q = Left(s,2)
    s = Mid(s,3,#MAX_PATH)
  ElseIf Left(s,2) = "\\"             ; unc path
    p = FindString(s,"\",3)
    If p > 0                          ; valid unc path
      q = Left(s,p)
      s = Mid(s,p+1,#MAX_PATH)
    Else                              ; invalid unc path
      q = ""
      s = ""
    EndIf
  EndIf
  ;
  If s > ""                           ; if there is something to find the parent of...
    If Right(s,1) = "\"               ; strip the trailing backslash to getpathpart() returns one level higher
      s = x_chop(s,0,1)
    EndIf
    If s > ""                         ; there must be something left, in which case we'll put the leading drive / unc path in front of it
      s = q+GetPathPart(s)
    EndIf
  EndIf
  ProcedureReturn s
  ;
EndProcedure

Procedure.s x_modifypath(path.s,modifier.s)                            ; modify path with modifier
  Protected file.s, root.s, p.i, modifierpart.s, pathpart.s
  ;
  ; *** modify absolute path with absolute or relative modifier
  ;
  ; in:     path.s     - a relative or absolute path, with an optional filename
  ;         modifier.s - a relative or absolute path, with an optional filename
  ; retval: .s         - the resulting path
  ;
  ; notes:
  ;
  ; - limited unc support
  ;
  ; example:
  ;
  ;   debug x_modifypath("c:\purebasic\","..\test\z.pb") ; returns "c:\purebasic\test\z.pb"
  ;
  If GetFilePart(modifier) > ""
    file = GetFilePart(modifier)
  Else
    file = GetFilePart(path)
  EndIf
  ;
  path = GetPathPart(path)
  modifier = GetPathPart(modifier)
  ;
  root = ""
  If Len(path) >= 3 And Mid(path,2,2)=":\"
    root = Left(path,3)
    If Len(path) = 3
      path = ""
    Else
      path = Mid(path,4,#MAX_PATH)
    EndIf
  EndIf
  If Len(modifier) >= 3 And Mid(modifier,2,2)=":\"
    root = ""
    path = modifier
    modifier = ""
  EndIf
  ;
  While modifier > ""
    If path > ""
      pathpart = GetFilePart(x_chop(path,0,1))
    Else
      pathpart = ""
    EndIf
    p = FindString(modifier,"\",1)
    modifierpart = Left(modifier,p-1)
    If p = Len(modifier)
      modifier = ""
    Else
      modifier = Mid(modifier,p+1,#MAX_PATH)
    EndIf
    ;
    If modifierpart = ".."
      If pathpart = ".."
        path = path+"..\"
      Else
        path = GetPathPart(x_chop(path,0,1))
      EndIf
    ElseIf modifierpart > ""
      path = path+modifierpart+"\"
    EndIf
    ;
  Wend
  ;
  ProcedureReturn root+path+file
EndProcedure

Procedure.s x_absolutepath(s.s)                                        ; change relative path to absolute path
  Protected d.s, f.s, p.s
  ;
  ; *** turn a path into an absolute path based on current directory and re-attach filename
  ;
  ; notes:
  ;
  ; - passed parameter is considered a filename if it does not end on a backslash
  ; - might (or might not) support unc paths
  ; - leaves absolute paths intact
  ;
  f = GetFilePart(s)
  p = GetPathPart(s)
  d = GetCurrentDirectory()
  ;
  If Left(p,3) = "..\"                                    ; a relative path pointing to a higher level
    While Left(p,3) = "..\"                               ; repeat until there are no levels left
      p = x_chop(p,3,0)
      d = x_getparentpathpart(d)
    Wend
    ProcedureReturn d+p+f
  ElseIf Mid(s,2,2) = ":\"                                ; absolute path to a local drive
    ProcedureReturn s
  ElseIf Left(s,2) = "\\"                                 ; an absolute unc path
    ProcedureReturn s
  ElseIf Left(s,1) = "\" And Left(d,2) = "\\"             ; root of an unc path (does this actually make sense?)
    ProcedureReturn "\\"+StringField(d,3,"\")+s
  ElseIf Left(s,1) = "\"                                  ; root of current path
    ProcedureReturn Left(d,3)+x_chop(s,1,0)
  Else                                                    ; anything else
    ProcedureReturn d+s
  EndIf
EndProcedure

Procedure.s x_validpathname(s.s)                                       ; check / reformat given path
  ;
  ; *** checks / reformats paths, does not check for existence
  ;
  ; in:        s.s                  - path
  ; retval:                         - reformatted string, empty if invalid
  ; out:       x_retval.i           - #true if valid, #false if invalid
  ;            x_retval_string.s    - same as retval
  ;
  ; notes:
  ;
  ; - relative paths are NOT supported
  ; - this isn't totally safe, but it helps a bit
  ;
  ; input and output examples
  ;
  ; c         - c:\
  ; c:        - c:\
  ; c:\       - c:\
  ; c:\test   - c:\test\
  ; \\media4\ - \\media4\
  ;
  x_retval = #False
  ;
  ; remove doublequote chars
  ;
  If Right(s,1) = Chr(34) And Left(s,1) = Chr(34)
    s = x_chop(s,1,1)
  EndIf
  ;
  ; trim
  ;
  s = Trim(s)
  ;
  ; single char drive letter
  ;
  If Len(s) = 1 And UCase(s) >= "A" And UCase(s) <= "Z"
    s = s+":\"
  EndIf
  ;
  ; ends on a backslash? if not add one
  ;
  If Len(s) >= 2 And Right(s,1) <> "\"
    s = s+"\"
  EndIf
  ;
  ; only accept if there are no illegal characters
  ;
  If FindString(s,"/") = 0 And FindString(s,"%") = 0 And FindString(s,",") = 0 And FindString(s,"|") = 0
    If FindString(s,"&") = 0 And FindString(s,"*") = 0 And FindString(s,"\\",2) = 0 And FindString(s,";",3) = 0 And FindString(s,";") = 0
      ;
      ; is it a 'root path' like C:\ ?
      ;
      If Mid(s,2,2) = ":\"
        If UCase(Left(s,1)) >= "A" And UCase(Left(s,1)) <= "Z"
          s = UCase(Left(s,1))+Mid(s,2)
          x_retval = #True
        EndIf
      ElseIf Left(s,2) = "\\"
        x_retval = #True
      EndIf
    EndIf
  EndIf
  ;
  If x_retval = #True
    x_retval_string = s
    ProcedureReturn s
  Else
    x_retval_string = ""
    ProcedureReturn ""
  EndIf
  ;
EndProcedure

Procedure.i x_createdirectory(folder.s)                                ; create a folder structure
  Protected n.i, path.s, pass.i
  ;
  ; in:        folder.s    - path
  ; retval:    0 or #False - cannot be created
  ;            1 or #True  - was succesfully created
  ; out:       x_retval.i  - as retval
  ;
  ; notes:
  ;
  ; - this function replaces the api makesuredirectorypathexists_() which doesn't work on unicode
  ; - it also replaces shcreatedirectoryex_() which requires opening a library
  ; - and it's cross platform :-)
  ; - should work with absolute and relative paths
  ; - no support for unc paths
  ;
  If Right(folder,1)<>"\"
    folder = folder+"\"
  EndIf
  folder = x_absolutepath(folder)
  ;
  x_retval = #True
  If (Mid(folder,2,1) = ":") And (Len(folder)<4 Or Mid(folder,3,1)<>"\")
    x_retval = #False                                ; error in path
  Else
    For pass = 1 To 2                                ; first pass: check for errors, second pass: create folders
      If Right(folder,1) <> "\"
        folder = folder+"\"
      EndIf
      x_parsefield(folder,"\")
      n = 0
      path = ""
      x_retval = #True
      While n < x_parse_field_n And x_retval = #True
        n = n+1
        path = path+x_parse_field(n)+"\"
        If Len(path) = 3 And Right(path,2) = ":\"
        ElseIf path = "\"
        ElseIf x_parse_field(n) = ""
          x_retval = #False                         ; error in path
        ElseIf pass = 2                             ; only create during pass 2
          If FileSize(path) = -2                    ; don't try to create if it already exists
          Else
            x_retval = CreateDirectory(path)          ; create it
          EndIf
        EndIf
      Wend
    Next pass
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_filetimetodate(*ft.FILETIME)                             ; convert windows file timestamp to pb's date format
  Protected ft2.FILETIME, st.SYSTEMTIME
  ;
  ; *** convert filetime to pb's date() format
  ;
  ; in:     pointer to filetime struct
  ; retval: date in pb format
  ;
  ; notes:
  ;
  ; - this procedure is to support winapi calls to GetFileTime_()
  ; - as of pb4.02 there is the GetFileDate() function
  ; - thanks to gpi for the help with this code, leave it to microsoft to make something complex out of something simple)
  ;
  ft2.FILETIME
  st.SYSTEMTIME
  ;
  FileTimeToLocalFileTime_(*ft,ft2)
  FileTimeToSystemTime_(ft2,st)
  ProcedureReturn Date(st\wYear,st\wMonth,st\wDay,st\wHour,st\wMinute,st\wSecond)
EndProcedure

Procedure.i x_datetofiletime(date.l,*ft.FILETIME)                      ; convert pb's date format to windows file timestamp
  Protected ft2.FILETIME, st.SYSTEMTIME
  ;
  ; *** convert pb date() format to api filetime time format
  ;
  ; in:     date.l       - date in pb format
  ;         *ft.FILETIME - pointer to an empty filetime struct that will contain the result
  ; retval: 0            - error
  ;         <> 0         - success
  ; out:    *ft.FILETIME - struct *ft.FILETIME will be overwritten
  ;
  ; notes:
  ;
  ; - as of pb4.02 there is the GetFileDate() function
  ; - thanks to gpi for the help with this code
  ;
  st.SYSTEMTIME
  ft2.FILETIME
  ;
  st\wYear=Year(date)
  st\wMonth=Month(date)
  st\wDayOfWeek=DayOfWeek(date)
  st\wDay=Day(date)
  st\wHour=Hour(date)
  st\wMinute=Minute(date)
  st\wSecond=Second(date)
  ;
  SystemTimeToFileTime_(st,ft2)
  ProcedureReturn LocalFileTimeToFileTime_(ft2,*ft)
EndProcedure

Procedure.i x_exist(filename.s)                                        ; returns true if a file exists
  ;
  ; *** check if a file exists (wrapper for filesize())
  ;
  ; in:       filename.s              - path + filename
  ; retval:   1 #True                 - exists
  ;           0 #False                - doesn't exist
  ;
  ; see also x_existex()
  ;
  If FileSize(filename) >= 0 And filename > ""
    ProcedureReturn #True
  Else
    ProcedureReturn #False
  EndIf
  ;
EndProcedure

Procedure.i x_existx(filename.s)                                       ; returns true if a file or drive exists
  Protected oldmode.l, attribute.l
  ;
  ; *** check if a file exists (as x_exist() but should skip error messages)
  ;
  ; returns true if the file exists
  ;
  If filename = ""
    ProcedureReturn #False
  Else
    oldmode = SetErrorMode_(1)
    attribute = GetFileAttributes_(@filename)
    SetErrorMode_(oldmode)
    If attribute >= 0
      ProcedureReturn #True
    Else
      GetLastError_()
      SetLastError_(0)
      ProcedureReturn #False
    EndIf
  EndIf
EndProcedure

Procedure.i x_folderexist(foldername.s)                                ; returns true if a folder exists
  If Right(foldername,1)<>"\"
    foldername = foldername+"\"
  EndIf
  If FileSize(foldername) = -2
    ProcedureReturn #True
  Else
    ProcedureReturn #False
  EndIf
EndProcedure

Procedure.i x_dir_simple(path.s,pattern.s="*.*",result_limit.i=50000)                            ; collect all filenames in a folder, windows wildcards
  Protected dir_nr.i, entry.i
  ; Global x_dir_n.i
  ;
  ; *** read all files from a folder matching pattern.s (simple alternative for x_dir() with windows / dos type of pattern matching)
  ;
  ; in:     path.s                    - starting path
  ;         pattern.s                 - one file pattern, allows windows (dos) wildcards * and ? (default *.*)
  ;         [ result_limit.i ]        - maximal number of results (default 100000)
  ; retval: > 0                       - number of entries found
  ;         = 0                       - no matches or non-existing directory
  ;         < 0                       - hit list limit
  ; out:    x_dir_file.x_dir_file()   - list containing results, first entry at position 0 (!)
  ;         x_dir_n.i                 - contains the number of results
  ;
  ; notes:
  ;
  ; - the wildcard '?' behaves differently in x_dir_simple() and x_dir() / x_matchpattern()!
  ; - in x_dir_simple() '?' stands for none or 1 character of any type (similar to dos)
  ; - in x_dir() the character '?' stands for 1 character of any type
  ; - the first entry in the list starts at position 0, as in selectelement(x_dir_file(),0)
  ; - this procedure does NOT convert relative paths to absolute paths, you may need x_absolutepath()
  ;
  ClearList(x_dir_file())
  x_dir_n = 0
  ;
  dir_nr = ExamineDirectory(#PB_Any,path,pattern)
  If dir_nr <> 0
    Repeat
      entry = NextDirectoryEntry(dir_nr)
      If entry <> 0
        If DirectoryEntryType(dir_nr) = #PB_DirectoryEntry_File
          x_dir_n = x_dir_n+1
          AddElement(x_dir_file())
          x_dir_file()\name = DirectoryEntryName(dir_nr)
          x_dir_file()\type = #PB_DirectoryEntry_File
        EndIf
      EndIf
    Until entry = 0 Or ( x_dir_n >= result_limit And result_limit <> -1)
    FinishDirectory(dir_nr)
  EndIf
  ;
  If dir_nr = 0
    ProcedureReturn 0
  ElseIf x_dir_n = result_limit And result_limit <> -1
    ProcedureReturn 0-x_dir_n
  ElseIf x_dir_n = 0
    ProcedureReturn 0
  Else
    SortStructuredList(x_dir_file(),#PB_Sort_Ascending,OffsetOf(x_dir_file\name),TypeOf(x_dir_file\name))
    ProcedureReturn x_dir_n
  EndIf
  ;
EndProcedure

Procedure.i x_dir(Path.s,pattern.s="*.*",mode.i=2,drill_limit.i=200,result_limit.i=50000)        ; collect al filenames in all (sub)folders, matchpattern wildcards, mode 2 = path+filename+drill
  Protected nr.i, deep.i, deepstart.i, type.i, z.i
  ; Global x_dir_n.i
  ;
  ; *** read contents from one or more subdirectories and store them in a linked list
  ;
  ; in:     path.s              - starting path
  ;         [ pattern.s ]       - one or more patterns using wildcards * ? and seperator | (default *.*)
  ;         [ mode.i = 0]       - 0 filename, no drilling
  ;                    1        - 1 filename and foldername+\, no drilling
  ;                    2        - 2 path+filename, drill (default)
  ;                    3        - 3 path+filename and path+foldername, drill, will include all folder names processed even when not containing matching files
  ;                    4        - 4 filename and [foldername], no drill
  ;                    5        - 5 path+filename and path+foldername, drill, will include only folders containing files with matching names
  ;                    6        - 6 path+filename and path+foldername, drill, will include files with matching filenames, and all files inside folders with matching foldernames
  ;         [ drill_limit.i ]   - number of levels (subdirs) to explore (default: 200)
  ;         [ result_limit.i ]  - stop after finding 'n' results (default 50000)
  ; retval: > 0                 - number of entries found (listed subdirs count!)
  ;         0                   - nothing found
  ;         < 0                 - hit result limit
  ; out:    x_dir_n.i           - number of elements in the list
  ;         x_dir_file.s()      - list with everything
  ;           \name             - file or folder name
  ;           \size             - size in bytes
  ;           \type             - #PB_DirectoryEntry_Directory if it's a folder, #PB_DirectoryEntry_File if it is a file
  ;           \deep             - how many levels deep the file was found, as seen from the starting path
  ;
  ; notes:
  ;
  ; - x_dir() does NOT support (relative) paths starting with ..\ or .\
  ; - x_dir() does NOT support drive only refs as C:FILENAME without the slash
  ; - x_dir() does NOT support wildcards in foldernames
  ; - * and ? behave different from dos / windows, see x_matchpattern()
  ; - *.* does NOT mean all files, it means all files with a dot in their names! use * for all files...
  ; - this procedure supports unc paths but be carefull with them
  ;
  If Right(path,1)<>"\"                             ; make sure path ends with a \
    path = path + "\"
  EndIf
  ;
  path = x_absolutepath(path)                       ; turn eventual relative paths into absolute paths
  ;
  deepstart = CountString(path,"\")                 ; count how many levels the given path is deep
  ClearList(x_dir_file.s())                         ; clear the result list
  x_dir_n = 0                                       ; set counter to zero
  ClearList(x_dir_folder.s())                       ; clear the folder list
  AddElement(x_dir_folder())                        ; store the first valid path in the folder list
  x_dir_folder() = path
  ResetList(x_dir_folder())
  ;
  Select mode                                      ; add starting element to the list
      Case 3, 5, 6
        x_dir_n = x_dir_n+1
        AddElement(x_dir_file())
        x_dir_file()\name = path
        x_dir_file()\type = #PB_DirectoryEntry_Directory
        x_dir_file()\size = 0
        x_dir_file()\deep = 0
        ; deepstart = deepstart-1
  EndSelect
  ;
  While NextElement(x_dir_folder()) And ( x_dir_n < result_limit Or result_limit = -1 )  ; go through every entry in the list of folders
    path = x_dir_folder()
    DeleteElement(x_dir_folder())
    deep = CountString(path,"\")-deepstart
    If deep > drill_limit And drill_limit <> -1                                           ; don't process if too deep
    Else                                                                                  ; or process within depth
      nr = ExamineDirectory(#PB_Any,path,"*.*")                                           ; examine each folder
      If nr > 0
        While NextDirectoryEntry(nr) And ( x_dir_n < result_limit Or result_limit = -1 )  ; check each entry
          type = DirectoryEntryType(nr)
          If Left(DirectoryEntryName(nr),1)="."                                           ; skip . and ..
          ElseIf type = #PB_DirectoryEntry_File                                           ; process all file entries
            If x_matchpattern(DirectoryEntryName(nr),pattern,1)=1                         ; add them if they match
              AddElement(x_dir_file())
              x_dir_n = x_dir_n+1
              x_dir_file()\name = path+DirectoryEntryName(nr)
              x_dir_file()\type = type
              x_dir_file()\size = DirectoryEntrySize(nr)
              x_dir_file()\deep = deep
            EndIf
          ElseIf type = #PB_DirectoryEntry_Directory                                      ; process all folder entries
            Select mode
                Case 2, 3, 5, 6                                                           ; add element to folder list for later drilling
                  AddElement(x_dir_folder())
                  x_dir_folder() = path+DirectoryEntryName(nr)+"\"
            EndSelect
            Select mode
                Case 1, 3, 4, 5, 6                                                        ; add element to file list for later listing
                  AddElement(x_dir_file())
                  x_dir_n = x_dir_n+1
                  x_dir_file()\name = path+DirectoryEntryName(nr)+"\"
                  x_dir_file()\type = type
                  x_dir_file()\size = 0
                  x_dir_file()\deep = deep
            EndSelect
          EndIf
        Wend
        FinishDirectory(nr)
      EndIf
    EndIf
    ResetList(x_dir_folder())
  Wend
  ClearList(x_dir_folder())
  ;
  SortStructuredList(x_dir_file(),#PB_Sort_Ascending,OffsetOf(x_dir_file\name),TypeOf(x_dir_file\name))
  ResetList(x_dir_file())
  ;
  ; process / filter the result list depending on mode
  ; basically the list has just been sorted above, but i might have to mess around a bit to match the requested mode / format
  ; (some stuff is mode dependant and done earlier in the code, this is the final touch up)
  ;
  Select mode
      Case 0
        ;
        ; mode 0 - filename, no drilling
        ; only files were stored, now strip out the path sections
        ;
        While NextElement(x_dir_file())
          x_dir_file()\name = GetFilePart(x_dir_file()\name)
        Wend
        ; SortStructuredList(x_dir_file(),#PB_Sort_Ascending,OffsetOf(x_dir_file\name),TypeOf(x_dir_file\name))
        ;
      Case 1
        ;
        ; mode 1 - filename and foldername\, no drilling
        ; strip out excess path, ie. keep only local level
        ;
        While NextElement(x_dir_file())
          If Right(x_dir_file()\name,1)="\"
            x_dir_file()\name = x_getdeepestpathpart(x_dir_file()\name)
          Else
            x_dir_file()\name = GetFilePart(x_dir_file()\name)
          EndIf
        Wend
      Case 2
        ;
        ; mode 2 - path+filename, drill
        ; list contains path+filename and no folders, don't do anything to it
        ;
      Case 3
        ;
        ; mode 3 - path+filename and path+foldername, drill
        ; list contains path+filename and path+foldername, don't do anything to it
        ;
      Case 4
        ;
        ; mode 4 - filename and [foldername], no drill
        ; like mode 1 but using square brackets to identify folders, then resorting
        ;
        While NextElement(x_dir_file())
          If Right(x_dir_file()\name,1)="\"
            x_dir_file()\name = "["+x_chop(x_getdeepestpathpart(x_dir_file()\name),0,1)+"]"
          Else
            x_dir_file()\name = GetFilePart(x_dir_file()\name)
          EndIf
        Wend
        SortStructuredList(x_dir_file(),#PB_Sort_Ascending,OffsetOf(x_dir_file\name),TypeOf(x_dir_file\name))
        ;
      Case 5
        ;
        ; mode 5 - path+filename and path+foldername, drill, include only folders and files if a filename in that folder contains the pattern
        ;
        deep = -1
        LastElement(x_dir_file())
        Repeat
          If x_dir_file()\type = #PB_DirectoryEntry_Directory
            If x_dir_file()\deep >= deep
              z = DeleteElement(x_dir_file())
              x_dir_n = x_dir_n-1
            Else
              deep = x_dir_file()\deep
              z = PreviousElement(x_dir_file())
            EndIf
          ElseIf x_matchpattern(GetFilePart(x_dir_file()\name),pattern,#x_parse_anycase)
            deep = x_dir_file()\deep
            z = PreviousElement(x_dir_file())
          Else
            z = DeleteElement(x_dir_file())
            x_dir_n = x_dir_n-1
          EndIf
        Until z = #False
        ;
        ; the output of mode 5 and 6 can be dumped into treegadgets:
        ;
        ;   ResetList(x_dir_file())
        ;   For n = 1 To x_dir_n
        ;     NextElement(x_dir_file())
        ;     AddGadgetItem(2,-1,x_dir_file()\name,0,x_dir_file()\deep)
        ;   Next n
        ;
      Case 6
        ;
        ; mode 6 - like mode 5 but it keeps files and keeps folders and all files in them that contains matching patterns
        ;
        deep = -1
        LastElement(x_dir_file())
        Repeat
          If x_dir_file()\type = #PB_DirectoryEntry_Directory                     ; it's a folder
            If x_matchpatternx(x_dir_file()\name,pattern,#x_parse_anycase)        ; folder and matches pattern, keep it
              deep = x_dir_file()\deep
              z = PreviousElement(x_dir_file())
            ElseIf x_dir_file()\deep >= deep                                      ; folder but to deep, remove it
              z = DeleteElement(x_dir_file())
              x_dir_n = x_dir_n-1
            Else
              deep = x_dir_file()\deep                                            ; folder above previous file or folder, keep it
              z = PreviousElement(x_dir_file())
            EndIf
          ElseIf x_matchpatternx(x_dir_file()\name,pattern,#x_parse_anycase)      ; it's a file and matches the pattern
            deep = x_dir_file()\deep                                              ; remember how deep and keep it
            z = PreviousElement(x_dir_file())
          Else
            z = DeleteElement(x_dir_file())                                       ; non matching file can be removed
            x_dir_n = x_dir_n-1
          EndIf
        Until z = #False
  EndSelect
  ;
  If x_dir_n >= result_limit And result_limit <> -1
    ProcedureReturn 0-x_dir_n
  Else
    ProcedureReturn x_dir_n
  EndIf
EndProcedure

Procedure.i x_dir_filefolder(filefolder.s,pattern.s="*",drill_limit.i=10,result_limit.i=1000)    ; checks if a given string is a path to a file or folder and fills x_dir_file() accordingly
  Protected folder.s, file.s
  ; Global x_dir_n.i
  ;
  ; *** tries to find one or more files based upon a file or folder specification
  ;
  ; (makes life easier when processing command line parameters)
  ;
  ; in:       filefolder.s       - file or foldername, path allowed
  ;           [ pattern.s ]      - filter all matches if filefolder turns out to be a folder, default * (no filtering)
  ;           [ drill_limit.l ]  - how many levels deep (default 10 don't drill, unlimited -1)
  ;           [ result_limit.l ] - max number of results (default 1000, unlimited -1)
  ; retval:   n > 0              - number of matching entries
  ;           0                  - none found
  ;           n < 0              - error
  ; out:      x_dir_file.s()     - all entries, see x_dir()
  ;           x_dir_n.l          - number of entries, see x_dir()
  ;
  ; see also: x_dir() x_absolutepath()
  ;
  ; notes:
  ;
  ; - if filefolder.s contains a wildcard, folderpattern.s is ignored
  ; - wildcards only apply to file names, not to folder names
  ; - this procedure is using x_dir() and does not convert relative paths itself, see x_absolutepath()
  ;
  ;
  If pattern = ""
    pattern = "*"
  EndIf
  ;
  ; step 1: test if its a file
  ;
  ; note that filesize() takes wildcards, so we have to make sure that won't cause problems
  ;
  If FileSize(filefolder) >= 0 And FindString(filefolder,"*",1) <= 0 And FindString(filefolder,"?",1) <= 0
    ClearList(x_dir_file())
    AddElement(x_dir_file())
    x_dir_n = 1
    x_dir_file()\name = filefolder
    ProcedureReturn 1
  EndIf
  ;
  ; step 2: test if its a folder
  ;
  If FileSize(filefolder) = -2
    ProcedureReturn x_dir(filefolder,pattern,2,drill_limit,result_limit)
  EndIf
  ;
  ; step 3: it's a folder, test it... (most likely means it the query was wrongly put together, but what the heck...)
  ;
  If Right(filefolder,1)="\"                                     ; must be a path
    ProcedureReturn x_dir(filefolder,pattern,2,drill_limit,result_limit)
  EndIf
  ;
  ; step 4: probably a path+pattern
  ;
  folder = GetPathPart(filefolder)
  file = GetFilePart(filefolder)
  ;
  ProcedureReturn x_dir(folder,file,2,drill_limit,result_limit)
  ;
EndProcedure

Procedure.q x_diskspace(drive.s)                                                                 ; return free / used / total diskspace
  ; Global x_diskspace_free.q, x_diskspace_total.q, x_diskspace_used.q
  ;
  ; *** return free / used / total diskspace
  ;
  ; in:      drive.s               - drive name / letter
  ; retval:                        - free disk space
  ; out:     x_diskspace_free.q    - free space in kbytes
  ;          x_diskspace_total.q   - total size of disk
  ;          x_diskspace_used.q    - total bytes in use
  ;
  ; note:    return values may be too large to display using debug etc. when using purebasic x86
  ;
  SetErrorMode_(#SEM_FAILCRITICALERRORS)
  x_retval = GetDiskFreeSpaceEx_(@drive, @x_diskspace_free, @x_diskspace_total, 0)
  SetErrorMode_(0)
  If x_retval = 0
    x_diskspace_free = 0
    x_diskspace_total = 0
    x_diskspace_used = 0
  Else
    x_diskspace_free = x_diskspace_free
    x_diskspace_used = x_diskspace_total-x_diskspace_free
    x_retval= x_diskspace_free
  EndIf
  ProcedureReturn x_retval
  ;
EndProcedure


; *** registry

Procedure.i x_registry_set(topkey.i,subkey.s,name.s,type.i,value.q,string.s,create.i)        ; set / create key / value / string in registry
  Protected handle.i , result.i
  ; Global x_retval.i, x_registry_q.q
  ;
  ; *** set / create key / value / string in registry
  ;
  ; in:        topkey.i
  ;            subkey.s
  ;            name.s
  ;            type.i = 0      - see msdn, not all types supported
  ;            value.q         - value to store
  ;            string.s        - string to store
  ;            create.i = 0    - open only existing keys (error out otherwise)
  ;                     = 1    - create key if it didn't exist
  ; retval:    0               - ok
  ;            <>              - error
  ; out:       x_retval        - as retval
  ;
  ; see also:  x_registry_get(), x_registry_getx(), x_registry_setx(), x_registry_setstring()
  ;
  ; notes:
  ;
  ; - see msdn for all registry types, not all types supported
  ;
  handle = 0                 ; placeholder for key handle
  result = 0                 ; placeholder for result
  ;
  If create = 0                                                             ; try to open an existing key
    x_retval = RegOpenKeyEx_(topkey,@subkey,0,#KEY_ALL_ACCESS,@handle)
    If x_retval = 0                                                         ; key does exist
      x_retval = RegQueryValueEx_(handle,name,0,0,0,0)                      ; can we open a name?
      If x_retval <> 0                                                      ; nope, close the key
        RegCloseKey_(@handle)
      EndIf
    EndIf
  Else
    x_retval = RegCreateKeyEx_(topkey,@subkey,0,0,#REG_OPTION_NON_VOLATILE,#KEY_ALL_ACCESS,0,@handle,@result)
  EndIf
  If x_retval = 0
    Select type
        Case #REG_DWORD, #REG_DWORD_LITTLE_ENDIAN
          x_retval = RegSetValueEx_(handle,@name,0,type,@value,4)
        Case #REG_QWORD, #REG_QWORD_LITTLE_ENDIAN
          x_retval = RegSetValueEx_(handle,@name,0,type,@value,8)
        Case #REG_SZ, #REG_LINK, #REG_EXPAND_SZ
          x_retval = RegSetValueEx_(handle,@name,0,type,@string,StringByteLength(string)+#x_charsize)
        Default
          x_retval = -1
    EndSelect
    RegCloseKey_(handle)
  EndIf
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_registry_setx(key.s,type.i,value.q,string.s,create.i)                          ; set / create key / value / string in registry
  Protected topkey_s.s , subkey.s , name.s , topkey.i
  ; Global x_retval.i
  ;
  ; *** set / create int or string value for registry key, with the key in 'human readable' format
  ;
  ; in:     key.s           - complete registry key
  ;         type.i          - type of entry to create
  ;         value.q         - value to store
  ;         string.s        - string to store
  ;         create.i = 0    - just open, don't create
  ;                  = 1    - create if it didn't exist yet
  ; retval: .i = 0          - ok
  ;            <> 0         - error
  ; out:    x_retval.i      - as retval
  ;
  x_retval = -1                                     ; default return value
  ;
  If x_parse(key,"\",0)
    topkey_s = UCase(x_parsed)
    Select topkey_s
        Case "#HKEY_CLASSES_ROOT"
          topkey = #HKEY_CLASSES_ROOT
        Case "#HKEY_CURRENT_CONFIG"
          topkey = #HKEY_CURRENT_CONFIG
        Case "#HKEY_CURRENT_USER"
          topkey = #HKEY_CURRENT_USER
        Case "#HKEY_DYN_DATA"
          topkey = #HKEY_DYN_DATA
        Case "#HKEY_LOCAL_MACHINE"
          topkey = #HKEY_LOCAL_MACHINE
        Case "#HKEY_PERFORMANCE_DATA"
          topkey = #HKEY_PERFORMANCE_DATA
        Case "#HKEY_USERS"
          topkey = #HKEY_USERS
    EndSelect
    If x_parsenext()                               ; any more backslashes?
      subkey = x_parsed
      While x_parsenext() = #True                  ; keep going until there are none left
        subkey = subkey+"\"+x_parsed               ; expand the subkey
      Wend
      name = x_parsed                              ; remainder must contain the name
      x_registry_set(topkey,subkey,name,type,value,string,create)
    EndIf
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_registry_get(topkey.i,subkey.s,name.s)                                         ; read int or string value from registry
  Protected buffer_m.i , buffer_l.i , handle.i
  ; Global x_retval.i , x_registry_q.q , x_registry_s.s , x_registry_type.l
  ;
  ; *** get string value from registry
  ;
  ; in:      topkey.i
  ;          subkey.s
  ;          name.s
  ; retval:  .i = 0              - ok
  ;             = -1             - error
  ;             = -2             - found but unknown type
  ;             <> 0             - error
  ; out:     x_retval.i          - as retval
  ;          x_registry_q.q      - found value as quad
  ;          x_registry_s.s      - as string
  ;          x_registry_type.l   - type of key read, -1 is unknown / unsupported
  ;
  ; example:
  ;
  ;   Debug x_registry_get(#HKEY_CURRENT_USER,"Control Panel\Desktop","WallPaper")
  ;
  ; explanation:
  ;
  ;   1. the windows registry contains keys (the #HKEY_xxxx stuff)
  ;   2. each key contains a number of subkeys, which can be multiple levels deep (like a folder structure)
  ;   3. each subkey can contain one or more 'named values' (sometimes called differently)
  ;   4. each named value can be of a certain type
  ;
  ; sometimes people refer to keys like
  ;
  ;   #HKEY_CURRENT_USER\Control Panel\Desktop\WallPaper = 1
  ;
  ; which means the named value 'WallPaper' of subkey 'Control Panel\Desktop' is set to 1, this can be done using
  ; the alternative x_registry_getx()
  ;
  ;   Debug x_registry_getx("#HKEY_CURRENT_USER\Control Panel\Desktop\WallPaper")
  ;
  ; notes:
  ;
  ; - on windows xp and later strings in the registry are stored as unicode, on ME, 98 and earlier they are ascii
  ; - use x_registry_q.q for retrieving 64 bit numbers on 32 bit platforms
  ; - windows converts string type itself without throwing a warning, you may loose special unicode characters
  ; - not all registry types are implemented
  ; - http://msdn.microsoft.com/en-us/library/ms724884(VS.85).aspx shows all types
  ;
  handle = 0                             ; placeholder for key handle
  x_retval = -1                          ; default return value
  x_registry_type = -1                   ; default return value
  ;
  ; note: to access 64 registries from win32 you need #KEY_WOW64_64KEY = $0100
  ; noteL to access 32 bit registries from win64 you need #KEY_WOW64_32KEY = $0200
  
  If RegOpenKeyEx_(topkey,@subkey,0,#KEY_ALL_ACCESS,@handle) = 0
    ;
    buffer_l = 2048                      ; buffer size, arbitrary choice of 2048 bytes (1024 characters unicode)
    buffer_m =AllocateMemory(buffer_l)   ; buffer for returned data
    ;
    If RegQueryValueEx_(handle,@name,0,@x_registry_type,buffer_m,@buffer_l) = 0
      ;
      ; data found, could be different things, see win32.hlp for all possibilities
      ;
      Select x_registry_type
          Case #REG_DWORD, #REG_QWORD, #REG_DWORD_LITTLE_ENDIAN, #REG_QWORD_LITTLE_ENDIAN
            ;
            ; a 32 bits long
            ; a 64 bits quad
            ;
            x_registry_q = PeekQ(buffer_m)
            x_registry_s = Str(x_registry_q)
            x_retval = 0
            ;
          Case #REG_SZ, #REG_EXPAND_SZ, #REG_LINK
            ;
            ; as string
            ;
            x_registry_s = x_peeks(buffer_m,buffer_l)
            x_registry_q = Val(x_registry_s)
            x_retval = 0
            ;
          Default
            ;
            ; other types would go here... none done yet
            ; hey, with this sample it should be easy to cook up something more specific :-)
            ;
            x_retval = -2
      EndSelect
    EndIf
    ;
    FreeMemory(buffer_m)
    RegCloseKey_(handle)
    ;
  EndIf
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_registry_getx(key.s)                                                           ; read int or string value from registry
  Protected topkey.i , topkey_s.s, subkey.s , name.s
  ; Global x_retval.i , x_registry_s.s , x_registry_type.l , x_registry_q.q
  ;
  ; *** as x_registry_get() but with all key-elements together
  ;
  ; in:      key.s               - topkey\subkey\name
  ; retval:  .i = 0              - ok
  ;             <> 0             - error
  ; out:     x_retval = -3       - unrecognized key
  ;                   = -2       - found but unknown type
  ;                   = -1       - error
  ;                   = 0        - ok
  ;                   <> 0       - error
  ;          x_registry_q.q      - value found
  ;          x_registry_s.s      - as string
  ;          x_registry_type.l   - type of key read, -1 is unknown
  ;
  ; notes: see x_registry_get()
  ;
  ; example:
  ;
  ;   x_registry_getx("#HKEY_CURRENT_USER\Control Panel\Desktop\WallPaper")
  ;
  x_parse(key,"\",0)
  topkey_s = UCase(x_parsed)
  x_retval = 0
  Select topkey_s
      Case "#HKEY_CLASSES_ROOT"
        topkey = #HKEY_CLASSES_ROOT
      Case "#HKEY_CURRENT_CONFIG"
        topkey = #HKEY_CURRENT_CONFIG
      Case "#HKEY_CURRENT_USER"
        topkey = #HKEY_CURRENT_USER
      Case "#HKEY_DYN_DATA"
        topkey = #HKEY_DYN_DATA
      Case "#HKEY_LOCAL_MACHINE"
        topkey = #HKEY_LOCAL_MACHINE
      Case "#HKEY_PERFORMANCE_DATA"
        topkey = #HKEY_PERFORMANCE_DATA
      Case "#HKEY_USERS"
        topkey = #HKEY_USERS
      Default
        x_retval = -3
  EndSelect
  If x_retval > -3
    If x_parsenext()                               ; any more backslashes?
      subkey = x_parsed
      While x_parsenext() = #True                  ; keep going until there are none left
        subkey = subkey+"\"+x_parsed               ; expand the subkey
      Wend
      name = x_parsed                              ; remainder must contain the name
      x_registry_get(topkey,subkey,name)
    Else
      x_retval = -3
    EndIf
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_registry_delete(topkey.i,subkey.s,name.s)                                      ; delete name and / or subkey from registry
  Protected handle.i
  ; Global x_retval.i
  ;
  ; *** delete a subkey and / or a named value
  ;
  ; in:     topkey.i
  ;         subkey.s
  ;         name.s      - leave empty to delete the whole subkey, fill in to delete a single named value
  ; retval: .i = 0      - succeeded
  ;            <> 0     - error
  ;
  If name = ""              ; delete a complete subkey including all named values
    handle = 0
    RegOpenKeyEx_(topkey,0,0,#KEY_ALL_ACCESS,@handle)
    If handle <> 0
      x_retval = RegDeleteKey_(handle,@subkey)
      RegCloseKey_(handle)
    EndIf
  Else                      ; delete a named value from a subkey
    handle = 0
    RegOpenKeyEx_(topkey,@subkey,0,#KEY_ALL_ACCESS,@handle)
    If handle <> 0
      x_retval = RegDeleteValue_(handle,@name)
      RegCloseKey_(handle)
    EndIf
  EndIf
  ;
  If handle = 0            ; couldn't open the key so indicate an error (x_retval <> 0)
    x_retval = -1
  EndIf
  ;
  ; don't need it now, but it may come in handy one day, here's how to delete a whole bunch of values
  ; from a subkey, but not the subkey itself... could just as well just delete and recreate the key I guess :-)
  ;
  ;   handle.l = 0
  ;   RegOpenkeyEx_(topkey,@subkey,0,#KEY_ALL_ACCESS,@handle)
  ;   If handle <> 0
  ;     buffer_max = 0
  ;     RegQueryInfoKey_(topkey,0,0,0,0,0,0,0,@buffer_max,0,0,0)
  ;     buffer_max = buffer_max+1
  ;     ;
  ;     buffer_l = buffer_max
  ;     buffer = Space(buffer_l)
  ;     While RegEnumValue_(handle,0,@buffer,@buffer_l,0,0,0,0) = 0
  ;       RegDeleteValue_(handle,@buffer)
  ;       Debug buffer
  ;       buffer_l = buffer_max
  ;       buffer = Space(buffer_l)
  ;     Wend
  ;     RegCloseKey_(handle)
  ;   EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_registry_deletex(key.s)                                                        ; delete name and / or subkey from registry
  Protected topkey_s.s , topkey.i , subkey.s , name.s
  ; Global x_retval.i
  ;
  ; *** delete name and / or subkey from registry
  ;
  ; to delete single named value specify the name:
  ;
  ;   x_registry_deletex("#HKEY_CURRENT_USER\Control Panel\x_lib\name")
  ;
  ; to delete all named entries and the key itself:
  ;
  ;   x_registry_deletex("#HKEY_CURRENT_USER\Control Panel\x_lib\")
  ;
  ; higher levels could have their own named values attached, so don't mix up the following three variants:
  ;
  ;   x_registry_deletex("#HKEY_CURRENT_USER\Control Panel\x_lib\name") - deletes a named value
  ;   x_registry_deletex("#HKEY_CURRENT_USER\Control Panel\x_lib\")     - deletes all named values under 'x_lib' and the subkey 'x_lib' itself
  ;   x_registry_deletex("#HKEY_CURRENT_USER\Control Panel\x_lib")      - deletes the named value 'x_lib' under subkey 'control panel'
  ;
  ; BE CAREFULL WHEN USING THIS ONE!
  ;
  x_retval = -1                                    ; default return value
  ;
  If x_parse(key,"\",0)
    topkey_s = UCase(x_parsed)
    Select topkey_s
        Case "#HKEY_CLASSES_ROOT"
          topkey = #HKEY_CLASSES_ROOT
        Case "#HKEY_CURRENT_CONFIG"
          topkey = #HKEY_CURRENT_CONFIG
        Case "#HKEY_CURRENT_USER"
          topkey = #HKEY_CURRENT_USER
        Case "#HKEY_DYN_DATA"
          topkey = #HKEY_DYN_DATA
        Case "#HKEY_LOCAL_MACHINE"
          topkey = #HKEY_LOCAL_MACHINE
        Case "#HKEY_PERFORMANCE_DATA"
          topkey = #HKEY_PERFORMANCE_DATA
        Case "#HKEY_USERS"
          topkey = #HKEY_USERS
    EndSelect
    If x_parsenext()                               ; any more backslashes?
      subkey = x_parsed
      While x_parsenext() = #True                  ; keep going until there are none left
        subkey = subkey+"\"+x_parsed               ; expand the subkey
      Wend
      name = x_parsed                              ; remainder must contain the name
      x_registry_delete(topkey,subkey,name)
    EndIf
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure


; *** console

Procedure.s x_waitkey()                                                                      ; console only, wait for a key to be pressed
  Protected k.s
  ;
  ; *** wait for a key to be pressed, return key in uppercase 1 char string
  ;
  ; console only
  ;
  Repeat
    k = Left(UCase(Inkey()),1)
    If k = ""
      Delay(30)
    EndIf
  Until k>""
  ProcedureReturn k
EndProcedure


; *** windows and desktop

Procedure.i x_dragbarheight()                                                                ; returns default size of dragbar aka caption
  ;
  ; *** return size of dragbar (caption)
  ;
  ; retval:  n                - height of dragbar aka. titlebar aka. caption
  ;
  ; note: this routine does NOT check if the dragbar exists or not, it just returns the getsystemmetrics_(#sm_cycaption) result
  ;
  ProcedureReturn GetSystemMetrics_(#SM_CYCAPTION)
EndProcedure

Procedure x_window_exteriorsize(window_nr.i)                                                 ; returns exterior size for specifed window
  Protected rect.RECT
  ; Global x_retval_x.i, x_retval_y.i, x_retval_width.i, x_retval_height.i
  ;
  ; *** returns coordinates and size of window exterior
  ;
  ; in:    window_nr            - pb window number
  ; out:   x_retval_x.i         - window x position
  ;        x_retval_y.i         - window y position
  ;        x_retval_width.i     - window width
  ;        x_retval_height.i    - window height
  ;
  GetWindowRect_(WindowID(window_nr),@rect)
  x_retval_x = rect\left
  x_retval_y = rect\top
  x_retval_width = rect\right-rect\left
  x_retval_height = rect\bottom-rect\top
  ;
EndProcedure

Procedure.i x_totaldesktops()                                                                ; return number and range of desktops
  Protected x1.i, y1.i, x2.i, y2.i, n.i
  ; Global x_desktops_x.i, x_desktops_y.i, x_desktops_width.i, x_desktops_height.i, x_desktops_n.i, n.i
  ;
  x1 = 0
  y1 = 0
  x2 = 0
  y2 = 0
  ;
  x_desktops_n = ExamineDesktops()
  ;
  n = 0
  While n < x_desktops_n
    x1 = x_min(x1,DesktopX(n))
    y1 = x_min(y1,DesktopY(n))
    x2 = x_max(x2,DesktopX(n)+DesktopWidth(n))
    y2 = x_max(y2,DesktopY(n)+DesktopHeight(n))
    n+1
  Wend
  ;
  x_desktops_x = x1
  x_desktops_y = y1
  x_desktops_width = (x2-x1)
  x_desktops_height = (y2-y1)
  x_retval = x_desktops_n
  ;
  ProcedureReturn x_desktops_n
EndProcedure

Procedure.i x_desktopx(nr.i)
  ;
  ProcedureReturn DesktopX(nr)
EndProcedure

Procedure.i x_desktopy(nr.i)
  ;
  ; see x_desktopx()
  ;
  ProcedureReturn DesktopY(nr)
EndProcedure

Procedure.i x_window_getdesktop(window_nr.l)                                                 ; returns the number of the desktop the window is currently on
  Protected x.i, y.i, n.i, nn.i, width.i, height.i, l.f, l_best.f
  ; Global x_retval.i, x_retval_x.i, x_retval_y.i
  ;
  ; *** returns the number of the monitor and the relative coordinates the window is currently on
  ;
  ; in:      window_nr.i     - pb window nr
  ; retval:  n               - desktop number (screen number)
  ; out:     x_retval_x.i    - relative coordinate of window to the upper left corner of the monitor it is on
  ;          x_retval_y.i    - same
  ;
  ; notes:
  ;
  ; - on windows each screen is called a 'desktop' which is pretty much (but not totally!) equivalent to a monitor
  ;
  x_retval = -1
  ;
  If GetWindowState(window_nr) = #PB_Window_Normal
    ;
    x = WindowX(window_nr)
    y = WindowY(window_nr)
    ;
    width = WindowWidth(window_nr)
    height = WindowHeight(window_nr)
    ;
    ; n = ExamineDesktops()
    ; nn = 0
    ; While nn < n
    ;   If x >= x_desktopx(nn) And x < x_desktopx(nn)+x_desktopwidth(nn)
    ;     If y >= x_desktopy(nn) And y < x_desktopy(nn)+x_desktopheight(nn)
    ;       x_retval = nn
    ;       x_retval_x = x-x_desktopx(nn)
    ;       x_retval_y = y-x_desktopy(nn)
    ;     EndIf
    ;   EndIf
    ;   nn+1
    ; Wend
    ;
    l_best = 9E+99
    n = ExamineDesktops()
    nn = 0
    While nn < n
      l = Pow((x+width/2)-(x_desktopx(nn)+DesktopWidth(nn)/2),2)+Pow((y+height/2)-(x_desktopy(nn)+DesktopHeight(nn)/2),2)
      If l < l_best
        l_best = l
        x_retval = nn
        x_retval_x = x-x_desktopx(nn)
        x_retval_y = y-x_desktopy(nn)
      EndIf
      nn = nn+1
    Wend
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_window_todesktop(window_nr.i,desktop_nr.i=-1,safety.i=50)                      ; moves / resizes a window so it becomes visible on a desktop with minimal specified size
  Protected r.RECT, tb_width.i, tb_height.i, b_width.i, b_height.i, b_c.i, x.i, y.i, width.i, height.i, window_h.i, state.i
  ; Global x_retval_x.i, x_retval_y.i, x_retval_width.i, x_retval_height.i
  ;
  ; *** resize and move a window to the desktop if outside
  ;
  ; in:     window_nr.i             - pb window number
  ;         [ desktop_nr.i = n  ]   - desktop (ie. monitor) that the window should be visible on, default -1
  ;                        = -1     - ensure it's visible on some desktop, no forcing (ie. monitor)
  ;         [ safety.i ]            - minimal size (visible section of window interior, default 50
  ; retval: #true                   - was (partially) outside the desktop and thus moved
  ;         #false                  - on the desktop, or iconized, or maximized
  ; out:    x_retval_x.i            - new adjusted coordinate left top corner
  ;         x_retval_y.i            - same
  ;         x_retval_width.i        - new adjusted INTERIOR width
  ;         x_retval_height.i       - new adjusted INTERIOR height
  ;
  ; notes:
  ;
  ; - not tested for ALL possible multi monitor setups :-) but seems to work fine...
  ; - on windows every screen is called a desktop which is pretty much (but not totally) equivalent to a monitor
  ;
  window_h = WindowID(window_nr)
  ;
  state = GetWindowState(window_nr)
  If state <> #PB_Window_Normal
    SetWindowState(window_nr,#PB_Window_Normal)
  EndIf
  ;
  ; find information on the window
  ;
  x = WindowX(window_nr)
  y = WindowY(window_nr)
  width = WindowWidth(window_nr)
  height = WindowHeight(window_nr)
  ;
  ;
  ; move the window to the specified desktop, if any given
  ;
  x_window_getdesktop(window_nr)
  If desktop_nr < 0                                                 ; stay on current desktop but make sure it's visible
    x_retval_x = x
    x_retval_y = y
  ElseIf x_retval = desktop_nr                                      ; it's already on the right desktop, don't do much
    x_retval_x = x
    x_retval_y = y
  ElseIf x_retval >= 0                                              ; move the window to the requested desktop
    x_retval_x = x_desktopx(desktop_nr) + x_retval_x
    x_retval_y = x_desktopy(desktop_nr) + x_retval_y
  Else                                                              ; window was not on a valid desktop, move to the requested desktop
    x_retval_x = x_limit(x_desktopx(desktop_nr),x_desktopx(desktop_nr)+DesktopWidth(desktop_nr)-width,x)
    x_retval_y = x_limit(x_desktopy(desktop_nr),x_desktopy(desktop_nr)+DesktopHeight(desktop_nr)-height,y)
  EndIf
  ;
  ; find the size of the border
  ;
  ; my old approach:
  ;
  ;   b_width = GetSystemMetrics_(#SM_CXSIZEFRAME)*2
  ;   b_height = GetSystemMetrics_(#SM_CYSIZEFRAME)*2
  ;   b_c = GetSystemMetrics_(#SM_CYCAPTION)
  ;
  ; the new approach
  ;
  r.RECT
  GetWindowRect_(window_h,@r)                                        ; retrieve exterior size
  b_width = ( r\right-r\left ) - width                               ; total border size including dragbar
  b_height = ( r\bottom-r\top ) - height                             ; same thing vertical
  ;
  ; get information on the positions and sizes of all desktops combined
  ;
  x_totaldesktops()
  ;
  ; find the position, width and height of the windows taskbar
  ;
  r.RECT
  SystemParametersInfo_(#SPI_GETWORKAREA,0,@r,0)                     ; get dimensions of workarea
  tb_width = GetSystemMetrics_(#SM_CXSCREEN)-(r\right-r\left+1)      ; deduct taskbar size
  tb_height = GetSystemMetrics_(#SM_CYSCREEN)-(r\bottom-r\top+1)
  ;
  ; now check and move if necessary
  ;
  If x_desktops_n > 1
    If (x_desktops_width / 3) = (GetSystemMetrics_(#SM_CXSCREEN) / 3)                                                     ; only one screen horizontally and no quick launch panel
      x_retval_width = x_limit(width,0,r\right-r\left-b_width)
      x_retval_x = x_limit(x_retval_x,r\left,r\right-x_retval_width-b_width)
    ElseIf ( x_desktops_x < 0 And r\left = 0 And tb_width > 0 And ((x_desktops_x+x_desktops_width) / 3 = GetSystemMetrics_(#SM_CXSCREEN) / 3) )
      x_retval_width = x_limit(width,safety*2,x_desktops_width-tb_width-safety)                                           ; primary monitor and taskbar on the right side
      x_retval_x = x_limit(x_retval_x,x_desktops_x,r\right-x_retval_width-b_width)
    ElseIf ( x_desktops_x = 0 And r\left > 0 And tb_width > 0)                                                            ; primary monitor and taskbar on the left side
      x_retval_width = x_limit(width,safety*2,x_desktops_width-tb_width-b_width)
      x_retval_x = x_limit(x_retval_x,tb_width,x_desktops_width-x_retval_width-b_width)
    Else                                                                                                                  ; no taskbar or in an unpleasant place...
      x_retval_width = x_limit(width,safety,x_desktops_width-b_width)
      x_retval_x = x_limit(x_retval_x,x_desktops_x,x_desktops_x+x_desktops_width-x_retval_width-b_width)
    EndIf
    If (x_desktops_height / 3) = ( GetSystemMetrics_(#SM_CYSCREEN) / 3)                                                   ; only one screen vertically
      x_retval_height = x_limit(height,safety,r\bottom-r\top-b_height)
      x_retval_y = x_limit(x_retval_y,r\top,r\bottom-x_retval_height-b_height)
    ElseIf ( x_desktops_y < 0 And r\top = 0 And tb_height > 0 And ( x_desktops_y+x_desktops_height ) / 3 = GetSystemMetrics_(#SM_CYSCREEN) / 3 )
      x_retval_height = x_limit(height,safety,x_desktops_height-tb_height-b_height)                                       ; primary monitor and taskbar on the bottom side
      x_retval_y = x_limit(x_retval_y,x_desktops_y,r\bottom-x_retval_height-b_height)
    ElseIf ( x_desktops_y = 0 And r\top > 0 And tb_height > 0)                                                            ; primary monitor and taskbar on the top side, or only one screen vertical
      x_retval_height = x_limit(height,safety,x_desktops_height-tb_height-b_height)
      x_retval_y = x_limit(x_retval_y,tb_height,x_desktops_height-x_retval_height-b_height)
    Else                                                                                                                  ; no taskbar or in an unpleasant place...
      x_retval_height = x_limit(height,safety,x_desktops_height-b_height)
      x_retval_y = x_limit(x_retval_y,x_desktops_y,x_desktops_y+x_desktops_height-x_retval_height-b_height)
    EndIf
    ;
  Else                                                                                                                    ; only one screen
    x_retval_width = x_limit(width,safety*2,r\right-r\left-b_width)
    x_retval_x = x_limit(x_retval_x,r\left,r\right-x_retval_width-b_width)
    x_retval_height = x_limit(height,safety,r\bottom-r\top-b_height)
    x_retval_y = x_limit(x_retval_y,r\top,r\bottom-x_retval_height-b_height)
  EndIf
  ;
  If x <> x_retval_x Or y <> x_retval_y Or width <> x_retval_width Or height <> x_retval_height
    ResizeWindow(window_nr,x_retval_x,x_retval_y,x_retval_width,x_retval_height)
    If state= #PB_Window_Maximize
      SetWindowState(window_nr,#PB_Window_Maximize)
    EndIf
    ProcedureReturn #True
  Else
    ProcedureReturn #False
  EndIf
EndProcedure

Procedure.i x_window_toprimary(window_nr.i,safety.i=50)                                      ; moves / resizes a window to the primary monitor
  ProcedureReturn x_window_todesktop(window_nr,0,safety)
EndProcedure

Procedure.i x_window_supersize(window_nr.i)                                                  ; maximize a window to cover all desktops
  If GetWindowState(window_nr) <> #PB_Window_Normal
    SetWindowState(window_nr,#PB_Window_Normal)
  EndIf
  x_totaldesktops()
  ResizeWindow(window_nr,x_desktops_x,x_desktops_y,x_desktops_width,x_desktops_height)
  ProcedureReturn x_window_todesktop(window_nr)
EndProcedure

Procedure.i x_window_toprimarycenter(window_nr.i)                                            ; move to and center on the primary screen
  If GetWindowState(window_nr) <> #PB_Window_Normal
    SetWindowState(window_nr,#PB_Window_Normal)
  EndIf
  ResizeWindow(window_nr,GetSystemMetrics_(#SM_CXSCREEN)/2-WindowWidth(window_nr)/2,GetSystemMetrics_(#SM_CYSCREEN)/2-WindowHeight(window_nr)/2,#PB_Ignore,#PB_Ignore)
  ProcedureReturn x_window_todesktop(window_nr)
EndProcedure

Procedure x_window_forcetofront(window_nr.i,window_h.i=0)                                    ; force any (other) window to the front
  Protected t1.i, t2.i
  ;
  ; *** bring any window to the front
  ;
  ; in:     window_nr.i = n         - use specified windows handle, ignored if window_h <> 0
  ;         [ window_h.i = 0 ]      - use pb window nr (default)
  ;                      = n        - use windows handle
  ;
  ; note: it should be possible to bring up the window of another process above your own, but it is not possible to
  ;       move your own window above all else if your process is not the active (focussed) one, still fooling around with this...
  ;
  If window_h = 0
    window_h = WindowID(window_nr)
  EndIf
  ;
  t1 = GetWindowThreadProcessId_(GetForegroundWindow_(),0)
  t2 = GetWindowThreadProcessId_(window_h,0)
  If t1 <> t2
    AttachThreadInput_(t1,t2,#True)
  EndIf
  SetForegroundWindow_(window_h)
  ;
  ; SetFocus_(window_h)
  ; SetWindowPos_(window_h,#HWND_TOP,0,0,0,0,#SWP_NOMOVE|#SWP_NOSIZE)
  ;
  Sleep_(256)
  If t1 <> t2
    AttachThreadInput_(t1,t2,#False)
  EndIf
  ;
  ; SetWindowPos_(window_h,#HWND_TOP,0,0,0,0,#SWP_NOMOVE|#SWP_NOSIZE)
  ;
EndProcedure

Procedure.i x_findwindow(pattern.s,ignorecase.i=#True,processid.i=0)                         ; find a window by name or processid
  Protected n.i, name.s, h.i, id.i
  ; Global x_window_name.s, x_window_h.i, x_process_id.i
  ;
  ; *** find a window by name or process id
  ;
  ; in:     pattern.s            - see x_matchpattern(), won't be used if empty
  ;         [ ignorecase.i ]     - set to #true if case doesn't matter
  ;         [ processid.i = 0]   - check for processid, ignored (default)
  ;                       = n    - find by processid, name will be ignored
  ; retval: #true                - found
  ; out:    x_window_h.i         - handle of the window found
  ;         x_process_id.i       - process id
  ;         x_window_name.s      - name of the window found
  ;
  ; notes:
  ;
  ; - the result from runprogram(... #pb_program_open) can be converted to a process id by the pb function programid()
  ;
  x_window_h = 0
  x_process_id = 0
  ;
  h = GetTopWindow_(0)
  n = 0
  While h <> 0
    name = Space(513)
    GetWindowText_(h,@name,512)                                     ; got the window name
    GetWindowThreadProcessId_(h,@id)                                ; and the process id of the owning process
    ;
    n = n+1
    If n > 500
      h = 0
    ElseIf processid = 0 And pattern = ""                           ; no conditions, return false
      h = 0
    ElseIf processid > 0 And pattern > ""
      If processid = id And x_matchpattern(name,pattern,ignorecase) = #True
        x_window_h = h
        x_process_id = id
        h = 0
      EndIf
    ElseIf processid > 0
      If processid = id
        x_window_h = h
        x_process_id = id
        h = 0
      EndIf
    ElseIf pattern > ""
      If x_matchpattern(name,pattern,ignorecase) = #True
        x_window_h = h
        x_process_id = id
        h = 0
      EndIf
    EndIf
    If h <> 0
      h = GetWindow_(h,#GW_HWNDNEXT)
    EndIf
  Wend
  ;
  If x_window_h <> 0
    x_window_name = name
    ProcedureReturn #True
  Else
    x_window_name = ""
    ProcedureReturn #False
  EndIf
  ;
EndProcedure


; *** icons

Procedure x_icon_createicon(width.i,height.i,image_h.i,mask_h.i)                             ; create icon from two bitmaps
  Protected icon.ICONINFO
  ;
  ; *** turn any two images into an icon
  ;
  ; note: you have to delete this icon manually when exiting the program using x_freeicion()
  ;
  ; how do icons behave?
  ;
  ; to replace a pixel: pixel in mask should be black, pixel in image should be whatever
  ; for no changes to a pixel: pixel in mask should be white, pixel in image should be black
  ;
  icon.ICONINFO
  icon\ficon = #True
  icon\hbmmask = mask_h
  icon\hbmcolor = image_h
  ProcedureReturn CreateIconIndirect_(@icon)
  ;
EndProcedure

Procedure x_icon_charactericon(width.i,height.i,text.s,color.i,depth.i=24)                   ; create an icon from a single character
  Protected image_nr.i, image_h.i, mask_nr.i, mask_h.i, d.i, icon_h.i
  ;
  ; *** create an icon using one or two characters as template
  ;
  ; in:  width.i      - width of the icon
  ;      height.i     - height of the icon
  ;      text.s       - character(s)
  ;      color.i      - color of the text, use RGB(red,green,blue)
  ;      depth.i      - image depth of the icon, default value 24, as of pb4.40b1 pb supports full alpha channels, and this confuses windows somewhat :-)
  ;
  ; note: you have to delete this icon manually when exiting the program using x_freeicion()
  ;
  ; step 1: draw the image
  ;
  image_nr = CreateImage(#PB_Any,width,height,depth)
  image_h = ImageID(image_nr)
  StartDrawing(ImageOutput(image_nr))
    DrawingMode(#PB_2DDrawing_Default)
    Box(0,0,width,height,0)                                                     ; background should be black so xor has no effect
    d = width/2-TextWidth(text)/2-1
    DrawingMode(#PB_2DDrawing_Transparent)
    DrawText(d,0,text,color)
  StopDrawing()
  ;
  ; step 2: the mask
  ;
  mask_nr = CreateImage(#PB_Any,width,height,depth)
  mask_h = ImageID(mask_nr)
  StartDrawing(ImageOutput(mask_nr))
    DrawingMode(#PB_2DDrawing_Default)
    Box(0,0,width,height,RGB(255,255,255))
    DrawingMode(#PB_2DDrawing_Transparent)
    DrawText(d,0,text,0)
  StopDrawing()
  ;
  icon_h = x_icon_createicon(width,height,image_h,mask_h)
  ;
  FreeImage(image_nr)
  FreeImage(mask_nr)
  ;
  ProcedureReturn icon_h
EndProcedure

Procedure x_icon_freeicon(icon_h.i)                                                          ; free icon object
  DeleteObject_(icon_h)
EndProcedure


; *** gadgets / toolbars / buttons

Procedure.i x_gadgetx(gadget_nr.i,window_nr.i)                                               ; convert gadget coordinate to window coordinate
  Protected window_h.i, gadget_h.i, rect.RECT
  ;
  ; *** turn gadget coordinate into client (window) coordinate
  ;
  ; objects on panelgadgets use a different origin (the one of the panel gadget) so we first look
  ; for the location of the object on the screen, then turn that into a location relative to the
  ; window the object is placed upon
  ;
  gadget_h = GadgetID(gadget_nr)
  window_h = WindowID(window_nr)
  rect.RECT
  GetWindowRect_(gadget_h,@rect)
  ScreenToClient_(window_h,@rect)
  ProcedureReturn rect\Left
EndProcedure

Procedure.i x_gadgety(gadget_nr.i,window_nr.i)                                               ; convert gadget coordinate to window coordinate
  Protected gadget_h.i, window_h.i, rect.RECT
  ;
  ; *** turn gadget coordinate into client coordinate
  ;
  ; see x_gadgetx()
  ;
  gadget_h = GadgetID(gadget_nr)
  window_h = WindowID(window_nr)
  rect.RECT
  GetWindowRect_(gadget_h,@rect)
  ScreenToClient_(window_h,@rect)
  ProcedureReturn rect\Top
EndProcedure

Procedure.i x_blockgadget_mousehookhandler(ncode.i,wparam.i,lparam.i)                        ; handle callbacks for x_disablegadget()
  Protected mx.i, my.i, rect.RECT
  ;
  ; see x_blockgadget()
  ;
  If wparam = #WM_LBUTTONUP Or wparam = #WM_LBUTTONDOWN                       ; only rmb or lmb messages
    ResetList(x_bggadget())
    While NextElement(x_bggadget()) <> 0
      If EventWindow() = x_bggadget()\windownr
        mx = DesktopMouseX()
        my = DesktopMouseY()
        rect.RECT
        GetWindowRect_(x_bggadget()\gadget_h,@rect)
        If mx >= rect\left And mx <= rect\right And my >= rect\Top And my <= rect\bottom And EventType()<>0
          ;
          ; save 5178: just added eventtype() to the line above to avoid confusion with requesters, not sure if it's the
          ; proper solution though it seems to work...
          ;
          ProcedureReturn 1
        EndIf
      EndIf
    Wend
  EndIf
  ;
  ProcedureReturn 0
EndProcedure

Procedure.i x_unblockgadgets(windownr.i)                                                     ; unblock all gadgets on specified window
  Protected found.i, retval.i
  ;
  ; *** remove the specified windownr and all gadgets on it from the list and unblock these gadgets
  ;
  ; in:      windownr.i = -1    - reenable all gadgets and windows. remove all hooks
  ;                     = n     - reenable all gadgets on specified window, remove hook
  ; retval:  -1                 - did not find so could not remove window from the list
  ;          0                  - removed window and hook and possibly gadgets
  ;
  ; see x_blockgadget() for more details on workings
  ;
  retval.i = -1                  ; i'm not, repeat NOT using x_retval!
  If windownr > 0
    ;
    ; delete all elements in the list of gadgets that belong to the specified window
    ;
    ResetList(x_bggadget())
    While NextElement(x_bggadget())
      If x_bggadget()\windownr = windownr
        DeleteElement(x_bggadget())
        retval = 0
      EndIf
    Wend
    ;
    ; now unhook the window and remove it from the list of windows
    ;
    found = #False
    If FirstElement(x_bgwindow())
      Repeat
        If x_bgwindow()\windownr = windownr
          found = #True
        EndIf
      Until found = #True Or NextElement(x_bgwindow()) = 0
      If found = #True
        UnhookWindowsHookEx_(x_bgwindow()\mousehook_h)
        DeleteElement(x_bgwindow())
      EndIf
    EndIf
    ;
  Else
    ;
    ; window was specified as -1 so unhook all windows, enable all gadgets and clear both lists
    ;
    ResetList(x_bgwindow())
    While NextElement(x_bgwindow()) <> 0
      UnhookWindowsHookEx_(x_bgwindow()\mousehook_h)
      retval = 0
    Wend
    ClearList(x_bgwindow())
    ClearList(x_bggadget())
    ;
  EndIf
  ;
  ProcedureReturn retval
EndProcedure

Procedure.i x_blockgadget(gadgetnr.i,windownr.i,disable.i)                                   ; disable gadget by intercepting clicks
  Protected found.i, instance_h.i, thread_id.i, window_h.i
  ; Global x_retval.i
  ; Global x_bgwindow.i(), x_bggadget.i()
  ;
  ; *** disable or enable gadgets by intercepting clicks
  ;
  ; in:     gadgetnr = m  windownr = n  disable = 1    - block events to gadget m on window n
  ;         gadgetnr = m                disable = 0    - unblock specified gadget
  ;         gadgetnr = -1 windownr = n  disable = 0    - unblock all gadgets belonging to specified window
  ;         gadgetnr = -1 windownr = -1 disable = 0    - unblock all gadgets, unhook all windows
  ;
  ; x_retval:   -1                                    - error
  ;             0                                     - no error
  ;
  ; see also purebasic 4.10 DisableGadget()
  ;
  ; notes:
  ;
  ; - does not stop people tabbing to the gadget
  ; - do not close windows once a hook has been set up for them, make sure you call x_unblockgadgets() with the proper windownumber, or x_blockgadget(-1,n,0)
  ; - nothing stops people from tabbing to the gadget, this one's just for those that have (quiet :-)) aspirations to write a visual designer :-)
  ;
  x_retval = -1
  If IsGadget(gadgetnr) And IsWindow(windownr) And disable = 1
    ;
    ; to disable a gadget i have to hook it's parentwindow
    ; the window is a valid one, time to hook it
    ;
    window_h = WindowID(windownr)                                  ; handle of window
    found = #False                                                 ; is the window already hooked?
    If FirstElement(x_bgwindow()) <> 0
      Repeat
        If x_bgwindow()\windownr = windownr
          found = #True
        EndIf
      Until found = #True Or NextElement(x_bgwindow()) = 0
    EndIf
    If found = #False                                              ; wasn't hooked yet
      AddElement(x_bgwindow())                                     ; new window to monitor
      x_bgwindow()\windownr = windownr
      instance_h = GetModuleHandle_(0)                             ; do the hooky thing
      thread_id = GetWindowThreadProcessId_(window_h, 0)
      x_bgwindow()\mousehook_h = SetWindowsHookEx_(#WH_MOUSE, @x_blockgadget_mousehookhandler(), instance_h, thread_id)
    EndIf
    ;
    ; window is added, now add gadget to blocking list
    ;
    found = #False
    If FirstElement(x_bggadget()) <> 0
      Repeat
        If x_bggadget()\gadgetnr = gadgetnr                        ; already existed
          found = #True
        EndIf
      Until found = #True Or NextElement(x_bggadget()) = 0
    EndIf
    If found = #False                                              ; didn't exist so add
      AddElement(x_bggadget())
      x_bggadget()\gadgetnr = gadgetnr
      x_bggadget()\gadget_h = GadgetID(gadgetnr)
      x_bggadget()\windownr = windownr
      x_retval = 0                                                 ; return success x_retval = 0
    EndIf
    ;
  ElseIf IsGadget(gadgetnr) And disable = 0
    ;
    found = #False
    If FirstElement(x_bggadget()) <> 0                               ; begin of list of gadgets
      Repeat
        If x_bggadget()\gadgetnr = gadgetnr                          ; found the gadget
          found = #True
          x_retval = 0
        EndIf
      Until found = #True Or NextElement(x_bggadget()) = 0
      If found = #True                                               ; found now delete
        windownr = x_bggadget()\windownr                             ; but first remember windownr
        DeleteElement(x_bggadget())                                  ; delete from list of gadgets
        found = #False
        If FirstElement(x_bggadget())                                ; now check if there were more gadgets on the same window
          Repeat
            If x_bggadget()\windownr = windownr
              found = #True
            EndIf
          Until found = #True Or NextElement(x_bggadget()) = 0       ; keep on going until found or end of list
        EndIf
        If found = #False                                            ; no more gadgets on this window
          x_unblockgadgets(windownr)                                 ; so unhook it
        EndIf
      EndIf
    EndIf
    ;
  ElseIf gadgetnr = -1 And disable = 0
    ;
    ; gadgetnr = -1 windownr = n  disable = 0    - unblock all gadgets belonging to specified window
    ; gadgetnr = -1 windownr = -1 disable = 0    - unblock all gadgets, unhook all windows
    ;
    x_retval = x_unblockgadgets(windownr)
    ;
  Else
    ;
    ; error
    ;
    x_retval = -1
    ;
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_getlisticontopline(gadget_nr.i)                                                ; what entry is currently shown on first line of a listicongadget
  ; Global x_retval.i
  ;
  ; *** get the item number of the first visible line at the top of a listicon gadget
  ;
  x_retval = SendMessage_(GadgetID(gadget_nr),#LVM_GETTOPINDEX,0,0)
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_getlisticonnextselected(gadget_nr.i, startpos.i)                               ; find the next selected entry in a listicongadget
  ; Global x_retval.i
  ;
  ; *** find the next selected entry in a listicongadget
  ;
  ; in:     gadget_nr.i       - pb listicongadget number
  ;         startpos.i = -1   - start at the top
  ;                    = n    - start searching FROM (not including!) the specified entry
  ; retval: n                 - number of entry found
  ;         -1                - no more found
  ;
  ; notes:
  ;
  ; - GetGadgetState() returns the number of the first item selected, or -1 if no items were selected
  ; - CountGadgetItems() counts the number of itmes
  ; - you can also simply loop through all items and use GetGadgetItemState() to check if they are selected or not
  ;
  If startpos = -1
    x_retval = SendMessage_(GadgetID(gadget_nr),#LVM_GETNEXTITEM,-1,#LVNI_SELECTED)
  Else
    x_retval = SendMessage_(GadgetID(gadget_nr),#LVM_GETNEXTITEM,startpos,#LVNI_SELECTED|#LVNI_BELOW)
  EndIf
  ProcedureReturn x_retval
EndProcedure

Procedure x_setlisticonvisibleline(gadget_nr.i,position.i)                                   ; make sure line n is visible
  SendMessage_(GadgetID(gadget_nr),#LVM_ENSUREVISIBLE,position,0)
EndProcedure

Procedure x_setlisticontopline(gadget_nr.i,position.i)                                       ; show line n as topline in listicon gadget
  ;
  ; *** selected line will become first line in gadget
  ;
  SendMessage_(GadgetID(gadget_nr),#LVM_ENSUREVISIBLE,position,0)
  position = 2*position-SendMessage_(GadgetID(gadget_nr),#LVM_GETTOPINDEX,0,0)
  SendMessage_(GadgetID(gadget_nr),#LVM_ENSUREVISIBLE,position,0)
EndProcedure

Procedure x_setlistviewitemtext(gadgetnr.i,itemnr.i,text.s)                                  ; set text for specific line in a listviewitem without scrolling
  Protected t.i, gadget_h.i
  ;
  ; *** set specific item of a listview gadget to a new text, just like setgadgetitemtext() but without causing a scroll
  ;
  ; in:             gadgetnr.i - gadgetid
  ;
  ; retval:                    - none
  ;
  ; out:                       - none
  ;
  ; notes:
  ;
  ; - this one is no longer necessary as SetGadgetItemText() no longer causes a scroll!
  ; - just left the code in here as a reminder and a sample, and perhaps may find some use someday :-)
  ; - quite slow compared with regular setgadgetitemtext() due to the many messages
  ;
  ; listviewgadgets can scroll up / down when overwriting an entry (this is due to the behaviour of the
  ; gadget under windows, where an entry is replaced by first sending a delete message causing the scroll,
  ; followed by an insert item message)
  ;
  gadget_h = GadgetID(gadgetnr)
  t = SendMessage_(gadget_h,#LB_GETTOPINDEX,0,0)
  SendMessage_(gadget_h,#WM_SETREDRAW,#False,0)
  SetGadgetItemText(gadgetnr,itemnr,text,0)
  SendMessage_(gadget_h,#LB_SETTOPINDEX,t,0)
  SendMessage_(gadget_h,#WM_SETREDRAW,#True,0)
EndProcedure

Procedure.i x_editorgadget(nr.i,x.i,y.i,width.i,height.i,text.s,tooltip.s,flags.i)           ; open a modified editor gadget
  Protected h.i
  ; Global x_retval.i
  ;
  ; *** create a wordwrap version of the editor gadget including default text and tooltip
  ;
  ; in:     nr.l              - gadget nr
  ;         x.l               - x position
  ;         y.l               - y position
  ;         width.l           - width
  ;         height.l          - height
  ;         text.s            - default text
  ;         tooltip.s         - tooltop
  ;         flags             - #x_gadget_wordwrap | #x_gadget_notify
  ; retval:                   - handle or gadgetnr (using nr = #pb_any)
  ; out:    x_retval.i        - as retval
  ;
  ; notes / steps:
  ;
  ;   1. create a parent window, then set up a callback to a procedure that will handle the callback
  ;
  ;      OpenWindow( ... )
  ;      SetWindowCallback(@callback_handler())
  ;
  ;   2. an example for a callback handler:
  ;
  ;      Procedure callback_handler(windowid.i,message.i,wparam.i,lparam.i)
  ;        Protected *nmhdr.NMHDR
  ;        Protected *nmlvcustomdraw.NMLVCUSTOMDRAW
  ;        Protected gadget_h.i, item_nr.i
  ;        Select message
  ;        ...
  ;        Case #WM_NOTIFY
  ;          *nmlvcustomdraw = lparam                              ; use to do custom drawing
  ;          *nmhdr.NMHDR = lparam                                 ; use to grab tab key
  ;          gadget_h = *nmhdr\hwndFrom                            ; handle of gadget that caused event
  ;          If GetAsyncKeyState_(#VK_TAB) = 32768                 ; this is all about processing tab keys etc.
  ;            Select gadget_h                                     ; we can process multiple gadgets upon a tab
  ;            Case GadgetID( x )                                  ; tab was pressed on gadget number 'x'
  ;              SetActiveGadget( x+1 )                            ; or whatever is the next gadget
  ;              ProcedureReturn 1                                 ; don't do anything else with this message
  ;            ...
  ;            EndSelect
  ;          EndIf
  ;        ...
  ;        EndSelect
  ;        ProcedureReturn #PB_ProcessPureBasicEvents              ; continue processing messages
  ;     EndProcedure
  ;
  ;  3. now create an editor gadget using x_editorgadget( ... , #x_gadget_notify )
  ;
  ;  4. the new editor gadget will send a #WM_NOTIFY message to the parent window (and the attached callback handler)
  ;     every time something happens to it, the callback handler can then take appropriate action depending on the
  ;     message and gadget (for example in the handler above the tab key selects the next gadget)
  ;
  If nr = #PB_Any
    nr = EditorGadget(#PB_Any,x,y,width,height)
    x_retval = nr
    h = GadgetID(nr)
  Else
    h = EditorGadget(nr,x,y,width,height)
    x_retval = h
  EndIf
  ;
  SetGadgetText(nr,text)
  GadgetToolTip(nr,tooltip)
  If ( flags & #x_gadget_wordwrap )
    SendMessage_(h, #EM_SETTARGETDEVICE, #Null, 0)
  EndIf
  If ( flags & #x_gadget_notify )
    SendMessage_(h, #EM_SETEVENTMASK, 0, #ENM_KEYEVENTS)
  EndIf
  ;
  ProcedureReturn x_retval
  ;
EndProcedure

Procedure x_editorgadget_select(gadgetnr.i,beginrow.i=-2,begincolumn.i=-2,endrow.i=-2,endcolumn.i=-2)    ; select section
  Protected range.CHARRANGE, id.i
  ;
  ; *** select a section of text in an editorgadget
  ;
  ; in:    gadgetnr.i            - gadget number
  ;        beginrow.i     =  n   - startpoint at row 'n'
  ;                         -1   - select from first line first column of text
  ;                         -2   - select all text, ignore all further parameters
  ;        begincolumn.i  =  n   - startpoint at column 'n'
  ;                         -1   - startpoint at column 0 ie. start of line
  ;                         -2   - select whole row, ignore all further parameters
  ;        endrow.i       =  n   - endpoint at row 'n'
  ;                         -1   - select until last line last column of text, ignore all further parameters
  ;                         -2   - endpoint will be the same as startpoint, ignore all further parameters
  ;        endcolumn.i    =  n   - endpoint at column 'n'
  ;                         -1   - endpoint at last character of current line
  ;                         -2   - endpoint at same column as start point
  ;
  ; call variations:
  ;
  ;        ( <gadget> )                                        - select whole gadget
  ;        ( <gadget> , <row> )                                - select a single full row
  ;        ( <gadget> , <row> , <column> )                     - place the cursor at the specified row / column
  ;        ( <gadget> , <row> , <column> , <row> , <column> )  - select all between startpoint and eindpoint
  ;
  ; examples:
  ;
  ;        ( 5 )                                               - select all text in gadget number 5
  ;        ( 5 , 0 )                                           - select top row in gadget number 5
  ;        ( 5 , 0 , 0 )                                       - set the cursor in front of the first character of the first line
  ;        ( 5 , 0 , 0 , 0 , 1 )                               - select the first character of the first line
  ;
  ; notes:
  ;
  ; - colums and rows start counting at 0 !!!!
  ; - credits to fr34k, gnozal, sparkie, hroudtwolf
  ; - topic on the forum http://www.purebasic.fr/english/viewtopic.php?t=6666&start=0
  ;
  
  id = GadgetID(gadgetnr)
  ;
  If beginrow = -1                                                           ; specifying row -1 selects from start of text
    beginrow = 0
    begincolumn = 0
  ElseIf beginrow = -2                                                       ; allowing a call (<gadget>) without parameters to select all text
    begincolumn = 0
    beginrow = 0
    endrow = -1
    endcolumn = -1
  EndIf
  ;
  If begincolumn = -1                                                        ; specifying begincolumn -1 selects from start of line
    begincolumn = 0
  ElseIf begincolumn = -2                                                    ; allowing a call (<gadget>,<n>) to select row <n> as a whole line
    begincolumn = 0
    endrow = beginrow
    endcolumn = -1
  EndIf
  ;
  range\cpMin = SendMessage_(id, #EM_LINEINDEX, beginrow, 0) + begincolumn
  ;
  If endrow = -1                                                             ; specifying endrow -1 selects to end of text
    endrow = SendMessage_(id, #EM_GETLINECOUNT, 0, 0)-1
    endcolumn = -1
  ElseIf endrow = -2                                                         ; specifying endrow -2 sets endpoint same as beginpoint
    endrow = beginrow
    endcolumn = begincolumn
  EndIf
  ;
  range\cpMax = SendMessage_(id, #EM_LINEINDEX, endrow, 0)
  ;
  If endcolumn = -2                                                          ; specifying endcolumn -2 sets endcolumn to startcolumn position
    endcolumn = begincolumn
  EndIf
  If endcolumn = -1                                                          ; specifying endcolumn -1 selects until end of line
    range\cpMax = range\cpMax + SendMessage_(id, #EM_LINELENGTH, range\cpMax, 0)
  Else
    range\cpMax = range\cpMax + endcolumn
  EndIf
  ;
  SendMessage_(id, #EM_EXSETSEL, 0, @range)
  ;
EndProcedure

Procedure x_editorgadget_frontcolor(gadgetnr.i,color.i)                                      ; set front color for selected
  Protected format.CHARFORMAT, id.i
  ;
  ; *** set current selection to specified color
  ;
  id = GadgetID(gadgetnr)
  format\cbSize = SizeOf(CHARFORMAT)
  format\dwMask = #CFM_COLOR
  format\crTextcolor = color
  SendMessage_(id, #EM_SETCHARFORMAT, #SCF_SELECTION, @format)
  ;
EndProcedure

Procedure x_editorgadget_backcolor(gadgetnr.i,color.i)                                       ; set back color for selected
  Protected format.CHARFORMAT2, id.i
  ;
  ; *** set background of current selection to specified color
  ;
  id = GadgetID(gadgetnr)
  format\cbSize = SizeOf(CHARFORMAT2)
  format\dwMask = #CFM_BACKCOLOR
  format\crBackColor = color
  SendMessage_(id, #EM_SETCHARFORMAT, #SCF_SELECTION, @format)
  ;
EndProcedure

Procedure x_editorgadget_setfontsize(gadgetnr.i,fontsize.i)                                  ; set fontsize for selected
  Protected format.CHARFORMAT, id.i
  ;
  ; *** set current selection to specified font size
  ;
  id = GadgetID(gadgetnr)
  format\cbSize = SizeOf(CHARFORMAT)
  format\dwMask = #CFM_SIZE
  format\yHeight = fontsize*20
  SendMessage_(id, #EM_SETCHARFORMAT, #SCF_SELECTION, @format)
  ;
EndProcedure

Procedure x_editorgadget_setfont(gadgetnr.i,fontname.s)                                      ; set font for selected
  Protected format.CHARFORMAT, id.i
  ;
  ; *** set current selection to specified font
  ;
  ; note: specifying the name is enough, no need to load the font
  ;
  id = GadgetID(gadgetnr)
  format\cbSize = SizeOf(CHARFORMAT)
  format\dwMask = #CFM_FACE
  PokeS(@format\szFaceName, fontname)
  SendMessage_(id, #EM_SETCHARFORMAT, #SCF_SELECTION, @format)
  ;
EndProcedure

Procedure x_editorgadget_setstyle(gadgetnr.i,style.i)                                        ; set style for selected
  Protected format.CHARFORMAT, id.i
  ;
  ; *** apply a specific style to selection
  ;
  ; in:    gadgetnr.i    - gadget number
  ;        style.i       - style, supported values #CFM_BOLD, #CFM_ITALIC, #CFM_UNDERLINE and #CFM_STRIKEOUT
  ;
  id = GadgetID(gadgetnr)
  format\cbSize = SizeOf(CHARFORMAT)
  format\dwMask = #CFM_ITALIC|#CFM_BOLD|#CFM_STRIKEOUT|#CFM_UNDERLINE
  format\dwEffects = style
  SendMessage_(id, #EM_SETCHARFORMAT, #SCF_SELECTION, @format)
  ;
EndProcedure

Procedure x_editorgadget_cut(gadgetnr.i)                                                     ; cut selected
  ;
  ; *** cut selection from editor gadget
  ;
  SendMessage_(GadgetID(gadgetnr), #WM_CUT,0,0)
  ;
EndProcedure

Procedure x_editorgadget_copy(gadgetnr.i)                                                    ; copy selected
  ;
  ; *** copy selection from editor gadget
  ;
  SendMessage_(GadgetID(gadgetnr), #WM_COPY,0,0)
  ;
EndProcedure

Procedure x_editorgadget_paste(gadgetnr.i)                                                   ; replace selected with buffer
  ;
  ; *** copy selection from editor gadget
  ;
  SendMessage_(GadgetID(gadgetnr), #WM_PASTE,0,0)
  ;
EndProcedure

Procedure x_editorgadget_replace(gadgetnr.i,text.s)                                          ; replace selected with text
  ;
  ; *** replace selection with text
  ;
  ProcedureReturn SendMessage_(GadgetID(gadgetnr),#EM_REPLACESEL,0,@text)
  ;
EndProcedure

Procedure x_editorgadget_cursorcolumn(gadgetnr.i)
  Protected range.CHARRANGE, id.i
  ;
  range.CHARRANGE
  id = GadgetID(gadgetnr)
  SendMessage_(id,#EM_EXGETSEL,0,range.CHARRANGE)
  ProcedureReturn (range\cpMax-(SendMessage_(id,#EM_LINEINDEX,SendMessage_(id,#EM_EXLINEFROMCHAR,0,range\cpMin),0)))
  ;
EndProcedure

Procedure x_editorgadget_cursorrow(gadgetnr)
  Protected range.CHARRANGE, id.i
  ;
  range.CHARRANGE
  id = GadgetID(gadgetnr)
  SendMessage_(id,#EM_EXGETSEL,0,@range)
  ProcedureReturn SendMessage_(id,#EM_EXLINEFROMCHAR,0,range\cpMin)
  ;
EndProcedure

Procedure x_toolbar_radiobutton(toolbar_nr.i,buttonnr.i,image_h.i,State.i=0,Tooltip.s="")    ; add a radio button to a toolbar
  Protected buttoninfo.TBBUTTONINFO
  ;
  ; *** add a radio button (group) to a toolbar
  ;
  ; in:     toolbar_nr.i     - toolbar to add the button to
  ;         buttonnr.        - nr of button to create
  ;         image_h.i        - handle of image to use
  ;         [ state.i ]      - selection state, default unselected
  ;         [ tooltip.s ]    - optional tooltip, default empty
  ;
  ToolBarImageButton(buttonnr,image_h)
  buttoninfo.TBBUTTONINFO
  buttoninfo\cbSize = SizeOf(buttoninfo)
  buttoninfo\dwMask = #TBIF_STYLE
  buttoninfo\fsStyle = #TBSTYLE_CHECKGROUP
  SendMessage_(ToolBarID(toolbar_nr),#TB_SETBUTTONINFOA,buttonnr,@buttoninfo)
  SendMessage_(ToolBarID(toolbar_nr),#TB_CHECKBUTTON,buttonnr,State)
  ;
  If Tooltip <> ""
    ToolBarToolTip(toolbar_nr,buttonnr,Tooltip)
  EndIf
EndProcedure

Procedure x_toolbar_togglebutton(toolbar_nr.i,buttonnr.i,image_h.i,State.i,Tooltip.s)        ; add a togglebutton to a toolbar
  Protected buttoninfo.TBBUTTONINFO
  ;
  ;  *** add a toggle (check) button to a toolbar
  ;
  ; in:      toolbar_nr.l - number of the toolbar
  ;          buttonnr.l   - buttonnr and message send when pressed
  ;          image_h.l    - windows handle of image
  ;          state.l      - 0 not pressed 1 pressed
  ;          tooltip.s    - added tooltip
  ; retval:  none
  ; out:     none
  ;
  ; notes:
  ;
  ; - deprecated, as of 4.x there's the native toolbarimagebutton() with the toggle flag
  ;
  ;
  ; it's a bit of a kludge :-) first add a regular button
  ; then use sendmessage_() to change its type and status
  ;
  ToolBarImageButton(buttonnr,image_h)
  buttoninfo.TBBUTTONINFO
  buttoninfo\cbSize = SizeOf(buttoninfo)
  buttoninfo\dwMask = #TBIF_STYLE
  buttoninfo\fsStyle = #TBSTYLE_CHECK
  SendMessage_(ToolBarID(toolbar_nr),#TB_SETBUTTONINFOA,buttonnr,@buttoninfo)
  SendMessage_(ToolBarID(toolbar_nr),#TB_CHECKBUTTON,buttonnr,State)
  ;
  ; finally add a tooltip
  ;
  If Tooltip <> ""
    ToolBarToolTip(toolbar_nr,buttonnr,Tooltip)
  EndIf
EndProcedure

Procedure x_toolbar_imagebutton(toolbar_nr.i,buttonnr.i,image_h,tooltip.s="")                ; add a regular imagebutton with tooltip
  ;
  ;  *** add a regular imagebutton to a toolbar
  ;
  ; in:      buttonnr.l   - buttonnr and message send when pressed
  ;          image_h.l    - windows handle of image
  ;          tooltip.s    - added tooltip
  ; retval:  none
  ; out:     none
  ;
  ; notes:
  ;
  ; - only used to save some typing when adding regular toolbarbuttons with tooltips
  ;
  ToolBarImageButton(buttonnr,image_h)
  If Tooltip <> ""
    ToolBarToolTip(toolbar_nr,buttonnr,Tooltip)
  EndIf
EndProcedure

Procedure.i x_button_imagetogglebutton(buttonnr.i,x.i,y.i,width.i,height.i,image_h.i,state.i=0,Tooltip.s="")    ; toggle image button, deprecated
  Protected h.i
  ;
  ; ***  draw an image button that can be toggled
  ;
  ; in:      buttonnr.i    - gadget number
  ;          x.i           - x coordinate
  ;          y.i           - y coordinate
  ;          width.i       - width
  ;          height.i      - height
  ;          image_h.i     - handle to icon or image
  ;          [ state.i ]   - state, default unselected
  ;          [ tooltip.s ] - tooltip string, default empty
  ; retval:  .i            - handle of button
  ;
  ; notes:
  ;
  ; - deprecated, as of pb 4.x there's the native buttonimagegadget() with the toggle flag
  ;
  ;
  h = ButtonImageGadget(buttonnr,x,y,width,height,image_h)
  SetWindowLong_(h,#GWL_STYLE,GetWindowLong_(h,#GWL_STYLE)|#BS_AUTOCHECKBOX|#BS_PUSHLIKE)
  SendMessage_(h,#BM_SETCHECK,state,0)
  ;
  If Tooltip <> ""
    GadgetToolTip(buttonnr,Tooltip)
  EndIf
  ;
  ProcedureReturn h
EndProcedure

Procedure x_button_setstate(button_h.i,state.i)                                              ; set state of a toggle button, deprecated
  ;
  ; *** set state of a toggle button
  ;
  ; in:    button_h.i    - handle of the button (not pb's gadget number!)
  ;        state.i       - state
  ;
  ; notes:
  ;
  ; - setgadgetstate() could not toggle the state of an imagebutton gadget in older versions of pb
  ; - deprecated
  ;
  SendMessage_(button_h,#BM_SETCHECK,state,0)
EndProcedure

Procedure.i x_button_getstate(button_h.i)                                                    ; get the state of a toggle button, deprecated
  ProcedureReturn SendMessage_(button_h,#BM_SETCHECK,0,0)
EndProcedure


; *** requesters

Procedure.i x_requesterposition_callback(window.i,message.i,wparam.i,lparam.i)               ; used for x_setrequesterposition()
  Protected result.i
  ; Global x_requesterposition_x.i, x_requesterposition_y.i, x_requesterposition_callback.i
  ; Global x_requesterposition_hook.i, x_requesterposition_window.i
  ;
  Select message
      Case #WM_DESTROY
        SetWindowLong_(window,#GWL_WNDPROC,x_requesterposition_callback)
      Case #WM_CREATE                            ; update requester position (special case for InputRequester)
        SetWindowPos_(window,#HWND_TOPMOST,x_requesterposition_x,x_requesterposition_y,0,0,#SWP_NOSIZE|#SWP_NOACTIVATE)
      Case #WM_INITDIALOG                        ; update requester position
        SetWindowPos_(Window,#HWND_TOPMOST,x_requesterposition_x,x_requesterposition_y,0,0,#SWP_NOSIZE|#SWP_NOACTIVATE)
        result = CallWindowProc_(x_requesterposition_callback,window,message,wparam,lparam)
        SetWindowPos_(Window,#HWND_TOPMOST,x_requesterposition_x,x_requesterposition_y,0,0,#SWP_NOSIZE|#SWP_NOACTIVATE)
        ProcedureReturn result
  EndSelect
  ProcedureReturn CallWindowProc_(x_requesterposition_callback,window,message,wparam,lparam)
EndProcedure

Procedure.i x_requesterposition_hook(idhook.i,wparam.i,lparam.i)                             ; used for x_setrequesterposition()
  Protected *x_requesterposition_cwp.CWPSTRUCT, windowclass.s, result.i, message.i, window.i
  ; Global x_requesterposition_x.i, x_requesterposition_y.i, x_requesterposition_callback.i
  ; Global x_requesterposition_hook.i, x_requesterposition_window.i
  ;
  *x_requesterposition_cwp.CWPSTRUCT = lParam
  message = *x_requesterposition_cwp\Message
  window = *x_requesterposition_cwp\hwnd
  ;
  ; special case for InputRequester
  ;
  windowClass = Space(255)
  GetClassName_(window,@windowclass,255)
  If LCase(windowclass)="inputrequester"
    message = #WM_INITDIALOG
  EndIf
  ;
  Select message
      Case #WM_INITDIALOG                            ; modify Requester callback
        x_requesterposition_callback = SetWindowLong_(window,#GWL_WNDPROC,@x_requesterposition_callback())
        result = CallNextHookEx_(x_requesterposition_hook, idHook,wParam,lParam)
        UnhookWindowsHookEx_(x_requesterposition_hook)
        x_requesterposition_hook = 0
      Default
        result = CallNextHookEx_(x_requesterposition_hook,idhook,wparam,lparam)
  EndSelect
  ;
  ProcedureReturn result
EndProcedure

Procedure x_setrequesterposition(x.i,y.i,windownr.i)                                         ; set x/y position of the next purebasic requester
  ; Global x_requesterposition_x.i, x_requesterposition_y.i, x_requesterposition_callback.i
  ; Global x_requesterposition_hook.i, x_requesterposition_window.i
  ;
  ; *** set x/y position of the next purebasic requester
  ;
  ; in:        x.l       - horizontal position
  ;            y.l       - vertical position
  ;            windownr  - number of parentwindow
  ; retval:    none
  ; out:       none
  ;
  ; from the purebasic forum, by eddy, dunno how it works, or even if it works properly... :-)
  ;
  ; note: only call this with a window open! otherwise it may cause all sorts of unpredictable errors, if neccessary open
  ;       an invisible window with x_requesterposition_window = OpenWindow(#PB_Any,x,y,0,0,#PB_Window_Invisible,"")
  ;
  x_requesterposition_x = x
  x_requesterposition_y = y
  x_requesterposition_window = windownr
  ;
  ; modify parent hook
  ;
  x_requesterposition_hook = SetWindowsHookEx_(#WH_CALLWNDPROC,@x_requesterposition_hook(),GetModuleHandle_(0),GetWindowThreadProcessId_(WindowID(x_requesterposition_window),0))
  ;
EndProcedure

Procedure.i x_silentrequester(icon.i,title.s,text.s,buttons.i,windownr.i)                    ; silent dialogbox with icon (silent messagerequester)
  Protected alert.MSGBOXPARAMS
  ;
  ; *** silent dialogbox with icon
  ;
  ; in:       icon.i     - icon, see below
  ;           title.s    - requester title (drag bar)
  ;           text.s     - text inside the requester
  ;           buttons.i  - one or more buttons with specific text, see below
  ;           windownr.i - parent window (pb number)
  ; retval:              - none
  ; out:                 - none
  ;
  ; (after sparky's example 27.01.2006)
  ;
  ; notes:
  ;
  ; - you can use the winapi messagebox_ constants for the regular messageRequester() to add icons, however these differ
  ; - you can turn regular requesters into silent ones using #QUICKNQUIET as flag:
  ;
  ;     MessageRequester("Silent","Important Quiet Message",#QUICKNQUIET)
  ;
  ; valid icon values:
  ;
  ; #IDI_APPLICATION       default application icon
  ; #IDI_INFORMATION       asterix
  ; #IDI_ASTERISK          same
  ; #IDI_ERROR             hand (halt)
  ; #IDI_HAND              same
  ; #IDI_WARNING           exclamation point
  ; #IDI_EXCLAMATION       same
  ; #IDI_QUESTION          question mark
  ; #IDI_WINLOGO           windows logo or default application
  ;
  ; equivalents for winapi messagebox_() (don't work here)
  ;
  ; #MB_ICONEXCLAMATION, #MB_ICONWARNING            an exclamation-point icon appears in the message box
  ; #MB_ICONINFORMATION, #MB_ICONASTERISK           an icon consisting of a lowercase letter i in a circle appears in the message box
  ; #MB_ICONQUESTION                                a question-mark icon appears in the message box
  ; #MB_ICONSTOP, #MB_ICONERROR, #MB_ICONHAND       a stop-sign icon appears in the message box
  ;
  ; buttons values:
  ;
  ; #MB_YESNO              same as #pb_messagerequester_yesno
  ; #MB_YESNOCANCEL        same as #pb_messagerequester_yesnocancel
  ; #MB_OK                 same as #pb_messagerequester_ok
  ; #MB_ABORTRETRYIGNORE
  ; #MB_CANCELTRYCONTINUE  2000 / xp only
  ; #MB_RETRYCANCEL
  ;
  ; return values:
  ;
  ; #IDABORT    = 3        abort
  ; #IDCANCEL   = 2        cancel (same as #pb_messagerequester_cancel)
  ; #IDCONTINUE = 11       continue
  ; #IDIGNORE   = 5        ignore
  ; #IDNO       = 7        no (same as #pb_messagerequester_no)
  ; #IDOK       = 1        ok
  ; #IDRETRY    = 4        retry
  ; #IDTRYAGAIN = 10       try again
  ; #IDYES      = 6        yes (same as #pb_messagerequester_yes)
  ;
  alert\cbSize = SizeOf(MSGBOXPARAMS)
  alert\hwndOwner = WindowID(windownr)
  alert\hInstance = 0
  alert\lpszText = @text
  alert\lpszCaption = @title
  alert\dwStyle = buttons | #MB_USERICON
  alert\lpszIcon = icon
  ;
  x_retval = MessageBoxIndirect_(alert)
  ;
  ProcedureReturn x_retval
EndProcedure


; *** mouse and pointers

Procedure.i x_leftmousebutton()                                                              ; return current state of left mousebutton
  ;
  ; retval: #True  - mouse button down
  ;         #False - mouse button up
  ;
  ; notes:
  ;
  ; - does not use events!
  ; - use #wm_lbuttonup and #wm_lbuttondown for event based handling
  ;
  If GetSystemMetrics_(#SM_SWAPBUTTON) = 0
    ProcedureReturn x_bool(GetAsyncKeyState_(#VK_LBUTTON))
  Else
    ProcedureReturn x_bool(GetAsyncKeyState_(#VK_RBUTTON))
  EndIf
EndProcedure

Procedure.i x_rightmousebutton()                                                             ; return current state of right mousebutton
  ;
  ; notes:
  ;
  ; - does not use events!
  ; - use #wm_lbuttonup and #wm_lbuttondown for event based handling
  ;
  If GetSystemMetrics_(#SM_SWAPBUTTON) = 0
    ProcedureReturn x_bool(GetAsyncKeyState_(#VK_RBUTTON))
  Else
    ProcedureReturn x_bool(GetAsyncKeyState_(#VK_LBUTTON))
  EndIf
EndProcedure

Procedure.i x_mousewheel(mode.i)                                                             ; returns number of lines up / down upon using the scrollwheel after a #WM_MOUSEWHEEL event
  Protected d.i
  ;
  ; *** returns number of lines up / down upon using the scrollwheel after a #WM_MOUSEWHEEL event
  ;
  ; in:     mode.i = 0   - one line per message, return -1 for up, +1 for down
  ;                = 1   - use windows setting for calculating number of lines, from -n to +n
  ; retval: < 0          - up
  ;         > 0          - down
  ;
  If mode = 0
    d = x_sgn( EventwParam() >> 16 )
  Else
    ;
    ; #SPI_GETWHEELSCROLLLINES = 104
    ;
    d = 0
    SystemParametersInfo_(104,0,@d,0)
    d = (EventwParam() >> 16)*d/120
    ;
  EndIf
  ProcedureReturn d
EndProcedure

Procedure.i x_windowmousex(windownr.i)                                                       ; return window related mouse coordinate
  Protected wmx.i, dmx.i
  ;
  ; *** return window related mouse coordinate
  ;
  ; the normal windowmousex() returns -1 for any position outside the window, this returns a coordinate relative to the window interior
  ; no matter if the mouse is in- or outside the window
  ;
  wmx = WindowMouseX(windownr)
  dmx = DesktopMouseX()
  If wmx >= 0
    x_windowmousedx = dmx-wmx
  EndIf
  ProcedureReturn DesktopMouseX()-x_windowmousedx
EndProcedure

Procedure.i x_windowmousey(windownr.i)                                                       ; return window related mouse coordinate
  Protected wmy.i, dmy.i
  ;
  ; *** return window related mouse coordinate
  ;
  ; the normal windowmousex() returns -1 for any position outside the window, this returns a coordinate relative to the window interior
  ; no matter if the mouse is in- or outside the window
  ;
  wmy = WindowMouseY(windownr)
  dmy = DesktopMouseY()
  If wmy >= 0
    x_windowmousedy = dmy-wmy
  EndIf
  ProcedureReturn DesktopMouseY()-x_windowmousedy
EndProcedure


; *** other gui related

Procedure.s x_waitkeywindow(mouse.i=1,escape.i=1)                                            ; window only, wait for a closewindow message, mousebutton, or keypress
  Protected k.s, event.i, exit.i
  ;
  ; *** wait until window is closed, or a mouse button, or a key is pressed
  ;
  ; in:     escape.i = 1    ; exit on escape key (default)
  ;         mouse.i = 1     ; exit on any mouseclick (default)
  ;
  ; notes:
  ;
  ; - window only
  ; - eats up all events
  ;
  exit = #False
  Repeat
    event = WaitWindowEvent()
    If event = #WM_KEYDOWN
      k.s = Chr(EventwParam())
    EndIf
  Until event = #PB_Event_CloseWindow Or (event = #WM_KEYUP And k.s = Chr(27)) Or (mouse = 1 And (event = #WM_RBUTTONUP Or event = #WM_LBUTTONUP))
  ProcedureReturn k.s
EndProcedure

Procedure x_emptyeventqueue()                                                                ; emtpy windows event queue
  Protected event.i, n.i
  ;
  Repeat
    event = WindowEvent()
    If event <> 0
      n = 0
    Else
      n = n+1
      Delay(10)
    EndIf
  Until n = 10
EndProcedure

Procedure x_setcursor(window_h.i,type.i)                                                     ; change mouse cursor
  ;
  ; can't remember... so a small procedure to help...
  ;
  ; #IDC_ARROW                Arrow cursor
  ; #IDC_IBEAM                I-beam cursor
  ; #IDC_WAIT                 Hourglass cursor
  ; #IDC_CROSS                Crosshair cursor
  ; #IDC_UPARROW              Up arrow cursor
  ; #IDC_SIZENWSE             Sizing cursor, points northwest And southeast
  ; #IDC_SIZENESW             Sizeing cursor, points northeast And southwest
  ; #IDC_SIZEWE               Sizing cursor, points west And east
  ; #IDC_SIZENS               Sizing cursor, points north And south
  ; #IDC_SIZEALL              Sizing cursor, points north, south, east, And west
  ; #IDC_NO                   "No" cursor
  ; #IDC_APPSTARTING          Application-starting cursor (arrow And hourglass)
  ; #IDC_HELP                 Help cursor (arrow And question mark)
  ; #IDI_APPLICATION          Application icon
  ; #IDI_HAND                 Stop sign icon
  ; #IDI_QUESTION             Question-mark icon
  ; #IDI_EXCLAMATION          Exclamation point icon
  ; #IDI_ASTERISK             Asterisk icon (letter "i" in a circle)
  ; #IDI_WINLOGO              Windows logo icon
  ;
  SetClassLong_(window_h,#GCL_HCURSOR,LoadCursor_(0,type))
EndProcedure


; *** threads / tasks / mailslots / timer / execute

Procedure.i x_mailslot_open(channelname.s,mode.i=#PB_UTF8)                                   ; open mailslot for receiving messages
  ; Global x_retval.i, x_x_firstinstance_instancename.s
  ;
  ; *** open mailslot for receiving messages
  ;
  ; in:      channelname.s         - name of channel to open, see notes
  ;          mode.i                - unicode mode (not used, ignore)
  ; retval:  .i = n                - channel number (mailslot handle)
  ;             = 0                - failed
  ; out:     x_retval.i            - as retval
  ;
  ; notes:
  ;
  ; - though i call them channels these are actually mailslots
  ; - unless the name of the channel starts with two backslashes it will use the name used in x_interapp_numberone() as part of the channel (mailslot) name
  ;
  ;
  ; mailslots can exist locally, or somewhere on the network, messages can have any size unless they are broadcasted
  ; onto a domain, in which case they are limited to 400 bytes
  ;
  ; (note: I ran into samples stating this could be 450 or 540, let's stick to 400 for safety reasons)
  ;
  ; the owner always creates a mailslot using the following format:
  ;
  ;   \\.\mailslot\<path>\<name>
  ;
  ; another computer on the network could address this slot by using the computername or even the domainname, we could also
  ; address whole groups:
  ;
  ;   \\<computername>\mailslot\<path>\<name>
  ;   \\<domainname>\mailslot\<path>\<name>
  ;   \\.\mailslot\myusers\peter
  ;   \\.\mailslot\myusers\jan
  ;   \\.\mailslot\myusers\*
  ;
  ; yes, this is a brilliant stupid wrapper for pretty much a single winapi call :-)
  ;
  If Left(channelname,2) <> "\\"
    If x_x_firstinstance_instancename <> ""
      channelname = "\\.\mailslot\"+x_x_firstinstance_instancename+"\"+channelname
    Else
      channelname = "\\.\mailslot\"+channelname
    EndIf
  EndIf
  ;
  x_retval = CreateMailslot_(@channelname,0,0,0)                           ; create mailslot
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_mailslot_close(channelhandle.i)                                                ; close open mailslot
  ;
  ; *** close open mailslot
  ;
  ; in:      channelhandle.i        - number of channel to close, as returned by x_mailslot_open()
  ; retval:  .i = 0                 - failed
  ;             <> 0                - success
  ;
  ProcedureReturn CloseHandle_(channelhandle)
EndProcedure

Procedure.i x_mailslot_sendmessage(channelname.s,message.s,mode.i=#PB_UTF8)                  ; send message to specified mailslot
  Protected file_h.i, message_p.i, message_l.i
  ; Global x_firstinstance_wakeupmessage.i
  ;
  ; *** send message to specified mailslot
  ;
  ; in:     channelname.s          - name of the channel to use, see also x_mailslot_open() for format
  ;         message.s              - message to send
  ;         [ mode.i ]             - unicode mode
  ; retval  .i = n                 - number of bytes written
  ;            = -1                - error
  ;
  ; writing using regular pb commands is possible, but reading needs winapi
  ;
  ;   CreateFile(1,mailslot_name)
  ;   WriteString(message)
  ;   CloseFile(1)
  ;
  ; complete channel (mailslot) name if needed
  ;
  If Left(channelname,2) <> "\\"
    If x_x_firstinstance_instancename <> ""
      channelname = "\\.\mailslot\"+x_x_firstinstance_instancename+"\"+channelname
    Else
      channelname = "\\.\mailslot\"+channelname
    EndIf
  EndIf
  ;
  ; i'm a bit surprised I need to specify #OPEN_EXISTING, but without it there's no go...
  ; (and the docs on (using)mailslots are outright wonderful... yeah, right!)
  ;
  file_h = CreateFile_(@channelname,#GENERIC_WRITE,#FILE_SHARE_READ,0,#OPEN_EXISTING,0,0)
  If file_h <= 0
    x_retval = -1
  Else
    ;
    ; turn the string into the requested format (unicode, asci, etc.) then write it out to the mailslot
    ;
    message_l = StringByteLength(message,mode)
    message_p = AllocateMemory(message_l+2)
    PokeS(message_p,message,message_l,mode)
    WriteFile_(file_h,message_p,message_l,@x_retval,0)
    CloseHandle_(file_h)
    FreeMemory(message_p)
  EndIf
  ;
  ; send out a broadcast message if requested, see x_firstinstance()
  ;
  If x_firstinstance_wakeupmessage > 0
    PostMessage_(#HWND_BROADCAST,x_firstinstance_wakeupmessage,0,0)
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_mailslot_getmessage(channelhandle.i,mode.i=#PB_UTF8)                           ; get message from specified mailslot
  Protected message_l.i, message_number.i, message_p.i
  ; Global x_retval_string.s, x_retval.i
  ;
  ; *** get message from specified mailslot
  ;
  ; in:      channelhandle.i       - channel as returned by x_mailslot_open
  ;          [ mode.i ]            - unicode mode
  ; retval:  .i = n                - length in BYTES of retrieved message
  ;             = 0                - no messages
  ;             = -1               - error
  ; out:     x_retval.i            - as retval
  ;          x_retval_string.s     - string received
  ;
  ;
  message_l = 0          ; size of message to read
  message_number = 0     ; number of messages in the queue (not used)
  x_retval = 0           ; size of message (in bytes) read, or the error code
  ;
  If GetMailslotInfo_(channelhandle,0,@message_l,@message_number,0) = 0
    ;
    ; error
    ;
    x_retval = -1
    x_retval_string = ""
    ;
  ElseIf message_l <= 0
    ;
    ; nothing in there
    ;
    x_retval = 0
    x_retval_string = ""
  Else
    ;
    ; read a message, convert from ascii, utf8 etc.
    ;
    message_p = AllocateMemory(message_l+2)
    ReadFile_(channelhandle,message_p,message_l,@x_retval,0)
    x_retval_string = PeekS(message_p,x_retval,mode)
    FreeMemory(message_p)
    ;
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_timer()                                                                        ; replacement for old gfabasic timer statement
  ;
  ; *** replacement for the old gfabasic timer statement
  ;
  ProcedureReturn GetTickCount_()
EndProcedure

Procedure.i x_runprogramex(executable.s,parameters.s="",working.s="",flags.i=-1,sender.i=-1)
  Protected temp.s
  ; Global x_retval.i
  ;
  ; *** execute a program, adjust for eventual 64 bit vs. 32 bit path
  ;
  ; simple replacement for the RunProgram() command
  ;
  ; execute specified program, if the specified exe isn't found, then check if it exists with or without '(x86') in its path, if so execute
  ; note that this does NOT support all possible parameters of the RunProgram() command
  ;
  ; in:          executable.s       - program to execute
  ;              [ paramaters.s ]   - command line parameters
  ;              [ working.s    ]   - working directory
  ;              [ flags.i      ]   - see purebasic command RunProgram()
  ;              [ sender.i     ]   - see purebasic command RunProgram()
  ;
  ; retval:      0                  - error
  ;              <> 0               - successful
  ;
  If x_exist(executable)
  Else
    temp = ReplaceString(executable,"\Program Files (x86)\","\Program Files\",#PB_String_NoCase)
    If x_exist(temp)
      executable = temp
    Else
      temp = ReplaceString(executable,"\Program Files\","\Program Files (x86)\",#PB_String_NoCase)
      If x_exist(temp)
        executable = temp
      EndIf
    EndIf
  EndIf
  ;
  If sender.i <> -1
    x_retval = RunProgram(executable,parameters,working,flags,sender)
  ElseIf flags.i <> -1
    x_retval = RunProgram(executable,parameters,working,flags)
  ElseIf parameters <> "" Or working <> ""
    x_retval = RunProgram(executable,parameters,working)
  Else
    x_retval = RunProgram(executable)
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_runprogram(cmdline.s,flags.i=0)                                                ; runs a commandline
  Protected programname.s, parameters.s, workingdirectory.s
  ; Global x_retval.i
  ;
  ; *** runs a commandline, takes a commandline and splits it into command and parameters
  ;
  ; in:    cmdline.s     - command line, the first part (the executable) may be within quotes
  ;        [ flags.i ]   - see purebasic command RunProgram(), default 0
  ;
  ;
  ; a single string is sliced into parameters for the RunProgram() command
  ;
  cmdline = Trim(cmdline)
  ;
  If x_parse(cmdline,Chr(34),#x_parse_anycase|#x_parse_trim|#x_parse_eol) And x_parsed = ""
    If x_parsenext() And x_parsed > ""
      programname = x_parsed
      parameters = x_parse
    EndIf
  EndIf
  ;
  ; a long line including spaces that points to a file, but without surrounding quotes
  ;
  If programname = "" And x_exist(cmdline)
    programname = cmdline
    parameters = ""
  EndIf
  ;
  ; a line with command and parameters, command may not include spaces...
  ;
  If x_parse(cmdline," ",#x_parse_anycase|#x_parse_trim|#x_parse_eol) And x_parsed > ""
    programname = x_parsed
    parameters = x_parse
  EndIf
  ;
  ; did we find something and can we execute it?
  ;
  x_retval = 0
  If programname > ""
    workingdirectory = GetPathPart(programname)
    x_retval = x_runprogramex(programname,parameters,workingdirectory,flags)
  EndIf
  ProcedureReturn x_retval
  ;
EndProcedure


; *** instances / elevation / system folders / shortcuts

Procedure.i x_firstinstance(instancename.s,wakeupmessage.s="",mode.i=#PB_Ascii)              ; claim mutex, try to become first instance
  Protected lasterror.i
  ; Global x_x_firstinstance_instancename.s, x_firstinstance_mutex_h.i, x_firstinstance_wakeupmessage.i, x_retval.i
  ;
  ; *** claim mutex, try to become first instance
  ;
  ; in:      instancename.s        - name of instance
  ;          [ wakeupmessage.s ]   - message that will be registered with windows and broadcasted to all apps upon activity, default none
  ;          [ mode.i ]            - unicode mode (not used, ignore)
  ; retval:  -2                    - error using / creating mutex
  ;          -1                    - another instance is already host
  ;          0                     - this instance is now host
  ;
  ; notes:
  ;
  ; - don't forget to call x_endfirstinstance() when done (to free up the mutex)
  ; - the 'wakeup' message should be as unique as possible
  ; - instancename.s will become part of the channel (mailslot) name, unless otherwise specified in x_mailslot_sendmessage()
  ; - regardless of being the first instance or not, the application will try to register the wakeupmessage
  ;
  ; I was actually considering calling this one 'take the bridge number one' :-)
  ;
  ; example code:
  ;
  ;   If x_firstinstance("appname","wakeup") = 0
  ;     ;
  ;     ; first instance
  ;     ;
  ;     h = x_mailslot_open("command")
  ;     Repeat
  ;       event = WaitWindowEvent(1024)
  ;       If x_mailslot_getmessage(h) > 0
  ;         ;
  ;         ; option 1 - process any messages received via the mailslot here, for example command line parameters etc.
  ;         ;
  ;       EndIf
  ;       If event = x_firstinstance_wakeupmessage
  ;         x_mailslot_getmessage(h)
  ;         ;
  ;         ; option 2 - instead of checking for any messages in the mailslot as above we could also opt for reacting to a broadcast message
  ;         ;
  ;       EndIf
  ;     Until event = #PB_Event_CloseWindow
  ;     CloseWindow(1)
  ;     x_mailslot_close(h)
  ;     x_endfirstinstance()
  ;     ;
  ;   Else
  ;     ;
  ;     ; later instance
  ;     ;
  ;     x_mailslot_sendmessage("...")
  ;     ;
  ;   EndIf
  ;
  ;
  ; always register the broadcast message, if one was given
  ;
  If wakeupmessage <> ""
    x_firstinstance_wakeupmessage = RegisterWindowMessage_(wakeupmessage)
  Else
    x_firstinstance_wakeupmessage = -1
  EndIf
  ;
  ; try to create a mutex
  ;
  x_x_firstinstance_instancename = instancename
  x_firstinstance_mutex_h = CreateMutex_(0,#True,@instancename)
  lasterror = GetLastError_()
  ;
  If x_firstinstance_mutex_h = 0
    ;
    ; error creating mutex
    ;
    x_retval = -2
    ;
  ElseIf lasterror = #ERROR_ALREADY_EXISTS
    ;
    ; mutex already in use
    ;
    CloseHandle_(x_firstinstance_mutex_h)
    x_retval = -1
    ;
  Else
    ;
    x_retval = 0
    ;
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure x_endfirstinstance()                                                               ; release mutex, stop being first instance
  ; Global x_firstinstance_mutex_h.i
  ;
  ; *** releasemutex, stop being first instance
  ;
  If x_firstinstance_mutex_h <> 0
    ReleaseMutex_(x_firstinstance_mutex_h)
    CloseHandle_(x_firstinstance_mutex_h)
  EndIf
  ;
EndProcedure

Procedure x_wakeupfirstinstance()                                                            ; send message to first instance to wake it up
  ;
  ; *** send message to first instance to wake it up
  ;
  ; notes:
  ;
  ; - this would wake up any 'first instance' of the application, and can be used to speed up mailslot handling etc.
  ;
  If x_firstinstance_wakeupmessage > 0
    PostMessage_(#HWND_BROADCAST,x_firstinstance_wakeupmessage,0,0)
  EndIf
  ;
EndProcedure

Procedure.s x_getspecialfolderlocation(type.i)                                               ; get special folder location
  Protected folderid.i
  ;
  ; *** retrieve a special folder name
  ;
  ; in:      type.i            - see notes
  ; retval:  .s                - string containing the path
  ; out:     x_retval_string.s - as retval
  ;          x_retval.i        - 0 aka #TRUE if successful
  ;
  ; notes:
  ;
  ; - alternative code in x_getknownfolderpath()
  ; - CSIDL has been superceeded by KNOWNFOLDERID (though CSIDL still works)
  ; - example code at the end of this procedure
  ; - for more information on possible values for type.i see the following URL's:
  ;
  ;     CSIDL - http://msdn.microsoft.com/en-us/library/bb762494%28VS.85%29.aspx
  ;     KNOWNFOLDERID - http://msdn.microsoft.com/en-us/library/bb762584%28VS.85%29.aspx
  ;
  ;
  ; first retrieve a pointer to the itemidlist structure of a 'special folder'
  ; followed by retrieving the path from that imtemidlist structure
  ;
  x_retval = SHGetSpecialFolderLocation_(0,type,@folderid)
  If x_retval = 0
    x_retval_string = Space(#MAX_PATH)
    If SHGetPathFromIDList_(folderid,@x_retval_string) = 0
      x_retval = 1
    ElseIf Right(x_retval_string,1) <> "\"
      x_retval_string = x_retval_string+"\"
    EndIf
    ;
    ; free the itemidlist
    ;
    CoTaskMemFree_(folderid)
  EndIf
  If x_retval <> 0
    x_retval_string = ""
  EndIf
  ProcedureReturn x_retval_string
  ;
  ; example usage:
  ;
  ;   AddGadgetItem(1,-1,"")
  ;   AddGadgetItem(1,-1,"ProgramFilename()"+#LF$+ProgramFilename())
  ;   AddGadgetItem(1,-1,"GetCurrentDirectory()"+#LF$+GetCurrentDirectory())
  ;   AddGadgetItem(1,-1,"GetHomeDirectory()"+#LF$+GetHomeDirectory())
  ;   AddGadgetItem(1,-1,"GetTemporaryDirectory()"+#LF$+GetTemporaryDirectory())
  ;   ;
  ;   AddGadgetItem(1,-1,"")
  ;   AddGadgetItem(1,-1,"SHGetSpecialFolderLocation_()")
  ;   AddGadgetItem(1,-1,"  #CSIDL_PERSONAL"+#LF$+x_getspecialfolderlocation(#CSIDL_PERSONAL))
  ;   AddGadgetItem(1,-1,"  #CSIDL_APPDATA"+#LF$+x_getspecialfolderlocation(#CSIDL_APPDATA))
  ;   AddGadgetItem(1,-1,"  #CSIDL_LOCAL_APPDATA"+#LF$+x_getspecialfolderlocation(#CSIDL_LOCAL_APPDATA))
  ;   AddGadgetItem(1,-1,"  #CSIDL_COMMON_APPDATA"+#LF$+x_getspecialfolderlocation(#CSIDL_COMMON_APPDATA))
  ;   AddGadgetItem(1,-1,"  #CSIDL_WINDOWS"+#LF$+x_getspecialfolderlocation(#CSIDL_WINDOWS))
  ;   AddGadgetItem(1,-1,"  #CSIDL_SYSTEM"+#LF$+x_getspecialfolderlocation(#CSIDL_SYSTEM))
  ;   AddGadgetItem(1,-1,"  #CSIDL_PROGRAM_FILES_COMMON"+#LF$+x_getspecialfolderlocation(#CSIDL_PROGRAM_FILES_COMMON))
  ;   AddGadgetItem(1,-1,"  #CSIDL_COMMON_DOCUMENTS"+#LF$+x_getspecialfolderlocation(#CSIDL_COMMON_DOCUMENTS))
  ;   AddGadgetItem(1,-1,"  #SHGFP_TYPE_CURRENT"+#LF$+x_getspecialfolderlocation(#SHGFP_TYPE_CURRENT))
  ;   AddGadgetItem(1,-1,"  #SHGFP_TYPE_DEFAULT"+#LF$+x_getspecialfolderlocation(#SHGFP_TYPE_DEFAULT))
  ;   ;
  ;   AddGadgetItem(1,-1,"")
  ;   AddGadgetItem(1,-1,"GetEnvironmentVariable()")
  ;   AddGadgetItem(1,-1,"  USERPROFILE"+#LF$+GetEnvironmentVariable("USERPROFILE"))
  ;   AddGadgetItem(1,-1,"  APPDATA"+#LF$+GetEnvironmentVariable("APPDATA"))
  ;   AddGadgetItem(1,-1,"  HOMEPATH"+#LF$+GetEnvironmentVariable("HOMEPATH"))
  ;   AddGadgetItem(1,-1,"  TEMP"+#LF$+GetEnvironmentVariable("TEMP"))
  ;   AddGadgetItem(1,-1,"  TMP"+#LF$+GetEnvironmentVariable("TMP"))
  ;   AddGadgetItem(1,-1,"  CLIENTNAME"+#LF$+GetEnvironmentVariable("CLIENTNAME"))
  ;   AddGadgetItem(1,-1,"  SYSTEMDRIVE"+#LF$+GetEnvironmentVariable("SYSTEMDRIVE"))
  ;   AddGadgetItem(1,-1,"  SYSTEMROOT"+#LF$+GetEnvironmentVariable("SYSTEMROOT"))
  ;   AddGadgetItem(1,-1,"  ALLUSERSPROFILE"+#LF$+GetEnvironmentVariable("ALLUSERSPROFILE"))
  ;   AddGadgetItem(1,-1,"  PATHEXT"+#LF$+GetEnvironmentVariable("PATHEXT"))
  ;
EndProcedure

Procedure.i x_relaunchandelevate(param.s="")                                                 ; (re)launch as elevated user
  Protected info.SHELLEXECUTEINFO, exe.s
  ;
  ; *** (re)launch as elevated user
  ;
  ; in:       param.s     - parameters passed on
  ; retval:   .i          - true if succesfull
  ; out:                  - none
  ;
  ; notes:
  ;
  ; - carefull when executing under a 'runas' as you may be a different user
  ; - use the optional param.s to pass along startup parameters and other info you want to pass along
  ; - original code by rescator
  ;
  ; example usage:
  ;
  ; - see x_isuseradmin() for an example of use
  ;
  ;
  exe.s = ProgramFilename()
  info.SHELLEXECUTEINFO
  info\hwnd            = GetForegroundWindow_()
  info\cbSize          = SizeOf(SHELLEXECUTEINFO)
  info\lpVerb          = @"runas"
  info\lpFile          = @exe.s
  info\lpParameters    = @param.s
  info\nShow           = #SW_SHOWNORMAL
  ;
  ProcedureReturn ShellExecuteEx_(info)
  ;
  ; notice by the original author:
  ;
  ; copyright (c) roger hagensen, emsai 2009
  ;
  ; this software is provided 'as-is', without any express or implied
  ; warranty. in no event will the authors be held liable for any damages
  ; arising from the use of this software.
  ;
  ; permission is granted to anyone to use this software for any purpose,
  ; including commercial applications, and to alter it and redistribute it
  ; freely, subject to the following restrictions:
  ;
  ; 1. the origin of this software must not be misrepresented; you must not
  ; claim that you wrote the original software. if you use this software
  ; in a product, an acknowledgment in the product documentation would be
  ; appreciated but is not required.
  ;
  ; 2. altered source versions must be plainly marked as such, and must not be
  ; misrepresented as being the original software
  ;
  ; 3. this notice may not be removed or altered from any source
  ; distribution
  ;
EndProcedure

Procedure.s x_getusername()                                                                  ; retrieve name of the user that started current program
  ;
  ; *** retrieve name of the user that started current program
  ;
  ; in:               - none
  ; retval:     .s    - username who started current program
  ; out:              - none
  ;
  ; notes:
  ;
  ; - alternative: GetEnvironmentVariable("USERNAME")
  ; - based on code by rescator, see x_relaunchandelevate()
  ;
  Protected result.s, size.i
  size.i = #UNLEN
  result.s = Space(size)
  If Not GetUserName_(@result.s,@size.i)
    result.s = ""
  EndIf
  ProcedureReturn result.s
EndProcedure

CompilerIf #PB_Compiler_Processor = #PB_Processor_x86                                        ; used by x_is64bitos()
  Prototype.l IsWow64Process_(hProcess.i,*Wow64Process)
  Prototype.i GetNativeSystemInfo_(*si.SYSTEM_INFO)
CompilerEndIf

Procedure.i x_is64bitos()                                                                    ; check if current program is running on a 64 bit os
  ;
  ; *** check if current program is running on a 64 bit os
  ;
  ; in:              - none
  ; retval:          - true if running on a 64 bit os
  ; out:             - none
  ;
  ; notes:
  ;
  ; - based on code by rescator, see x_relaunchandelevate()
  ;
  CompilerIf #PB_Compiler_Processor = #PB_Processor_x64
    ;
    ; 64 bit code can only run on a 64 bit os, so if this section runs it must be a 64 bit os
    ;
    ProcedureReturn #True
    ;
  CompilerElse
    ;
    ; 32 bit code could be running on a 64 bit os, so check if that is the case
    ;
    #PROCESSOR_ARCHITECTURE_AMD64 = 9
    Protected result.i = #False
    Protected dll.i
    Protected si.SYSTEM_INFO
    Protected iswow64.l = #False
    Protected GetNativeSystemInfo_.GetNativeSystemInfo_
    Protected IsWow64Process_.IsWow64Process_
    ;
    dll = OpenLibrary(#PB_Any,"kernel32.dll")
    If dll
      IsWow64Process_ = GetFunction(dll,"IsWow64Process")               ; only xp sp2 or later, and 2003 sp1 or later have this function
      If IsWow64Process_(GetCurrentProcess_(),@iswow64)
        result = iswow64
      EndIf
      If Not IsWow64Process_
        GetNativeSystemInfo_ = GetFunction(dll,"GetNativeSystemInfo")   ; only windows xp or later has this function
        GetNativeSystemInfo_(si)
        If si\wProcessorArchitecture = #PROCESSOR_ARCHITECTURE_AMD64
          result = #True
        EndIf
      EndIf
      CloseLibrary(dll)
    EndIf
    ProcedureReturn result
    ;
  CompilerEndIf
  ;
EndProcedure

Prototype.l CheckTokenMembership_(TokenHandle.i,SidToCheck.i,*IsMember)                      ; used by x_isuseradmin()
Prototype.i FreeSid_(*pSid)
Prototype.l AllocateAndInitializeSid_(*pIdentifierAuthority,nSubAuthorityCount.b,dwSubAuthority0.l,dwSubAuthority1.l,dwSubAuthority2.l,dwSubAuthority3.l,dwSubAuthority4.l,dwSubAuthority5.l,dwSubAuthority6.l,dwSubAuthority7.l,*pSid)

Procedure.i x_isuseradmin()                                                                  ; is current program a member of local administrator group
  Protected result.i, dll.i
  Protected CheckTokenMembership_.CheckTokenMembership_
  Protected AllocateAndInitializeSid_.AllocateAndInitializeSid_
  Protected FreeSid_.FreeSid_
  Protected *AdministratorsGroup.SID_IDENTIFIER_AUTHORITY
  Protected NtAuthority.SID_IDENTIFIER_AUTHORITY
  ;
  ; *** is current program a member of local administrator group
  ;
  ; in:                      - none
  ; retval:        .i        - true if member of local administrator group
  ; out:                     - none
  ;
  ; notes:
  ;
  ; - based on code by rescator, see x_relaunchandelevate()
  ;
  ; example of use:
  ;
  ;   If Not x_isuseradmin()
  ;     MessageRequester("Example instructions:","You can cancel the elevation and this example will actually handle this."+#LF$+#LF$+"The user will vary depending on the elevation."+#LF$+#LF$+"It's advised to open a window/msgbox before trying the elevation,"+#LF$+"otherwise a canceled/failed elevation may cause the program to open behind other windows,"+#LF$+"XP still has issues with this.")
  ;     If x_relaunchandelevate()
  ;       End
  ;     Else
  ;       MessageRequester("Elevation status:","We are not elevated (failed or canceled)!")
  ;     EndIf
  ;   Else
  ;     MessageRequester("Elevation status:","We are elevated!")
  ;   EndIf
  ;
  ;
  result.i = #True
  ;
  dll = OpenLibrary(#PB_Any,"advapi32.dll")
  If dll
    CheckTokenMembership_ = GetFunction(dll,"CheckTokenMembership")
    AllocateAndInitializeSid_ = GetFunction(dll,"AllocateAndInitializeSid")
    FreeSid_ = GetFunction(dll,"FreeSid")
    If Not (CheckTokenMembership_ Or AllocateAndInitializeSid_ Or FreeSid_)
      result=#False
    EndIf
    If result
      ;
      ; NtAuthority=SECURITY_NT_AUTHORITY [0,0,0,0,0,5]
      ;
      NtAuthority\Value[0]=0
      NtAuthority\Value[1]=0
      NtAuthority\Value[2]=0
      NtAuthority\Value[3]=0
      NtAuthority\Value[4]=0
      NtAuthority\Value[5]=5
      result = AllocateAndInitializeSid_(NtAuthority,2,#SECURITY_BUILTIN_DOMAIN_RID,#DOMAIN_ALIAS_RID_ADMINS,0,0,0,0,0,0,@*AdministratorsGroup)
      ;
      If result And *AdministratorsGroup
        If Not CheckTokenMembership_(#Null,*AdministratorsGroup,@result)
          result = #False
        EndIf
        FreeSid_(*AdministratorsGroup)
      EndIf
    EndIf
    CloseLibrary(dll)
  Else
    result = #False
  EndIf
  ;
  ProcedureReturn result
  ;
EndProcedure

Prototype.i SHGetKnownFolderPath_(*rfid,dwFlags.l,hToken.i,*ppszPath)                        ; used by x_getknownfolderpath()
Prototype.i SHGetFolderPath_(hwndOwner.i,nFolder.l,hToken.i,dwFlags.l,*pszPath)

Procedure.s x_getknownfolderpath(what.i, all.i=#False)                                       ; return standard folder
  Protected folder.s, result.i
  Protected dll.i, rfid.GUID, *path,path.s
  Protected csidl.l
  Protected SHGetKnownFolderPath_.SHGetKnownFolderPath_,SHGetFolderPath_.SHGetFolderPath_
  ;
  ; *** return standard folder
  ;
  ; in:          what.i    - use x_folderxxxx constants below
  ;              all.i     - #true for 'all users'
  ; retval:      .s        - path
  ; out:                   - none
  ;
  ; notes:
  ;
  ; - for alternative code see x_getspecialfolderlocation()
  ; - code by rescator, see also x_relaunchandelevate()
  ;
  #x_folderstartmenu = 1 ; startmenu / programs path
  #x_folderdesktop   = 2 ; desktop path (icons)
  #x_folderprograms  = 3 ; advised install location for applications / programs
  #x_foldersettings  = 4 ; advised location for application settings / configs
  #x_folderdocuments = 5 ; documents folder location
  ;
  #KF_FLAG_CREATE = $00008000
  ;
  result.i = #True
  ;
  dll=OpenLibrary(#PB_Any,"shell32.dll")
  If dll
    SHGetKnownFolderPath_ = GetFunction(dll,"SHGetKnownFolderPath")
    If Not SHGetKnownFolderPath_
      result = #False
    EndIf
    If result                                        ; Windows 6.x
      Select what
          Case #x_folderstartmenu
            If all                                       ; FOLDERID_CommonPrograms CSIDL_COMMON_PROGRAMS {0139D44E-6AFE-49F2-8690-3DAFCAE6FFB8}
              rfid\Data1=$0139D44E
              rfid\Data2=$6AFE
              rfid\Data3=$49F2
              rfid\Data4[0]=$86
              rfid\Data4[1]=$90
              rfid\Data4[2]=$3D
              rfid\Data4[3]=$AF
              rfid\Data4[4]=$CA
              rfid\Data4[5]=$E6
              rfid\Data4[6]=$FF
              rfid\Data4[7]=$B8
            Else                                         ; FOLDERID_Programs CSIDL_PROGRAMS {A77F5D77-2E2B-44C3-A6A2-ABA601054A51}
              rfid\Data1=$A77F5D77
              rfid\Data2=$2E2B
              rfid\Data3=$44C3
              rfid\Data4[0]=$A6
              rfid\Data4[1]=$A2
              rfid\Data4[2]=$AB
              rfid\Data4[3]=$A6
              rfid\Data4[4]=$01
              rfid\Data4[5]=$05
              rfid\Data4[6]=$4A
              rfid\Data4[7]=$51
            EndIf
          Case #x_folderdesktop
            If all                                       ; FOLDERID_PublicDesktop CSIDL_COMMON_DESKTOPDIRECTORY {C4AA340D-F20F-4863-AFEF-F87EF2E6BA25}
              rfid\Data1=$C4AA340D
              rfid\Data2=$F20F
              rfid\Data3=$4863
              rfid\Data4[0]=$AF
              rfid\Data4[1]=$EF
              rfid\Data4[2]=$F8
              rfid\Data4[3]=$7E
              rfid\Data4[4]=$F2
              rfid\Data4[5]=$E6
              rfid\Data4[6]=$BA
              rfid\Data4[7]=$25
            Else                                         ; FOLDERID_Desktop CSIDL_DESKTOP {B4BFCC3A-DB2C-424C-B029-7FE99A87C641}
              rfid\Data1=$B4BFCC3A
              rfid\Data2=$DB2C
              rfid\Data3=$424C
              rfid\Data4[0]=$B0
              rfid\Data4[1]=$29
              rfid\Data4[2]=$7F
              rfid\Data4[3]=$E9
              rfid\Data4[4]=$9A
              rfid\Data4[5]=$87
              rfid\Data4[6]=$C6
              rfid\Data4[7]=$41
            EndIf
          Case #x_folderprograms
            If all                                       ; FOLDERID_ProgramFiles CSIDL_PROGRAM_FILES {905e63b6-c1bf-494e-b29c-65b732d3d21a}
              rfid\Data1=$905E63B6
              rfid\Data2=$C1BF
              rfid\Data3=$494E
              rfid\Data4[0]=$B2
              rfid\Data4[1]=$9C
              rfid\Data4[2]=$65
              rfid\Data4[3]=$B7
              rfid\Data4[4]=$32
              rfid\Data4[5]=$D3
              rfid\Data4[6]=$D2
              rfid\Data4[7]=$1A
            Else                                         ; FOLDERID_LocalAppData CSIDL_LOCAL_APPDATA {F1B32785-6FBA-4FCF-9D55-7B8E7F157091}
              rfid\Data1=$F1B32785
              rfid\Data2=$6FBA
              rfid\Data3=$4FCF
              rfid\Data4[0]=$9D
              rfid\Data4[1]=$55
              rfid\Data4[2]=$7B
              rfid\Data4[3]=$8E
              rfid\Data4[4]=$7F
              rfid\Data4[5]=$15
              rfid\Data4[6]=$70
              rfid\Data4[7]=$91
            EndIf
          Case #x_foldersettings
            If all                                       ; FOLDERID_ProgramData CSIDL_COMMON_APPDATA {62AB5D82-FDC1-4DC3-A9DD-070D1D495D97}
              rfid\Data1=$62AB5D82
              rfid\Data2=$FDC1
              rfid\Data3=$4DC3
              rfid\Data4[0]=$A9
              rfid\Data4[1]=$DD
              rfid\Data4[2]=$07
              rfid\Data4[3]=$0D
              rfid\Data4[4]=$1D
              rfid\Data4[5]=$49
              rfid\Data4[6]=$5D
              rfid\Data4[7]=$97
            Else                                         ; FOLDERID_RoamingAppData CSIDL_APPDATA {3EB685DB-65F9-4CF6-A03A-E3EF65729F3D}
              rfid\Data1=$3EB685DB
              rfid\Data2=$65F9
              rfid\Data3=$4CF6
              rfid\Data4[0]=$A0
              rfid\Data4[1]=$3A
              rfid\Data4[2]=$E3
              rfid\Data4[3]=$EF
              rfid\Data4[4]=$65
              rfid\Data4[5]=$72
              rfid\Data4[6]=$9F
              rfid\Data4[7]=$3D
            EndIf
          Case #x_folderdocuments
            If all                                       ; FOLDERID_PublicDocuments CSIDL_COMMON_DOCUMENTS {ED4824AF-DCE4-45A8-81E2-FC7965083634}
              rfid\Data1=$ED4824AF
              rfid\Data2=$DCE4
              rfid\Data3=$45A8
              rfid\Data4[0]=$81
              rfid\Data4[1]=$E2
              rfid\Data4[2]=$FC
              rfid\Data4[3]=$79
              rfid\Data4[4]=$65
              rfid\Data4[5]=$08
              rfid\Data4[6]=$36
              rfid\Data4[7]=$34
            Else                                         ; FOLDERID_Documents #CSIDL_PERSONAL {FDD39AD0-238F-46AF-ADB4-6C85480369C7}
              rfid\Data1=$FDD39AD0
              rfid\Data2=$238F
              rfid\Data3=$46AF
              rfid\Data4[0]=$AD
              rfid\Data4[1]=$B4
              rfid\Data4[2]=$6C
              rfid\Data4[3]=$85
              rfid\Data4[4]=$48
              rfid\Data4[5]=$03
              rfid\Data4[6]=$69
              rfid\Data4[7]=$C7
            EndIf
          Default
            result=#False
      EndSelect
      If result
        result=SHGetKnownFolderPath_(rfid,#KF_FLAG_CREATE,#Null,@*path)
        If (result=#S_OK) And *path
          folder.s=PeekS(*path)
        EndIf
        If *path
          CoTaskMemFree_(*path)
        EndIf
      EndIf
    Else                                             ; legacy (windows 5.x)
      SHGetFolderPath_ = GetFunction(dll,"SHGetFolderPathW")
      If SHGetFolderPath_
        result=#True
      EndIf
      If result
        Select what
            Case #x_folderstartmenu
              If all
                csidl = #CSIDL_COMMON_PROGRAMS
              Else
                csidl = #CSIDL_PROGRAMS
              EndIf
            Case #x_folderdesktop
              If all
                csidl=#CSIDL_COMMON_DESKTOPDIRECTORY
              Else
                csidl=#CSIDL_DESKTOP
              EndIf
            Case #x_folderprograms
              If all
                csidl=#CSIDL_PROGRAM_FILES
              Else
                csidl=#CSIDL_LOCAL_APPDATA
              EndIf
            Case #x_foldersettings
              If all
                csidl=#CSIDL_COMMON_APPDATA
              Else
                csidl=#CSIDL_APPDATA
              EndIf
            Case #x_folderdocuments
              If all
                csidl=#CSIDL_COMMON_DOCUMENTS
              Else
                csidl=#CSIDL_PERSONAL
              EndIf
            Default
              result=#False
        EndSelect
        If result
          path.s=Space(#MAX_PATH)
          If SHGetFolderPath_(#Null,csidl,#Null,#SHGFP_TYPE_CURRENT,@path.s)=#S_OK
            folder.s=Trim(path.s)
          EndIf
        EndIf
      EndIf
    EndIf
    If Len(folder.s)
      If Not (Right(folder.s,1)="\")
        folder.s+"\"
      EndIf
    EndIf
    CloseLibrary(dll)
  EndIf
  ProcedureReturn folder.s
EndProcedure

Procedure.i x_createshelllink(target.s, link.s, arguments.s, description.s, workingfolder.s, iconfile.s, iconindex.i)
  Protected hRes.l, mem.s, ppf.IPersistFile
  ;
  ; *** create a shortcut (on the desktop)
  ;
  ; in:      target.s        - program / object to be launched
  ;          link.s          - place + name of created shortcut, see example below
  ;          arguments.s     - command line parameters
  ;          description.s   - description
  ;          workingfolder.s - working directory
  ;          iconfile.s      - icon file / file containing the icon
  ;          iconindex.i     - icon index in icon file
  ; retval:           - none
  ; out:              - none
  ;
  ; example:
  ;
  ; target.s = "c:\software\utils\wakiewakie.exe"
  ; link.s = x_getspecialfolderlocation(#CSIDL_DESKTOPDIRECTORY) + "wakiewakie demo.lnk"
  ; x_createshelllink(target, link, "help", "", GetPathPart(target), target, 0)
  ;
  ; notes:
  ;
  ; - original code by rashad
  ; - see purebasic forum for alternatives and more inforamtion
  ; - http://www.purebasic.fr/english/viewtopic.php?f=12&t=44477&p=340289#p340289
  ; - http://www.forums.purebasic.com/english/viewtopic.php?f=13&t=44464
  ;
  CompilerIf #PB_Compiler_Unicode
    Protected psl.IShellLinkW
  CompilerElse
    Protected psl.IShellLinkA
  CompilerEndIf
  ;
  DataSection
    CLSID_ShellLink:
    Data.l $00021401
    Data.w $0000,$0000
    Data.b $C0,$00,$00,$00,$00,$00,$00,$46
    IID_IShellLink:
    CompilerIf #PB_Compiler_Unicode
      Data.l $000214F9
    CompilerElse
      Data.l $000214EE
    CompilerEndIf
    Data.w $0000,$0000
    Data.b $C0,$00,$00,$00,$00,$00,$00,$46
    IID_IPersistFile:
    Data.l $0000010b
    Data.w $0000,$0000
    Data.b $C0,$00,$00,$00,$00,$00,$00,$46
  EndDataSection
  ;
  ; make shure COM is active
  ;
  CoInitialize_(0)
  hRes = CoCreateInstance_(?CLSID_ShellLink, 0, 1, ?IID_IShellLink, @psl)
  ;
  If hRes = 0
    psl\SetPath(target)
    psl\SetArguments(arguments)
    psl\SetDescription(description)
    psl\SetWorkingDirectory(workingfolder)
    psl\SetIconLocation(iconfile, iconindex)
    ;
    ; query IShellLink for the IPersistFile interface for saving the
    ; link in persistent storage
    ;
    hRes = psl\QueryInterface(?IID_IPersistFile, @ppf)
    ;
    If hRes = 0
      ;
      ; save the link
      ;
      hRes = ppf\Save(link, #True)
      ppf\Release()
      ;
    EndIf
    psl\Release()
  EndIf
  ;
  ; shut down COM
  ;
  CoUninitialize_()
  ;
  ProcedureReturn hRes
EndProcedure


; *** logfiles

Procedure x_log_set(filename.s="",autocr.i=1,keepopen.i=1,maxsize.i=65536,inhibit.i=0,autotimestamp.i=1,copytomailslot.s="",leader.s="",copytowindow.i=0,resizablewindow.i=0,hiddenwindow.i=0,copytodebug.i=0,copytoconsole.i=0)   ; logfile settings
  ; Global x_log_autocr.i, x_log_autotimestamp.i, x_log_copytodebug.i, x_log_filename.s
  ; Global x_log_file_h.i, x_log_indent.i, x_log_keepopen.i, x_log_isopen.i, x_log_copytowindow.i
  ; Global x_log_resizablewindow.i, x_log_hiddenwindow.i, x_log_inhibit.i, x_log_maxsize.i
  ; Global x_log_copytomailslot.i, x_log_copytoconsole.i, x_log_mailslotname.s, x_log_leader.s
  ;
  ; *** set options
  ;
  ; in:     file:      [ filename.s ]        - "" to keep last opened aka current file, if none was specified before it will be called "log.log"
  ;                    [ autocr.i ]          - true (1 = default) to add crlf to every line, only applies to file output
  ;                    [ keepopen.i ]        - true (1 = default) to keep the file open, false (0) to reopen and close every time (slower but safer)
  ;                    [ maxsize.i ]         - half the logfile if it gets larger than specified size (minimal size 1024+ bytes), default 65536 bytes
  ;                    [ inhibit.i ]         - true (1) to block use of logfile (for use on read only media), false (0 = default) allows file activity
  ;         timestamp: [ autotimestamp.i ]   - true (1 = default) adds timestamp to every x_log() entry, false (0) leaves it out
  ;         mailslot:  [ copytomailslot.s ]  - specify a (windows api) mailslot name to send a copy to ("" = default aka no mailslot, see notes)
  ;                    [ leader.s ]          - string inserted in front of the message sent to the mailslot ("" = default)
  ;         window:    [ copytowindow.i ]    - true (1) to send output to a window which will open on first use, false (0 = default) no window, see also x_log_window()
  ;                    [ resizablewindow.i ] - true (1) to add resize options to log window, false (0 = default) non resizable window
  ;                    [ hiddenwindow.i ]    - true (1) creates a window that is invisble, false (0 = default) uses a regular visible window
  ;         debug:     [ copytodebug.i ]     - true (1) to send output to debug window as well, false (0 = default) keeps the messages to itself
  ;         console:   [ copytoconsole.i ]   - true (1) to send output to the console as well, false (0 = default) do not copy to console
  ; retval: none
  ; out:    none
  ;
  ; notes:
  ;
  ; - CALL THIS FUNCTION AT LEAST ONCE BEFORE USING ANY OTHER X_LOG() ROUTINE!
  ; - TO OUTPUT TO A WINDOW X_LOG_WINDOW() NEEDS TO BE CALLED ASAP AFTER X_LOG_SET!!
  ;
  ; - all x_log() routines that write to the logfile automatically open the file
  ; - autotimestamp and indent will be ignored if autocr is off
  ; - with indent > 0 autotimestamp will only work with x_log_noindent
  ; - x_log_timestamp() ignores indent and timestamp settings and outputs given string with a timestamp
  ; - x_log_tab() ignores indent and timestamp settings and outputs given string with a specified indent
  ; - use the hidden window option if you want to show the logwindow at the end
  ; - mailslots provide a good way for sharing parameters and logs between multiple instances
  ; - the only way to disable mailslots afterwards is by x_log_copytomailslot.i = #False
  ; - you can change the leader in mailslot messages by x_log_leader.s = "your message "
  ; - file format is UTF8 without BOM (which is pretty much like ascii for most purposes :-))
  ;
  ; example:
  ;
  ;   x_log_set("",1,1,65536,0,1,"","",0,1)
  ;   x_log_window(0,0,400,600,#PB_Window_ScreenCentered)
  ;   Repeat
  ;     event = WaitWindowEvent()
  ;     Select event
  ;     Case #PB_Event_SizeWindow
  ;       x_log_window()
  ;     EndSelect
  ;   Until event = #PB_Event_CloseWindow
  ;
  ; related functions:
  ;
  ; the x_log() routines allow flexible creation of logfiles, for different purposes
  ; (yeah, I went a bit overboard, I admit it :-) but hey, it works and allows all sorts of fun stuff)
  ;
  ;   x_log_write()     - writes a string to the logfile, takes into account autocr
  ;   x_log()           - same but takes into account auto timestamp and indent setting
  ;   x_log_tab()       - writes a string but ignores timestamp and indent
  ;   x_log_timestamp() - writes string and timestamp but ignores indent
  ;   x_log_set()       - set options, filename etc.
  ;   x_log_close()     - forces closing
  ;   x_log_delete()    - deletes logfile
  ;   x_log_indent()    - set indent size
  ;   x_log_q()         - display comment plus int in char dec hex bin
  ;   ...               - and more :-)
  ;
  ;
  ; generic
  ;
  If autotimestamp <> -1
    x_log_autotimestamp = x_bool(autotimestamp)
  EndIf
  ;
  If filename.s = ""
    If x_log_filename = ""
      filename = "log.log"
    EndIf
  EndIf
  If filename.s <> ""
    If filename <> x_log_filename
      If x_log_isopen = #True
        x_ll_closefile(x_log_file_h)
        x_log_isopen = #False
      EndIf
      x_log_filename = filename
    EndIf
  EndIf
  If autocr <> -1
    x_log_autocr = x_bool(autocr)
  EndIf
  If keepopen <> -1
    x_log_keepopen = x_bool(keepopen)
  EndIf
  If maxsize <> -1
    x_log_maxsize = maxsize
  EndIf
  If inhibit <> -1
    x_log_inhibit = x_bool(inhibit)
  EndIf
  ;
  ; mailslot
  ;
  If copytomailslot <> ""
    x_log_copytomailslot = #True
    x_log_mailslotname = copytomailslot
    x_log_leader = leader
  EndIf
  ;
  ; window
  ;
  If copytowindow <> -1
    x_log_copytowindow = x_bool(copytowindow)
  EndIf
  If resizablewindow <> -1
    x_log_resizablewindow = x_bool(resizablewindow)
  EndIf
  If hiddenwindow <> -1
    x_log_hiddenwindow = x_bool(hiddenwindow)
  EndIf
  ;
  ; debug
  ;
  If copytodebug <> -1
    x_log_copytodebug = x_bool(copytodebug)
  EndIf
  ;
  ; console
  ;
  If copytoconsole <> -1
    x_log_copytoconsole = x_bool(copytoconsole)
  EndIf
  ;
EndProcedure

Procedure x_log_delete()                                                                     ; close and delete the logfile
  ; Global x_log_file_h.i, x_log_filename.s
  ;
  ; *** close logfile if it was open, then delete logfile if it existed
  ;
  ; note: setting the inhibit flag blocks deletion of the file
  ;
  If x_log_isopen = #True
    x_ll_closefile(x_log_file_h)
    x_log_isopen = #False
  EndIf
  If FileSize(x_log_filename.s) >= 0 And x_log_inhibit = #False
    DeleteFile(x_log_filename.s)
  EndIf
  ;
EndProcedure

Procedure x_log_indent(n.i)                                                                  ; set indent size for logfile
  ; Global x_log_indent.i
  ;
  ; *** set indent size for logfile
  ;
  x_log_indent = n
  ;
EndProcedure

Procedure x_log_window(x.i=-1,y.i=-1,width.i=-1,height.i=-1,flags=0)                         ; open or resize a window to display logfile output
  Protected w_flags.i
  ; Global x_log_window_nr.i, x_log_listicon_nr.i, x_log_listicon_h.i, x_log_copytowindow.i
  ; Global x_log_resizablewindow.i
  ;
  ; *** open or resize log window with set dimensions
  ;
  ; notes:
  ;
  ; - opens a new window if none existed yet, resizes one if it did
  ; - call with -1,-1,-1,-1 to force gadget adjustment (to handle resize messages from within a main event loop)
  ;
  If IsWindow(x_log_window_nr) = 0
    w_flags = #PB_Window_SystemMenu
    w_flags = w_flags | ( x_log_resizablewindow * (#PB_Window_SizeGadget|#PB_Window_MaximizeGadget|#PB_Window_MinimizeGadget) )
    w_flags = w_flags | ( x_log_hiddenwindow * (#PB_Window_Invisible) )
    w_flags = w_flags | flags
    x_log_window_nr = OpenWindow(#PB_Any,x,y,width,height,"Log: "+x_log_filename,w_flags)
    ; ButtonGadget(200,10,10,10,10,"test")
    x_window_todesktop(x_log_window_nr)
    x_log_listicon_nr = ListViewGadget(#PB_Any,1,1,WindowWidth(x_log_window_nr),WindowHeight(x_log_window_nr))
    x_log_listicon_h = GadgetID(x_log_listicon_nr)
    SendMessage_(x_log_listicon_h,#LB_SETHORIZONTALEXTENT,5000,0)
  Else
    If width = -1
      width = WindowWidth(x_log_window_nr)
    EndIf
    If height = -1
      height = WindowHeight(x_log_window_nr)
    EndIf
    If x = -1
      x = WindowX(x_log_window_nr)
    EndIf
    If y = -1
      y = WindowY(x_log_window_nr)
    EndIf
    ResizeWindow(x_log_window_nr,x,y,width,height)
    ResizeGadget(x_log_listicon_nr,1,1,width,height)
  EndIf
  x_log_copytowindow = #True
  ;
EndProcedure

Procedure x_log_hidewindow(hide.i)                                                           ; hide or unhide log window
  ; Global x_log_hiddenwindow.i, x_log_window_nr.i
  ;
  ; *** hide or show logwindow
  ;
  ; in: hide.i = 0       - show
  ;              1       - hide
  ;
  x_log_hiddenwindow = x_bool(hide)
  HideWindow(x_log_window_nr,x_log_hiddenwindow,#PB_Window_NoActivate)
  ;
EndProcedure

Procedure x_log_write(s.s)                                                                   ; write a string to the logfile
  Protected l.i, n.i, m.i, b.i, p.i
  ; Global x_log_autocr.i, x_log_autotimestamp.i, x_log_copytodebug.i, x_log_filename.s
  ; Global x_log_file_h.i, x_log_indent.i, x_log_keepopen.i, x_log_isopen.i, x_log_copytowindow.i
  ; Global x_log_resizablewindow.i, x_log_hiddenwindow.i, x_log_inhibit.i, x_log_maxsize.i
  ; Global x_log_copytomailslot.i, x_log_mailslotname.s, x_log_leader.s, x_log_copytoconsole.i
  ;
  ; *** write a string to the logfile
  ;
  ; in:      s.s         - string to write
  ;
  If x_log_inhibit = #False And x_log_filename<>""
    ;
    ; only do file things if not inhibited and if a filename was given
    ;
    ; step 1: open logfile if not done so yet
    ;
    If x_log_isopen = #False
      x_log_file_h = x_ll_openfile(x_log_filename,#x_file_rw|#x_file_shared)
    EndIf
    l = x_ll_lof(x_log_file_h)
    ;
    ; step 2: trim logfile if required
    ;
    If x_log_maxsize > 0 And l > x_log_maxsize And l > 1024
      x_ll_fileseek(x_log_file_h,x_log_maxsize/2)
      x_ll_readstring(x_log_file_h,#PB_UTF8)
      p = x_ll_loc(x_log_file_h)
      b = l-p
      m = AllocateMemory(b)
      x_ll_readdata(x_log_file_h,m,b)
      x_ll_closefile(x_log_file_h)
      x_log_delete()
      x_log_file_h = x_ll_openfile(x_log_filename,#x_file_rw|#x_file_shared)
      x_ll_writedata(x_log_file_h,m,b)
      FreeMemory(m)
    Else
      x_ll_fileseek(x_log_file_h,l)
    EndIf
    x_log_isopen = #True
  EndIf
  ;
  ; step 3: write to log file
  ;
  If x_log_inhibit = #False
    If x_log_autocr = #True
      x_ll_writestring(x_log_file_h,s+#CRLF$)
    Else
      x_ll_writestring(x_log_file_h,s)
    EndIf
  EndIf
  ;
  ; step 4: close logfile on every write if requested
  ;
  If x_log_keepopen = #False And x_log_isopen = #True
    x_ll_closefile(x_log_file_h)
    x_log_isopen = #False
  EndIf
  ;
  ; step 5: copy log stuff to other receivers, such as debug, log window, console, and mailslot
  ;
  If x_log_copytowindow = #True
    If IsWindow(x_log_window_nr)=0
      x_log_window(9999,9999,GetSystemMetrics_(#SM_CXSCREEN),9999)
    EndIf
    AddGadgetItem(x_log_listicon_nr,-1,s)
  EndIf
  If x_log_copytodebug = #True
    Debug s
  EndIf
  If x_log_copytoconsole = #True
    PrintN(s)
  EndIf
  If x_log_copytomailslot = #True
    x_mailslot_sendmessage(x_log_mailslotname,x_log_leader+s)
  EndIf
  ;
EndProcedure

Macro x_log_linenr()                                                                         ; write line nr of source code to the logfile
  x_log_write("[ "+Str(#PB_Compiler_Line)+" ]")
EndMacro

Procedure x_log(s.s="")                                                                      ; write a string with timestamp or indent to the logfile
  Protected o.s
  ; Global x_log_indent.i, x_log_autocr.i, x_log_autotimestamp.i
  ;
  ; *** add text to logfile
  ;
  ; in:      [ s.s ]                   - string to log
  ; retval:                            - none
  ; out:                               - none
  ;
  ; this will add a text the logfile, or any other specified output, it will add a timestamp if requested...
  ; this is your do it all, handle it all, standard entry for adding entries to the log files...
  ;
  ; notes:
  ;
  ; - see x_log_set() for use of the log functions
  ; - indented lines do not show a timestamp
  ; - standard timestap length is equivalent to 18 spaces, ie. an indent of 18
  ; - x_log_leader.s is only added to messages send to a mailslot
  ; - you can send messages to mailslots by setting x_log_mailslotname.s and calling x_log_mailslot() bypassing x_log_set()
  ;
  If x_log_autocr = #True
    If x_log_indent = 0
      If x_log_autotimestamp = #True
        o = FormatDate("%dd/%mm/%yy %hh:%ii:%ss ",Date())
      EndIf
    Else
      o = Space(x_log_indent)
    EndIf
  EndIf
  x_log_write(o+s)
  ;
EndProcedure

Procedure x_log_timestamp(s.s)                                                               ; write timestamp + string to logfile (no indent)
  ;
  ; *** add entry with timestamp but no indent
  ;
  x_log_write(FormatDate("%dd/%mm/%yy %hh:%ii:%ss ",Date())+s)
  ;
EndProcedure

Procedure x_log_tab(tabsize.i,s.s)                                                           ; write indent + string to logfile (no timestamp)
  ;
  ; *** add entry with specified indent
  ;
  x_log_write(Space(tabsize)+s)
  ;
EndProcedure

Procedure x_log_int(s.s,q.q)                                                                 ; write a comment and int var to logfile
  Protected o.s
  ;
  ; *** write a comment and an int value to the logfile using regular x_log())
  ;
  If q > 32 And q <= 'z'
    o = "'"+Chr(q)+"'"
  ElseIf q >= 0 And q <= 255
    Select q
        Case 0
          o = "#False"
        Case 1
          o = "#True"
        Case 10
          o = "<lf>"
        Case 13
          o = "<cr>"
        Case 32
          o = "<space>"
        Default
          o = "chr("+Str(q)+")"
    EndSelect
  Else
    o = ""
  EndIf
  o = LSet(o,7)
  ;
  x_log(LSet(s,20)+" =  "+o+" "+x_strex(q,"           #")+"  $"+RSet(Hex(q),8,"0")+"  %"+RSet(Bin(q),32,"0"))
  ;
EndProcedure

Procedure x_log_cr()                                                                         ; write cr or force empty line
  ; Global x_log_autocr.i
  ;
  ; *** write an cr or emtpy line to the logfile
  ;
  If x_log_autocr = #True
    x_log_write("")
  Else
    x_log_write(Chr(13)+Chr(10))
  EndIf
EndProcedure

Procedure x_log_close()                                                                      ; close logfile (a logwindow stays open)
  ; Global x_log_file_h.i, x_log_isopen.i
  ;
  ; *** make sure the logfile is closed
  ;
  If x_log_isopen = #True
    x_ll_closefile(x_log_file_h)
    x_log_isopen = #False
  EndIf
  ;
EndProcedure

Procedure x_log_inhibit(inhibit.i)                                                           ; stop and block usage of logfile
  ; Global x_log_inhibit.i
  ;
  x_log_close()
  x_log_inhibit = x_bool(inhibit)
EndProcedure

Procedure x_log_closewindow()                                                                ; close log window
  ; Global x_log_window_nr.i
  ;
  ; *** close log window
  ;
  If IsWindow(x_log_window_nr)<>0
    CloseWindow(x_log_window_nr)
  EndIf
EndProcedure

Procedure x_log_waitwindow()                                                                 ; wait for user to close log window
  Protected event.i
  ; Global x_log_window_nr.i
  ;
  ; *** wait for user to close log window
  ;
  ; note: eats up all events
  ;
  If IsWindow(x_log_window_nr)<>0
    x_log_hidewindow(0)
    x_emptyeventqueue()
    Repeat
      event = WaitWindowEvent()
      If event = #PB_Event_SizeWindow
        x_log_window(-1,-1,-1,-1)
      EndIf
    Until event = #WM_KEYUP Or event = #PB_Event_CloseWindow Or event = #WM_RBUTTONUP
  EndIf
EndProcedure


; *** network

Procedure.i x_checkipaddress(s.s)                                                            ; check if it is a valid ip address
  Protected l.i, n.i, v.i, x.s
  ;
  x_retval = #True
  ;
  If CountString(s,".") <> 3
    x_retval = #False
  EndIf
  ;
  For n = 1 To 4
    l = Len(StringField(s,n,"."))
    v = Val(StringField(s,n,"."))
    If l = 0 Or l > 3
      x_retval = #False
    EndIf
    If v < 0 Or v > 255
      x_retval = #False
    EndIf
  Next n
  ;
  l = Len(s)
  n = 0
  While n < l
    n = n+1
    x = Mid(s,n,1)
    If x = "."
    ElseIf x < "0" Or x > "9"
      x_retval = #False
    EndIf
  Wend
  If s.s = "0.0.0.0"
    x_retval = #False
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.s x_getipaddress(host.s)                                                           ; find (default) ip address for given hostname
  Protected *host.HOSTENT, *m, wsa.WSADATA, s.s
  ;
  ; *** find ip address for given hostname
  ;
  ; notes:
  ;
  ; - based on code by flype
  ; - see also x_ws_nametoip()
  ; - DO NOT USE simultaneously with the x_ws functions
  ; - this procedure interferes with the x_ws functions as it calls wsacleanup_() !!!
  ;
  If host > ""
    If WSAStartup_ ((1<<8|1), wsa.WSADATA) = #NOERROR
      *m = AllocateMemory(#MAX_COMPUTERNAME_LENGTH+1)
      PokeS(*m,host,#MAX_COMPUTERNAME_LENGTH,#PB_Ascii)
      *host.HOSTENT = gethostbyname_(*m)
      WSACleanup_()
      FreeMemory(*m)
      If *host
        ;
        ; purebasic 5.21 LTS seems to have a bug when returning the result of a peeks() to procedurereturn
        ;
        ;   ProcedureReturn PeekS(inet_ntoa_(PeekL(PeekL(*host\h_addr_list))),#MAX_COMPUTERNAME_LENGTH,#PB_Ascii)
        ;
        ; using an intermediate variable seems to fix the problem
        ;
        s = PeekS(inet_ntoa_(PeekL(PeekL(*host\h_addr_list))),#MAX_COMPUTERNAME_LENGTH,#PB_Ascii)
        ProcedureReturn s
      EndIf
    EndIf
  EndIf
EndProcedure

Procedure.s x_getmacaddress(ip.s)                                                            ; find mac address for given ip address
  Protected n.i, i.i, l.i, *ip, *mac, mac.s
  ;
  ; *** find mac address for given ip address
  ;
  ; notes:
  ;
  ; - based on code by flype
  ;
  l = Len(ip)
  n = 6
  ;
  *ip = AllocateMemory(l+1)
  *mac = AllocateMemory(n)
  PokeS(*ip,ip,l,#PB_Ascii)
  If SendARP_(inet_addr_(*ip),0,*mac, @n) = #NO_ERROR
    For i = 0 To n - 2
      mac = mac + RSet(Hex(PeekA(*mac+i)),2,"0") + ":"
    Next
    mac = mac + RSet(Hex(PeekA(*mac+i)),2,"0")
  Else
    mac = ""
  EndIf
  FreeMemory(*ip)
  FreeMemory(*mac)
  ProcedureReturn mac
EndProcedure

Procedure.s x_gethostname()                                                                  ; get host name of this machine
  Protected size.l, host.s
  ;
  ; *** get host name of this machine
  ;
  ; notes:
  ;
  ; - based on code by flype
  ; - there is a purebasic hostname() function but it requires initnetwork()
  ;
  size.l = #MAX_COMPUTERNAME_LENGTH+1
  host.s = Space(size)
  GetComputerName_(@host, @size)
  ProcedureReturn host
EndProcedure

Procedure.i x_ping(ip_address.s, timeout.i=0)                                                ; send imcp echo message ie. ping ip address
  Protected bip.l, icmp_h.i, message.q
  Protected reply.x_ping_reply
  ; Global x_ping_timeout
  ;
  ; *** send imcp echo message ie. ping ip address
  ;
  ; notes:
  ;
  ; - inspired by code by freak and jhpjhp
  ;
  If timeout <= 0
    timeout = x_ping_timeout
  EndIf
  ;
  If CountString(ip_address,".") = 3
    bip = MakeIPAddress(Val(StringField(ip_address,1,".")),Val(StringField(ip_address,2,".")),Val(StringField(ip_address,3,".")),Val(StringField(ip_address,4,".")))
  EndIf
  ;
  message = Random($FFFFF)
  icmp_h = IcmpCreateFile_()
  IcmpSendEcho_(icmp_h,bip,@message,8,0,@reply,SizeOf(x_ping_reply),timeout)
  IcmpCloseHandle_(icmp_h)
  ;
  If reply\reply\status Or PeekQ(@reply\buffer) <> message
    ProcedureReturn -1
  Else
    ProcedureReturn reply\reply\RoundTripTime
  EndIf
  ;
EndProcedure

Procedure.i x_threadping(*ping.x_threadping_status)                                          ; send imcp echo message ie. ping ip address
  Protected reply.x_ping_reply, message.q, icmp_h.i, bip.l
  ; Global x_ping_timeout.i
  ;
  ; *** thread version to achieve async ping (couldn't get icmpsendecho2_() to work)
  ;
  ; in:      threadping_p.i   - address of a x_threadping_status structure
  ; global:  x_ping_timeout.i - must be larger than 0
  ; retval:  n                - threadnumber
  ; out:     \status.i = 0    - in progress
  ;                      -1   - failed (timeout or other error)
  ;                      -2   - other error
  ;                      >0   - reply time (assume a time of 0 is impossible)
  ;
  ; example
  ;
  ;   x_init()
  ;   ping.x_threadping_status
  ;   ping\bip = MakeIPAddress(192,168,0,99)
  ;   ping\status = 0
  ;   x_threadping(@ping)
  ;
  ; or async (in a thread, check the structure field \status for result
  ;
  ;   IncludeFile "..\x_lib\x_lib.pb"
  ;   x_init()
  ;   OpenWindow(1,10,10,100,100,"Test",#PB_Window_ScreenCentered|#PB_Window_SystemMenu)
  ;   ping.x_threadping_status
  ;   ping\bip = MakeIPAddress(192,168,0,99)
  ;   ping\status = 0
  ;   CreateThread(@x_threadping(),@ping)
  ;   Repeat
  ;     event = WaitWindowEvent(1024)
  ;   Until event = #PB_Event_CloseWindow Or ping\status > 0
  ;   x_end()
  ;
  reply.x_ping_reply
  *ping\status = #x_threadping_inprogress
  ;
  message = Random($FFFFF)
  icmp_h = IcmpCreateFile_()
  If IcmpSendEcho_(icmp_h,*ping\bip,@message,8,0,@reply,SizeOf(x_ping_reply),x_ping_timeout) = 0
    *ping\status = #x_threadping_error
  ElseIf reply\reply\status Or PeekQ(@reply\buffer) <> message
    *ping\status = #x_threadping_timeout
  Else
    If reply\reply\RoundTripTime = 0
      *ping\status = 1
    Else
      *ping\status = reply\reply\RoundTripTime
    EndIf
  EndIf
  IcmpCloseHandle_(icmp_h)
  ;
EndProcedure

Procedure.i x_wakeonlan(mac.s, ip.s="255.255.255.255", port.l=0)                             ; send magic packet (wakeonlan)
  Protected p.i, n.i
  Protected bip.l, bport.w, socket.l, keepalive.b, servermode.l, sockaddr.SOCKADDR_IN, send.i
  Protected Dim magicpacket.b(102)
  ;
  ; *** send a magic packet to another computer
  ;
  ; notes
  ;
  ; - need to call either InitNetwork() or x_ws_start() before calling this procedure
  ; - specifying port and IP might be useful on WAN's but for LAN's just broadcast to 255.255.255.255:0
  ;
  ; build the magic packet
  ;
  p = @magicpacket(0)
  ;
  ; header is 16 bytes $FF
  ;
  x_pokehex(p, "$ FF FF FF FF FF FF")
  ;
  ; followed by 16x mac address
  ;
  For n = 1 To 16
    x_pokehex(p+n*6,mac)
  Next n
  ;
  ; open a socket and send the message
  ;
  x_retval = 0
  ;
  socket     = SOCKET_(#AF_INET, #SOCK_DGRAM, #IPPROTO_UDP)
  If socket
    keepalive  = 1
    If setsockopt_(socket, #SOL_SOCKET, #SO_BROADCAST, @keepalive, 1) = 0
      sockaddr.SOCKADDR_IN
      sockaddr\sin_family = #AF_INET
      sockaddr\sin_port   = htons_(port)
      ;
      ; in ascii mode this inet_addr_() works fine, but not in unicode mode...
      ; need to do this the long and ugly way...
      ;
      ;   sockaddr\sin_addr   = inet_addr_(ip)
      ;
      sockaddr\sin_addr = MakeIPAddress(Val(StringField(ip,1,".")),Val(StringField(ip,2,".")),Val(StringField(ip,3,".")),Val(StringField(ip,4,".")))
      ;
      servermode = 0
      If ioctlsocket_(socket,#FIONBIO,@servermode) = 0
        send = sendto_(socket, p, 102, 0, @sockaddr, 16)
        If send=102 And WSAGetLastError_()=0
          x_retval = 1
        EndIf
      EndIf
    EndIf
    closesocket_(socket)
  EndIf
  ;
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_ws_start()                                                                     ; initialize winsock dll and start winsock services
  ; global x_ws_library.i, x_ws_wsastartup.WSASTARTUP
  ;
  ; *** intitialise winsock dll and start winsock services
  ;
  ; in:                                  - none
  ; retval:    <> 0                      - dll number
  ;            =  0                      - failed to load dll
  ; out:       x_ws_library.i              - same as retval
  ;
  ;
  ; *** declare and load winsock library
  ;
  x_ws_library = OpenLibrary(#PB_Any,"WSOCK32.DLL")
  ;
  ; *** start winsock services
  ;
  ; winsock dll: wsastartup
  ;
  WSAStartup_($101,@x_ws_wsastartup)
  ;
  ;
  ; the remainder of this procedure contains data I have not translated yet, or constants already provided by purebasic
  ; i've left this in here when I converted the code from my old ws_lib file for potential future usage / reference
  ;
  ;
  ; *** socket types
  ;
  ; #SOCK_STREAM = 1            ; stream socket
  ; #SOCK_DGRAM = 2             ; datagram socket
  ; #SOCK_RAW = 3               ; raw-protocol interface
  ; #SOCK_RDM = 4               ; reliably-delivered message
  ; #SOCK_SEQPACKET = 5         ; sequenced packet stream
  ;
  ; *** address / protocol families
  ;
  ; AF_INET values are for the address family, and PF_INET values are for the protocol family
  ;
  ; #AF_INET = 2                ; internetwork: UDP, TCP, etc.
  ;
  ; AF_UNSPEC = 0               ; unspecified
  ; AF_UNIX = 1                 ; local to host (pipes, portals)
  ; AF_IMPLINK = 3              ; arpanet imp addresses
  ; AF_PUP = 4                  ; pup protocols: e.g. BSP
  ; AF_CHAOS = 5                ; mit CHAOS protocols
  ; AF_NS = 6                   ; XEROX NS protocols
  ; AF_ISO = 7                  ; ISO protocols
  ; AF_OSI = AF_ISO             ; OSI is ISO
  ; AF_ECMA = 8                 ; european computer manufacturers
  ; AF_DATAKIT = 9              ; datakit protocols
  ; AF_CCITT = 10               ; CCITT protocols, X.25 etc
  ; AF_SNA = 11                 ; IBM SNA
  ; AF_DECnet = 12              ; DECnet
  ; AF_DLI = 13                 ; Direct data link interface
  ; AF_LAT = 14                 ; LAT
  ; AF_HYLINK = 15              ; NSC Hyperchannel
  ; AF_APPLETALK = 16           ; AppleTalk
  ; AF_NETBIOS = 17             ; NetBios-style addresses
  ; AF_MAX = 18
  ;
  ; *** protocols
  ;
  ; #IPPROTO_TCP = 6            ; tcp
  ; #IPPROTO_RAW = 255          ; raw IP packet
  ;
  ; IPPROTO_IP = 0              ; dummy for IP
  ; IPPROTO_ICMP = 1            ; control message protocol
  ; IPPROTO_GGP = 2             ; gateway^2 (deprecated)
  ; IPPROTO_PUP = 12            ; pup
  ; IPPROTO_UDP = 17            ; user datagram protocol
  ; IPPROTO_IDP = 22            ; xns idp
  ; IPPROTO_ND = 77             ; UNOFFICIAL net disk proto
  ; IPPROTO_MAX = 256           ; ?
  ;
  ; *** ioctlsocket
  ;
  ; couldn't get this call to work, use wsaasyncselect instead
  ;
  ; #FIONREAD=1074030207      ; enable / disable non blocking mode
  ;
  ; FIONBIO = 2147772030        ; determine amount of data that can be read
  ; FIONBIO_NONBLOCK = 1        ; use as parameter to switch to non blocking
  ; FIONBIO_BLOCKING = 0        ; use as parameter to siwtch to blocking
  ; FIOASYNC = 125              ; ? not implemented in winsock under windows
  ;
  ; *** setsockopt
  ;
  ; #SOL_SOCKET = $FFFF
  ; #SO_KEEPALIVE = $8          ; send keepalive messages
  ; #SO_LINGER = $80            ; linger on close if unsent data is present
  ; #TCP_NODELAY = $1           ; disable nagle algorithm
  ;
  ; SO_DONTLINGER               ; don't block close() waiting for unsent data to be sent
  ; SO_BROADCAST                ; allow transmission of broadcast messages on the socket
  ; SO_DEBUG                    ; record debugging information
  ;                             ; (equivalent to setting SO_LINGER with l_onoff = zero)
  ; SO_DONTROUTE                ; don't route: send directly to interface
  ; SO_OOBINLINE                ; receive out of band data in the normal data stream
  ; SO_RCVBUF                   ; buffer size for receives
  ; SO_REUSEADDR                ; allow bind to reuse an address
  ; SO_SNDBUF                   ; specify send buffer size
  ;
  ; *** wsaasyncselect
  ;
  ; parameters for notification
  ;
  ; #FD_READ = 1                            ; readiness for reading
  ; #FD_WRITE = 2                           ; readiness for writing
  ; #FD_OOB = 4                             ; arrival of out of band data
  ; #FD_ACCEPT = 8                          ; incoming connects
  ; #FD_CONNECT = 16                        ; completed connections
  ; #FD_CLOSE = 32                          ; socket closure
  ; #FD_QOS = 64                            ; qos changes
  ; #FD_GROUP_QOS = 128                     ; qos changes
  ; #FD_ROUTING_INTERFACE_CHANGE = 256      ; ?
  ; #FD_ADDRESS_LIST_CHANGE = 512           ; ?
  ; #FD_ALL_EVENTS = 1024                   ; all events
  ;
  ; *** send / recv
  ;
  ; #MSG_OOB = $1                           ; process out of band data
  ; #MSG_PEEK = $2                          ; peek at incoming messages
  ; #MSG_DONTROUTE = $4                     ; send without using routing tables
  ; #MSG_PARTIAL = $8000                    ; ?
  ;
  ; *** structures
  ;
  ; TIMEVAL
  ; WSADATA
  ; SOCKADDR
  ; HOSTENT
  ;   \h_name       ; +0  pointer to the official name of the host (pc)
  ;   \h_aliases    ; +4  pointer to a null-terminated array of alternate names
  ;   \h_addrtype   ; +8  type of address being returned (windows sockets always PF_INET)
  ;   \h_length     ; +10 length, in bytes, of each address (for PF_INET, this is always 4)
  ;   \h_addr_list  ; +12 pointer to a null-terminated list of addresses for the host (addresses in network byte order)
  ;
  ; for ws_fd_set see x_init()
  ;
  ; *** error codes (1) used
  ;
  ; regular microsoft c error constants
  ;
  ; #WSABASEERR = 10000
  ; #WSAEINTR = #WSABASEERR+4
  ; #WSAEWOULDBLOCK = #WSABASEERR+35
  ;
  ; winsock equivalents or berkeley constants
  ;
  ; #WSAENOTSOCK = #WSABASEERR + 38
  ;
  ; *** error codes (2) not (yet) used or converted to gfa / pure
  ;
  ; regular microsoft c error constants
  ;
  ; #define WSAEBADF                (WSABASEERR+9)
  ; #define WSAEACCES               (WSABASEERR+13)
  ; #define WSAEFAULT               (WSABASEERR+14)
  ; #define WSAEINVAL               (WSABASEERR+22)
  ; #define WSAEMFILE               (WSABASEERR+24)
  ;
  ; winsock equivalents or berkeley constants
  ;
  ; #define WSAEINPROGRESS          (WSABASEERR+36)
  ; #define WSAEALREADY             (WSABASEERR+37)
  ; #define WSAEDESTADDRREQ         (WSABASEERR+39)
  ; #define WSAEMSGSIZE             (WSABASEERR+40)
  ; #define WSAEPROTOTYPE           (WSABASEERR+41)
  ; #define WSAENOPROTOOPT          (WSABASEERR+42)
  ; #define WSAEPROTONOSUPPORT      (WSABASEERR+43)
  ; #define WSAESOCKTNOSUPPORT      (WSABASEERR+44)
  ; #define WSAEOPNOTSUPP           (WSABASEERR+45)
  ; #define WSAEPFNOSUPPORT         (WSABASEERR+46)
  ; #define WSAEAFNOSUPPORT         (WSABASEERR+47)
  ; #define WSAEADDRINUSE           (WSABASEERR+48)
  ; #define WSAEADDRNOTAVAIL        (WSABASEERR+49)
  ; #define WSAENETDOWN             (WSABASEERR+50)
  ; #define WSAENETUNREACH          (WSABASEERR+51)
  ; #define WSAENETRESET            (WSABASEERR+52)
  ; #define WSAECONNABORTED         (WSABASEERR+53)
  ; #define WSAECONNRESET           (WSABASEERR+54)
  ; #define WSAENOBUFS              (WSABASEERR+55)
  ; #define WSAEISCONN              (WSABASEERR+56)
  ; #define WSAENOTCONN             (WSABASEERR+57)
  ; #define WSAESHUTDOWN            (WSABASEERR+58)
  ; #define WSAETOOMANYREFS         (WSABASEERR+59)
  ; #define WSAETIMEDOUT            (WSABASEERR+60)
  ; #define WSAECONNREFUSED         (WSABASEERR+61)
  ; #define WSAELOOP                (WSABASEERR+62)
  ; #define WSAENAMETOOLONG         (WSABASEERR+63)
  ; #define WSAEHOSTDOWN            (WSABASEERR+64)
  ; #define WSAEHOSTUNREACH         (WSABASEERR+65)
  ; #define WSAENOTEMPTY            (WSABASEERR+66)
  ; #define WSAEPROCLIM             (WSABASEERR+67)
  ; #define WSAEUSERS               (WSABASEERR+68)
  ; #define WSAEDQUOT               (WSABASEERR+69)
  ; #define WSAESTALE               (WSABASEERR+70)
  ; #define WSAEREMOTE              (WSABASEERR+71)
  ;
  ; extended windows sockets errors
  ;
  ; #define WSASYSNOTREADY          (WSABASEERR+91)
  ; #define WSAVERNOTSUPPORTED      (WSABASEERR+92)
  ; #define WSANOTINITIALISED       (WSABASEERR+93)
  ; #define WSAEDISCON              (WSABASEERR+101)
  ; #define WSAENOMORE              (WSABASEERR+102)
  ; #define WSAECANCELLED           (WSABASEERR+103)
  ; #define WSAEINVALIDPROCTABLE    (WSABASEERR+104)
  ; #define WSAEINVALIDPROVIDER     (WSABASEERR+105)
  ; #define WSAEPROVIDERFAILEDINIT  (WSABASEERR+106)
  ; #define WSASYSCALLFAILURE       (WSABASEERR+107)
  ; #define WSASERVICE_NOT_FOUND    (WSABASEERR+108)
  ; #define WSATYPE_NOT_FOUND       (WSABASEERR+109)
  ; #define WSA_E_NO_MORE           (WSABASEERR+110)
  ; #define WSA_E_CANCELLED         (WSABASEERR+111)
  ; #define WSAEREFUSED             (WSABASEERR+112)
  ;
  ; error return codes for gethostbyname() and gethostbyaddr()
  ;
  ; #define WSAHOST_NOT_FOUND       (WSABASEERR+1001)           ; host not found
  ; #define WSATRY_AGAIN            (WSABASEERR+1002)           ; host not found or server fail
  ; #define WSANO_RECOVERY          (WSABASEERR+1003)           ; non recoverable
  ; #define WSANO_DATA              (WSABASEERR+1004)           ; no data record of requested type
  ;
  ; QOS related error return codes
  ;
  ; #define  WSA_QOS_RECEIVERS               (WSABASEERR + 1005)        ; at least one reserve
  ; #define  WSA_QOS_SENDERS                 (WSABASEERR + 1006)        ; at least one path
  ; #define  WSA_QOS_NO_SENDERS              (WSABASEERR + 1007)        ; no senders
  ; #define  WSA_QOS_NO_RECEIVERS            (WSABASEERR + 1008)        ; no receivers
  ; #define  WSA_QOS_REQUEST_CONFIRMED       (WSABASEERR + 1009)        ; reserve confirmed
  ; #define  WSA_QOS_ADMISSION_FAILURE       (WSABASEERR + 1010)        ; not enough resources
  ; #define  WSA_QOS_POLICY_FAILURE          (WSABASEERR + 1011)        ; bad credentials
  ; #define  WSA_QOS_BAD_STYLE               (WSABASEERR + 1012)        ; unknown / conflicting style
  ; #define  WSA_QOS_BAD_OBJECT              (WSABASEERR + 1013)        ; filterspec or providerspec
  ; #define  WSA_QOS_TRAFFIC_CTRL_ERROR      (WSABASEERR + 1014)        ; flowspec
  ; #define  WSA_QOS_GENERIC_ERROR           (WSABASEERR + 1015)        ; general error
  ;
EndProcedure

Procedure.s x_ws_biptoip(bip.l)                                                              ; convert binairy ip address to dotted ip address
  ;
  ; *** convert binairy network order ip address to human readable dotted address
  ;
  ; in:        bip.l  - binary network order ip address
  ; retval:   .s      - string with dotted address
  ;
  ProcedureReturn(Str(PeekA(@bip))+"."+Str(PeekA(@bip+1))+"."+Str(PeekA(@bip+2))+"."+Str(PeekA(@bip+3)))
EndProcedure

Procedure.s x_ws_biptoname(bip.l)                                                            ; find hostname for binairy ip (reverse dns)
  Protected *retval
  Protected s.s
  ;
  ; *** find the name of a specific ip (reverse dns)
  ;
  ; in:        bip.l    - binary network order ip address
  ; retval:    .s       - name of host (reverse dns)
  ;
  ; notes:
  ;
  ; - only ip v4
  ; - the gethostbyaddr function has been deprecated by the introduction of the getnameinfo function
  ; - still seems to work though :-)
  ;
  *retval = gethostbyaddr_(@bip,4,#PF_INET)
  If *retval > 0
    s = PeekS(PeekL(*retval),-1,#PB_Ascii)
    ProcedureReturn s
  Else
    ProcedureReturn ""
  EndIf
EndProcedure

Procedure.l x_ws_iptobip(ip.s)                                                               ; convert dotted ip address to binairy ip address
  Protected buffer_p.i, bip.l
  ;
  ; *** convert dotted ip address to binary form
  ;
  ; in:       ip.s     - string in human readable format
  ; retval:   .l       - ip address in binary form
  ;
  ; notes:
  ;
  ; - only ip v4
  ;
  buffer_p = AllocateMemory(Len(ip)+1)
  PokeS(buffer_p,ip,-1,#PB_Ascii)
  bip = inet_addr_(buffer_p)
  FreeMemory(buffer_p)
  ProcedureReturn bip
EndProcedure

Procedure.l x_ws_nametobip(name.s)                                                           ; find binairy ip for hostname (dns lookup)
  Protected bip_p.l, buffer_p.i
  ;
  ; *** find the bip of a specific server (dns lookup) and return it in binairy form
  ;
  ; in:       name.s       - name of host
  ; retval:   .l           - ip address in binary format
  ;
  ; notes:
  ;
  ; - only ip v4
  ; - gethostbyname_() was deprecated
  ; - still works though :-)
  ;
  buffer_p = AllocateMemory(Len(name)+1)
  PokeS(buffer_p,name,-1,#PB_Ascii)
  bip_p = gethostbyname_(buffer_p)
  FreeMemory(buffer_p)
  If bip_p <> 0
    CompilerIf #PB_Compiler_Processor = #PB_Processor_x64
      ProcedureReturn PeekI(PeekI(PeekI(bip_p+24)))
    CompilerElse
      ProcedureReturn PeekL(PeekL(PeekL(bip_p+12)))
    CompilerEndIf
  Else
    ProcedureReturn 0
  EndIf
EndProcedure

Procedure.s x_ws_nametoip(name.s)                                                            ; find dotted ip for hostname (dns lookup)
  ProcedureReturn(x_ws_biptoip(x_ws_nametobip(name)))
EndProcedure

Procedure.s x_ws_iptoname(ip.s)                                                              ; find hostname for dotted ip (reverse dns)
  Protected s.s
  ;
  s.s = x_ws_biptoname(x_ws_iptobip(ip))
  ProcedureReturn s
EndProcedure

Procedure.s x_ws_hostname()                                                                  ; find hostname of this machine
  Protected size.l, host.s
  ;
  ; notes:
  ;
  ; - there is a purebasic hostname() function but it requires initnetwork()
  ;
  size.l = #MAX_COMPUTERNAME_LENGTH+1
  host.s = Space(size)
  GetComputerName_(@host, @size)
  ProcedureReturn host
EndProcedure

Procedure x_ws_stop()                                                                        ; stop using winsock and close library
  ;
  WSACleanup_()                 ; clean up and deregister winsock stuff
  CloseLibrary(x_ws_library)    ; stop using the dll
  ;
EndProcedure


; *** other

Procedure x_closedebugger()                                                                  ; close the debugger window
  Protected window_h.i
  ;
  ; *** hide the debugger window
  ;
  window_h = FindWindow_(#Null, "PureBasic Debugger")
  If window_h<>0
    SendMessage_(window_h,#WM_CLOSE,0,0)
  EndIf
  window_h = FindWindow_(#Null,"PureBasic - Debug Output")
  If window_h<>0
    SendMessage_(window_h,#WM_CLOSE,0,0)
  EndIf
EndProcedure

Procedure.s x_sqliteversion(instance.i)
  Protected s.s
  ;
  ; *** returns the SQLite version used by PureBasic
  ;
  ; note:
  ;
  ; - library must be initialised with UseSQLiteDatabase() before this function is being called
  ;
  OpenDatabase(instance,":memory:","","")
  DatabaseQuery(instance,"select sqlite_version()")
  NextDatabaseRow(instance)
  s = GetDatabaseString(0,0)
  FinishDatabaseQuery(instance)
  ProcedureReturn s
  ;
EndProcedure

Procedure.i x_rect_over(x.i,y.i,box_x.i,box_y.i,box_width.i,box_height.i)                    ; check if a point is inside a box
  ;
  ; just to save some typing and improve readability of some code
  ;
  x_retval = #False
  If x > box_x And x < box_x+box_width And y > box_y And y < box_y+box_height
    x_retval = #True
  EndIf
  ProcedureReturn x_retval
EndProcedure

Procedure.i x_rect_overlap(x1.i,y1.i,width1.i,height1.i,x2.i,y2.i,width2.i,height2.i)        ; check if two rectangles overlap
  Protected x0.i, y0.i, width0.i, height0.i
  ; Global x_retval.i, x_retval_x.i, x_retval_y.i, x_retval_width.i, x_retval_height.i
  ;
  ; *** check if two rectangles overlap, returns coordinates of overlapping section
  ;
  x_retval= #False
  If x1 <= x2+width2
    If x1+width1 >= x2
      If y1 <= y2+height2
        If y1+height1 >= y2
          ;
          ; this should be overlapping, i hope :-)
          ;
          x_retval = #True
          If x1 > x2
            x_retval_x = x1
          Else
            x_retval_x = x2
          EndIf
          If x1+width1 > x2+width2
            x_retval_width = x2+width2-x_retval_x
          Else
            x_retval_width = x1+width1-x_retval_x
          EndIf
          If y1 > y2
            x_retval_y = y1
          Else
            x_retval_y = y2
          EndIf
          If y1+height1 > y2+height2
            x_retval_height = y2+height2-x_retval_y
          Else
            x_retval_height = y1+height1-x_retval_y
          EndIf
        EndIf
      EndIf
    EndIf
  EndIf
  ProcedureReturn x_retval
EndProcedure

Procedure x_littlehelp(title.s,text.s,pointer.i=0,flags.i=-1)                                ; show a few small help windows
  Protected text_size.i, page_size.i, page_number.i, page_count.i, event.i
  Protected p.i, n.i, l.i,  h.i, width.i, height.i, w_showhelp.i, g_text.i, g_counter.i, g_previous.i, g_next.i, g_done.i, font_nr.i
  Protected Dim page_start.i(20), Dim page_size.i(20)
  ;
  ; *** displays a little window with the given text, and allows to browse through that text
  ;
  ; in:   title.s    - window title
  ;       text.s     - multi line text seperated by CRLF's
  ;       pointer.i  - pointer to text in memory ending on #00 $00, ignored if test.s <> ""
  ;       flags.i    - text format, #PB_Ascii, #PB_Unicode, -1
  ;
  ; typically used in windows / console apps when launched without command line parameters, or parameters such as /help or /?
  ; or when I'm too lazy to include a decent help file
  ;
  If text = "" And pointer > 0
    text = x_peeks(pointer,-1,flags)
  EndIf
  ;
  ExamineDesktops()
  text_size = CountString(text,#CRLF$)+1
  page_size = DesktopHeight(0)/23
  height = page_size*17
  ;
  page_count = 1+(text_size-1)/page_size
  width = 600
  ;
  p = 1
  n = 0
  Repeat
    n = n+1
    page_start.i(n) = p
    l = 0
    Repeat
      p = FindString(text,#CRLF$,p+1)
      l = l+1
    Until l >= page_size Or p = 0
    If p > 0
      page_size(n) = p-page_start(n)
      p = p+2
    Else
      page_size(n) = Len(text)-page_start(n)+1
    EndIf
  Until p <= 0
  ;
  font_nr = LoadFont(#PB_Any,"Courier New",9)
  w_showhelp = OpenWindow(#PB_Any,10,10,width+16,height+16,title,#PB_Window_SystemMenu|#PB_Window_ScreenCentered)
  AddKeyboardShortcut(w_showhelp,#PB_Shortcut_Return,1)
  AddKeyboardShortcut(w_showhelp,#PB_Shortcut_Escape,1)
  AddKeyboardShortcut(w_showhelp,#PB_Shortcut_Space,2)
  AddKeyboardShortcut(w_showhelp,#PB_Shortcut_Down,2)
  AddKeyboardShortcut(w_showhelp,#PB_Shortcut_Up,3)
  ;
  g_text = TextGadget(#PB_Any,16,16,width,height-60,"")
  g_counter = TextGadget(#PB_Any,16,height-12,32,16,"")
  g_previous = ButtonGadget(#PB_Any,width-140,height-12,50,22,"<")
  g_next = ButtonGadget(#PB_Any,width-90,height-12,50,22,">")
  g_done = ButtonGadget(#PB_Any,width-40,height-12,50,22,"Ok")
  SetGadgetFont(g_text,FontID(font_nr))
  SetGadgetFont(g_counter,FontID(font_nr))
  ;
  page_number = 1
  SetGadgetText(g_text,Mid(text,page_start(page_number),page_size(page_number)))
  SetGadgetText(g_counter,Str(page_number)+"/"+Str(page_count))
  Repeat
    If page_number = 1
      DisableGadget(g_previous,1)
    Else
      DisableGadget(g_previous,0)
    EndIf
    If page_number = page_count
      DisableGadget(g_next,1)
    Else
      DisableGadget(g_next,0)
    EndIf
    event = WaitWindowEvent()
    Select event
        Case #PB_Event_Menu
          Select EventMenu()
              Case 1
                event = #PB_Event_CloseWindow
              Case 2
                page_number = page_number+1
                If page_number > page_count
                  page_number = page_count
                EndIf
              Case 3
                page_number = page_number-1
                If page_number < 1
                  page_number = 1
                EndIf
          EndSelect
          SetGadgetText(g_text,Mid(text,page_start(page_number),page_size(page_number)))
          SetGadgetText(g_counter,Str(page_number)+"/"+Str(page_count))
        Case #PB_Event_Gadget
          Select EventGadget()
              Case g_previous
                page_number = page_number-1
              Case g_next
                page_number = page_number+1
              Case g_done
                event = #PB_Event_CloseWindow
          EndSelect
          SetGadgetText(g_text,Mid(text,page_start(page_number),page_size(page_number)))
          SetGadgetText(g_counter,Str(page_number)+"/"+Str(page_count))
    EndSelect
  Until event = #PB_Event_CloseWindow
  CloseWindow(w_showhelp)
  FreeFont(font_nr)
  ;
EndProcedure

Macro x_debug(text,)                                                                         ; show in debugger line number + message
  Debug Str(x_timer())+"["+Str(#PB_Compiler_Line)+"] "+text
EndMacro

; IDE Options = PureBasic 5.61 (Windows - x86)
; CursorPosition = 11
; Folding = AAAAAAAAAAgBAAAAAAAAAAAAAAAAAAAAAAAAAw
; Markers = 3648
; EnableAsm
; Executable = ..\..\x_lib.exe
; CurrentDirectory = C:\software\purebasic\_projects\
; CompileSourceDirectory
; EnableCompileCount = 1747
; EnableBuildCount = 16
; EnableExeConstant
; EnableUnicode